/*
	FORTRAN Interface to Unix qsort functions
*/

#ifdef cray
#define qsorti_ QSORTI
#define qsrtpi_ QSRTPI
#define qsortf_ QSORTF
#define qsrtpf_ QSRTPF
#define qsrt2p_ QSRT2P
#define qsrt2pf_ QSRT2PF
#endif

/*--------------------------------------------------------------------*/

void qsorti_(n,ix)
/*
	CALL QSORTI(N,IX)

	Integer sort, in-place

	INTEGER N - number of elements to be sorted
	INTEGER IX(N) - integer array to be sorted
*/
int *n;
int *ix;
{
	int icompar();
	void qsort();

	qsort( (char *)ix, (unsigned)(*n), sizeof(int), icompar );
}

int icompar(a,b)
/*
	Integer comparison for Unix quick sort routine
*/
int *a,*b;
{
	if (*a < *b)
		return(-1);
	else if (*a > *b) 
		return(1);
	else
		return(0);
}

/*--------------------------------------------------------------------*/

void qsrtpi_(n,ix)
/*
	CALL QSRTPI(N,IX)

	Integer sort, in-place, with permutation returned

	INTEGER N - number of elements to be sorted
	INTEGER IX(2,N) - integer array to be sorted.
		IX(1,I) contains values to be sorted, IX(2,I) is carried along
		with IX(1,I).  If IX(2,I) is initialized to I, then on output
		IX(2,I) will contain the sort permutation.
*/
int *n;
int *ix;
{
	int icompar();
	void qsort();

	qsort( (char *)ix, (unsigned)(*n), 2*sizeof(int), icompar );
}

/*--------------------------------------------------------------------*/

void qsortf_(n,x)
/*
	CALL QSORTF(N,X)

	Real sort, in-place

	INTEGER N - number of elements to be sorted
	REAL X(N) - real array to be sorted
*/
int *n;
float *x;
{
	int fcompar();
	void qsort();

	qsort( (char *)x, (unsigned)(*n), sizeof(float), fcompar );
}

int fcompar(a,b)
/*
	float comparison for Unix quick sort routine
*/
float *a,*b;
{
	if (*a < *b)
		return(-1);
	else if (*a > *b) 
		return(1);
	else
		return(0);
}

/*--------------------------------------------------------------------*/

void qsrtpf_(n,x)
/*
	CALL QSRTPF(N,X)

	Real sort, in-place, with permutations returned

	INTEGER N - number of elements to be sorted
	REAL X(2,N) - real array to be sorted
		X(1,I) contains values to be sorted, X(2,I) is carried along
		with X(1,I).  If X(2,I) is initialized to FLOAT(I), then on 
		output X(2,I) will contain the sort permutation.
*/
int *n;
float *x;
{
	int fcompar();
	void qsort();

	qsort( (char *)x, (unsigned)(*n), 2*sizeof(float), fcompar );
}

/*--------------------------------------------------------------------*/

void qsrt2p_(n,ix,nrec,maxtrc)
/*
	CALL QSRT2P(N,IX,NREC,MAXTRC)

	2-level heirarchical sort, in-place, with permutations returned

	INTEGER N - number of elements to be sorted
	INTEGER IX(3,N) - integer array to be sorted.
		IX(1,I) - First level for sort (i.e. record number)
		IX(2,I) - Second level for sort (i.e. trace number)
		IX(3,I) - If sequential on input (1 - N), will hold 
					permutation on output
	INTEGER NREC - output number of records detected
	INTEGER MAXTRC - output max traces per record
*/
int *n;
int ix[][3];
int *nrec,*maxtrc;
{
	int icompar(), icmpar2();
	void qsort();
	int i, irec, len;
	unsigned nn;
	char *p;

	*nrec = 1;
	*maxtrc = 1;
	if (*n < 2) return;

/* Sort on record number */
	len = 3*sizeof(int);
	qsort( (char *)ix, (unsigned)(*n), len, icompar );

/* Initialize counters and pointer to first record */
	irec = ix[0][0];
	nn = 1;
	p = (char *)(ix);
	
/* Loop and sort within each record */
	for (i=1; i<*n; i++) {
		if (irec != ix[i][0]) {
/* new record detected, sort previous record */
			if (nn > 1) qsort( p, nn, len, icmpar2 );  
/* set new record number, pointer, and counters */
			irec = ix[i][0];
			p = (char *)(&ix[i][0]);
			*nrec += 1;
			if ( nn > (unsigned)(*maxtrc) ) *maxtrc = (int)(nn);
			nn = 0;
		}
		nn++;
	}
/* sort last record and return */
	if (nn > 1) qsort( p, nn, len, icmpar2 );  
	if ( nn > (unsigned)(*maxtrc) ) *maxtrc = (int)(nn);
}

int icmpar2( a, b )
/*
	Second-level compare for heirarchical sort
*/
int *a,*b;
{
/* point to second element */
	a++;
	b++;
/* and compare */
	if (*a < *b)
		return(-1);
	else if (*a > *b) 
		return(1);
	else
		return(0);
}

/*--------------------------------------------------------------------*/

void qsrt2pf_(n,xi,nrec,maxtrc)
/*
	CALL QSRT2PF(N,XI,NREC,MAXTRC)

	2-level heirarchical sort, in-place, with permutations returned

	INTEGER N - number of elements to be sorted
	REAL XI(3,N) - real array to be sorted.
		XI(1,I) - First level for sort (i.e. record number)
		XI(2,I) - Second level for sort (i.e. trace number)
		XI(3,I) - If sequential on input (1 - N), will hold 
					permutation on output
	INTEGER NREC - output number of records detected
	INTEGER MAXTRC - output max traces per record
*/
int *n;
float xi[][3];
int *nrec,*maxtrc;
{
	int fcompar(), fcmpar2();
	void qsort();
	int i, len;
	float rec;
	unsigned nn;
	char *p;

	*nrec = 1;
	*maxtrc = 1;
	if (*n < 2) return;

/* Sort on record number */
	len = 3*sizeof(float);
	qsort( (char *)xi, (unsigned)(*n), len, fcompar );

/* Initialize counters and pointer to first record */
	rec = xi[0][0];
	nn = 1;
	p = (char *)(xi);
	
/* Loop and sort within each record */
	for (i=1; i<*n; i++) {
		if (rec != xi[i][0]) {
/* new record detected, sort previous record */
			if (nn > 1) qsort( p, nn, len, icmpar2 );  
/* set new record number, pointer, and counters */
			rec = xi[i][0];
			p = (char *)(&xi[i][0]);
			*nrec += 1;
			if ( nn > (unsigned)(*maxtrc) ) *maxtrc = (int)(nn);
			nn = 0;
		}
		nn++;
	}
/* sort last record and return */
	if (nn > 1) qsort( p, nn, len, fcmpar2 );  
	if ( nn > (unsigned)(*maxtrc) ) *maxtrc = (int)(nn);
}

int fcmpar2( a, b )
/*
	Second-level compare for heirarchical sort
*/
float *a,*b;
{
/* point to second element */
	a++;
	b++;
/* and compare */
	if (*a < *b)
		return(-1);
	else if (*a > *b) 
		return(1);
	else
		return(0);
}
