C*----------------------------------------------------------------------
C*
C*  NAME:
C*    fanfp, fanfa, fanfb, fanfc - Seismic Processing Benchmark 
C*        process FANF
C*     
C*  SYNOPSIS:
C*    integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
C*    real otr(ldim,maxtrc), ra(nra), sa(nra)
C*    logical abort
C*    call fanfp( ldim, maxtrc, nra, nsa, abort, ipr )
C*    call fanfa( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                abort, ipr )
C*    call fanfb( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                ntri, ntro, abort, ipr )
C*    call fanfc( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                ntro, abort, ipr )
C*
C*  DESCRIPTION: 
C*    Reads data from seismic benchmark format files
C*
C*  PARAMETERS: 
C*    Record 1
C*    Field  Columns   Description
C*      1     11-15    low cut dip in microseconds/ft
C*      2     16-20    low pass dip
C*      3     21-25    high pass dip
C*      4     26-30    high cut dip
C*      5     31-35    Max frequency in Hz
C*      6     36-40    Trace separation in ft
C*      7     41-45    padding factor in time direction
C*      8     46-50    padding factor in distance direction
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*     Include files:
C*       ./coord.inc
C*       ./fanf.inc
C*
C*     Libraries:
C*       libbench.a
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        12/31/91 - C. Mosher, ARCO E&PT 
C*
C*      References:
C*
C*      Modified:
C*
C*----------------------------------------------------------------------

C*----------------------------------------------------------------------
C*   NAME:
C*     fanfp
C*     
C*----------------------------------------------------------------------
C*   DESCRIPTION:
C*
C*   This process reads 'FANF' parameters, checks syntax, and sets
C*   memory requirements for the processing phase
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*      Subroutines:
C*        getparm - read parameter record
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/06/90 - C. Mosher
C*
C*      References:
C*        SPARC System Documentation
C*
C*      Modified:
C*
C*----------------------------------------------------------------------
C
	subroutine fanfp( ldim, maxtrc, nra, nsa, abort, ipr )
C
C  Arguments:
C
C  Output leading dimension of trace array required for the PROC step
	integer LDIM
C  Output max number of traces per group to be output during PROC step
	integer MAXTRC
C  Number of words of reserved storage required in PROC step
	integer NRA
C  Number of words of working storage required in PROC step
	integer NSA
C  Abort flag, set to .TRUE. by this process if errors detected
	logical ABORT
C  Print unit number that this routine should use
	integer IPR

C
C  Common blocks:
C
C  parameters for process FANF
	include 'fanf.inc'
C  data structure definition
	include 'group.inc'

C
C  Local declarations:
C
C  parameter record buffer
	character*80 buf
C  parameter record index
	integer index

C
C  Start of executable statements
C

C  Get parameter record with velocity
	index = 1
	call getparm('FANF',index,buf,ier)
	if (ier.ne.0) then
		write (ipr,*) 'No parameter records for process FANF'
		abort = .true.
		return
	endif

C  Read dip values and padding factors
	read (buf,'(10x,6f5.0,2i5)') dlc,dlp,dhp,dhc,delx,fmax,ilt,ilx
	if (dlc.eq.0 .and. dhc.eq.0 .and. dhp.eq.0 .and. dhc.eq.0) then
		write (ipr,*) ' FANF ERROR: illegal dip range'
		abort = .true.
		return
	endif
	ddl = dlp-dlc
	ddh = dhc-dhp

C  Set transform lengths
	call seiftm(ns,magn,n2)
	magn = magn + ilt
	n2 = 2**magn
	nyq = n2/2 + 1
	call seiftm(ntpg,magm,m2)
	magm = magm + ilx
	m2 = 2**magm
	myq = m2/2 + 1

C  Number of frequencies to process
	df = 1./(0.001*msr*n2)
	if (fmax.ne.0.) then
		nf = 1 + nint(fmax/df)
	else
		nf = nyq
	endif
	fmax = df*(nf-1)
	
	write (ipr,9010) dlc,dlp,dhp,dhc,delx,fmax,n2,m2
 9010 format (' Process FANF    : 2D Fourier Fan Filter',/,
     +        ' Dip pass range  : ',4f8.1,/,
     +        ' Trace separation: ',f8.1,/,
     +        ' Max frequency   : ',f8.1,/,
     +        ' Transform size  : ',2i8)

C  Set wavenumber increment in 1/micro-ft
	dk = 1.e6/(delx*m2)

C  Set memory requirements
	LDIM = ns
	MAXTRC = ntpg
	NRA = 0

C  Need work space to hold transform
	NSA = 2*nf*m2
C  And one additional trace
	ksa = NSA + 1
	NSA = NSA + n2 + 2

	return
	end

C*----------------------------------------------------------------------
C*   NAME:
C*     fanfa, fanfb, fanfc
C*     
C*----------------------------------------------------------------------
C*   DESCRIPTION:
C*
C*   Processing phase routines for fanfing Seismic Benchmark 
C*     Format files
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/01/92 - C. Mosher
C*
C*      References:
C*        Seismic Benchmark Users Guide
C*
C*      Modified:
C*
C*----------------------------------------------------------------------

C*----------------------------------------------------------------------
C*  The 'a' routine does nothing
C*----------------------------------------------------------------------

	subroutine fanfa( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		abort, ipr )
	integer ldim, maxtrc, nra, nsa, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort

	return
	end

C*----------------------------------------------------------------------
C*  The 'b' routine applies nmo
C*----------------------------------------------------------------------

	subroutine fanfb( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntri, ntro, abort, ipr )
	integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort

	include 'group.inc'
	include 'fanf.inc'

C  Forward 2d fft
	call seif2rc(n2, m2, ldim, otr, nf, sa, nf, ns, ntri, sa(ksa))
C  Apply dip filter
	call fanfdip( nf, sa )
C  Inverse 2d fft
	call seif2cr(n2, m2, ldim, otr, nf, sa, nf, ns, ntri, sa(ksa))

	ntro = ntri
	return
	end

C*----------------------------------------------------------------------
C*  'c' routine does nothing
C*----------------------------------------------------------------------
c
	subroutine fanfc( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntro, abort, ipr )
	integer ldim, maxtrc, nra, nsa, ntro, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort

	call sysmsg('Process FANF completed')

	return
	end
C
C*----------------------------------------------------------------------
C
      subroutine fanfdip( ldimc, otr )
C
C  APPLY DIP FILTER TO FK DATA
C
C  LDIMC - Complex leading dimension
C  OTR - input complex transform
C
C  Parameters used from 'fanf.inc':
C  NF,DF - number of frequencies, sample frequency in HZ
C  M2,DK - number of spatial frequencies, wavenumber increment
C  MYQ - spatial nyquist sample
C  FMAX - frequency filter trapezoid
C  DLC,DLP,DHP,DHC - dip filter trapezoid
C  DDL,DDH - dip taper parameters on low and high sides
C
      INTEGER LDIMC
      COMPLEX OTR(LDIMC,*)
      INCLUDE 'fanf.inc'
      INTEGER I,J
      REAL F,DIP
C
      DO 20 I=2,NF
C  CHECK FOR FREQUENCY OUT OF PASS RANGE
          F = DF*(I-1)
          IF (F.GT.FMAX) THEN
              DO J=1,M2
                  OTR(I,J) = (0.,0.)
              ENDDO
              GOTO 20
          ENDIF
C  LOOP OVER K AND SCALE
      DO 10 J=1,M2
C  SET DIP VALUE, VALUES ABOVE SPATIAL NYQUIST ARE NEGATIVE
          IF (J.LE.MYQ) THEN
              DIP = DK*(J-1)/F
          ELSE
              DIP = -DK*(M2-J+2)/F
          ENDIF
C  SCALE ACCORDING TO PASS RANGE
          IF (DIP.LE.DLC .OR. DIP.GE.DHC) THEN
              OTR(I,J) = (0.,0.)
          ELSE IF (DIP.GT.DLC .AND. DIP.LT.DLP) THEN
              OTR(I,J) = OTR(I,J)*(DIP-DLC)/DDL
          ELSE IF (DIP.GT.DHP .AND. DIP.LT.DHC) THEN
              OTR(I,J) = OTR(I,J)*(DHC-DIP)/DDH
          ENDIF
  10  CONTINUE
  20  CONTINUE
      RETURN
      END
C
C*----------------------------------------------------------------------
