C*----------------------------------------------------------------------
C*
C*  NAME:
C*    geomp, geoma, geomb, geomc - Seismic Processing Benchmark 
C*        process GEOM
C*     
C*  SYNOPSIS:
C*    integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
C*    real otr(ldim,maxtrc), ra(nra), sa(nra)
C*    logical abort
C*    call geomp( ldim, maxtrc, nra, nsa, abort, ipr )
C*    call geoma( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                abort, ipr )
C*    call geomb( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                ntri, ntro, abort, ipr )
C*    call geomc( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                ntro, abort, ipr )
C*
C*  DESCRIPTION: 
C*    Generates empty traces and xyz coordinates
C*
C*  PARAMETERS: 
C*    Record 1
C*    Field  Columns   Description
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*     Include files:
C*       ./coord.inc
C*       ./geom.inc
C*
C*     Libraries:
C*       libbench.a
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        12/31/91 - C. Mosher, ARCO E&PT 
C*
C*      References:
C*
C*      Modified:
C*
C*----------------------------------------------------------------------

C*----------------------------------------------------------------------
C*   NAME:
C*     geomp
C*     
C*----------------------------------------------------------------------
C*   DESCRIPTION:
C*
C*   This process reads 'GEOM' parameters, checks syntax, and sets
C*   memory requirements for the processing phase
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*      Subroutines:
C*        getparm - read parameter record
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/06/90 - C. Mosher
C*
C*      References:
C*        SPARC System Documentation
C*
C*      Modified:
C*
C*----------------------------------------------------------------------
C
	subroutine geomp( ldim, maxtrc, nra, nsa, abort, ipr )
C
C  Arguments:
C
C  Output leading dimension of trace array required for the PROC step
	integer LDIM
C  Output max number of traces per group to be output during PROC step
	integer MAXTRC
C  Number of words of reserved storage required in PROC step
	integer NRA
C  Number of words of working storage required in PROC step
	integer NSA
C  Abort flag, set to .TRUE. by this process if errors detected
	logical ABORT
C  Print unit number that this routine should use
	integer IPR

C
C  Common blocks:
C
	include 'coord.inc'
	include 'group.inc'
	include 'para.inc'
	include 'geom.inc'

C
C  Local declarations:
C
C  parameter record buffer
	character*80 buf

C
C  Start of executable statements
C

C  Read first parameter record
	call getparm('GEOM',1,buf,ier)
	if (ier.ne.0) goto 999
	read (buf,'(6x,i1,3x,5i5,1x,a8)') idbug,jmsr,jns,jntpg,jngpl,
     +    jnlin,jtype
	if (jtype.ne.'stak    ') jtype = 'shot    '
C
	write (ipr,9020) jmsr,jns,jntpg,jngpl,jnlin,jtype
 9020 format (/,' Geometry Parameters',/,
     +          ' Sample rate (ms)  : ',i6,/,
     +          ' Samples per trace : ',i6,/,
     +          ' Traces per shot   : ',i6,/,
     +          ' Shots per line    : ',i6,/,
     +          ' Number of lines   : ',i6,/,
     +          ' Group Type        : ',2x,a8)

C  Check for block data allocation
	if (buf(8:10).eq.'BLK') then
		jmode = 1
		call jdcomp( jngpl, jngpn, jnlive )
	else
		jmode = 0
		jnlive = jngpl
	endif
C
C  Read start location for source, source moveup
	call getparm('GEOM',2,buf,ier)
	if (ier.ne.0 .or. buf(8:10).ne.'SXY') goto 999
	read (buf,'(10x,9f5.0)') sxy0,dsxy,dlsxy
	write (ipr,9030) sxy0,dsxy,dlsxy
 9030 format (' Start shot XYZ    : ',3f7.1,/,
     +        ' Shot moveup       : ',3f7.1,/,
     +        ' Shot line moveup  : ',3f7.1)
C
C  Read start location for receiver 1, receiver interval
	call getparm('GEOM',3,buf,ier)
	if (ier.ne.0 .or. buf(8:10).ne.'RXY') goto 999
	read (buf,'(10x,6f5.0)') rxy0,drxy
	write (ipr,9040) rxy0,drxy
 9040 format (' Start rcvr XYZ    : ',3f7.1,/,
     +        ' Receiver moveup   : ',3f7.1)
C
C  Cable moveup
	call getparm('GEOM',4,buf,ier)
	if (ier.ne.0 .or. buf(8:10).ne.'CXY') goto 999 
	read (buf,'(10x,6f5.0)') dcxy,dlcxy
	write (ipr,9050) dcxy,dlcxy
 9050 format (' Cable moveup      : ',3f7.1,/,
     +        ' Cable line moveup : ',3f7.1)
C
C  CDP binning geometry
	call getparm('GEOM',5,buf,ier)
	if (ier.ne.0 .or. buf(8:10).ne.'CDP') goto 999
	read (buf,'(10x,5f5.0,2i5)') x0,y0,dx,dy,rot
	write (ipr,9060) x0,y0,dx,dy,rot
 9060 format (' CDP Grid orgin    : ',2f10.1,/,
     +        ' CDP, Line spacing : ',2f10.1,/,
     +        ' Grid rotation     : ',f10.1)
C  Convert rotation to radians
	rot = rot * atan(1.)/45.

C  Check for ramp specification
	call getparm('GEOM',6,buf,ier)
	if (ier.eq.0 .and. buf(8:10).eq.'RMP') then
		read (buf,'(10x,2f5.0)') xv1,xv2
		if (xv1.eq.0. .and. xv2.eq.0.) xv2 = 1.
		write (ipr,9070) xv1,xv2
 9070   format (' Data ramp values  : ',2f10.1)
	else
		xv1 = 0.0
		xv2 = 0.0
	endif

C  For stacked data generation, shuffle down values
	if (jtype.eq.'stak    ') then
		jntpg = jngpl
		jngpl = jnlin
		jnlin = 1
	endif

C  Set memory requirements
	ldim = jns
	maxtrc = jntpg
	nra = 0
	nsa = 0

C  Set attrinbutes
	call geomatr()

	return
C
C  Error exit
C
 999  continue
	write (ipr,'(/,'' *** GEOM: Geometry syntax error'')')
	abort = .true.
	return
	end

C*----------------------------------------------------------------------
C*   NAME:
C*     geoma, geomb, geomc
C*     
C*----------------------------------------------------------------------
C*   DESCRIPTION:
C*
C*   Processing phase routines for Seismic Benchmark 
C*     geometry generation
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/01/92 - C. Mosher
C*
C*      References:
C*        Seismic Benchmark Users Guide
C*
C*      Modified:
C*
C*----------------------------------------------------------------------

C*----------------------------------------------------------------------
C*  The 'a' routine initializes group and line number
C*----------------------------------------------------------------------

	subroutine geoma( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		abort, ipr )
	implicit none
	integer ldim, maxtrc, nra, nsa, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort

	include 'group.inc'
	include 'para.inc'
	include 'geom.inc'

C  Set start group and increment based on data allocation mode
	if (jmode.eq.0) then
		kgrp1 = node + 1
		kgrpi = nodes
		kgrpn = jngpl
	else
		call jdcomp( jngpl, jngpn, jnlive )
		call jrange( jngpl, kgrp1, kgrpn )
		if (kgrp1.eq.0) kgrp1 = 1
		kgrpi = 1
	endif
	kgroup = kgrp1 - kgrpi
	kline = 1

C  Debug flag
	if (idbug.ne.0 .and. node.eq.master) then
		idbug = ipr
	else
		idbug = 0
	endif

C  Set attributes
	call geomatr()

	return
	end

C*----------------------------------------------------------------------
C*  'b' routine does nothing
C*----------------------------------------------------------------------

	subroutine geomb( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntri, ntro, abort, ipr )
	integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort

	ntro = 0

	return
	end

C*----------------------------------------------------------------------
C*  The 'c' routine generates geometry and empty traces
C*----------------------------------------------------------------------

	subroutine geomc( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntro, abort, ipr )
	integer ldim, maxtrc, nra, nsa, ntro, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort
C
C  Common blocks
	include 'coord.inc'
	include 'group.inc'
	include 'para.inc'
	include 'geom.inc'

c  Increment group counter
	kgroup = kgroup + kgrpi

c  check for line increment
	if (kgroup.gt.kgrpn) then
		kline = kline + 1
		if (kline.gt.nlin) then
			ntro = 0
			call sysmsg('Process GEOM completed')
			return
		endif
		kgroup = kgrp1
	endif
C
C Set attributes
	call geomatr()
C
C  Set first source and receiver
	do i=1,3
		sxyz(i,1) = sxy0(i) + dsxy(i)*(kgroup-1) + dlsxy(i)*(kline-1)
		rxyz(i,1) = rxy0(i) + dcxy(i)*(kgroup-1) + dlcxy(i)*(kline-1)
	enddo
C
C  Loop and set receiver coords
	if (idbug.ne.0) then
		write (idbug,9010) (sxyz(i,1),i=1,3)
 9010	format (/,' Source XYZ = ',3f8.1,/,
     +		' Trace        Receiver XYZ       CDPN  LNNO')
	endif
C
	do j=1,ntpg
C  Increment receiver
		if (gtype.eq.'stak    ') then
C  Coincident source and receive for stacked data
			do i=1,3
				sxyz(i,j) = sxy0(i) + dsxy(i)*(j-1) + 
     +				dlsxy(i)*(kgroup-1)
				rxyz(i,j) = sxyz(i,j)
			enddo
		else
C  increment receiver and replicate source for shot data
			rxyz(1,j) = rxyz(1,1) + drxy(1) * (j-1)
			rxyz(2,j) = rxyz(2,1) + drxy(2) * (j-1)
			rxyz(3,j) = rxyz(3,1) + drxy(3) * (j-1)
			sxyz(1,j) = sxyz(1,1)
			sxyz(2,j) = sxyz(2,1)
			sxyz(3,j) = sxyz(3,1)
		endif
C  Set midpoint coordinates
		call cdplin(sxyz,rxyz(1,j),dpx,dpy,icdp,ilin)
		if (idbug.ne.0) 
     +	write (idbug,'(i6,3f8.1,2i6)') j,(rxyz(i,j),i=1,3),icdp,ilin
	enddo

C  Set ramp if specified
	if (xv1.ne.0. .or. xv2.ne.0.)
     +  call geomramp( ldim, ns, ntpg, otr, xv1, xv2 )
C
C  Set output attributes
	ntro = ntpg
	iline = kline
	igroup = kgroup
	pmode = 0
C
	return
	end

C*----------------------------------------------------------------------

	subroutine cdplin(sxy,rxy,dpx,dpy,icdp,ilin)
c  sxy,rxy - input source,receiver x,y,z coordinates
c  dpx,dpy - output midpoint location
c  icdp,ilin - output cdp and line numbers
	dimension sxy(3),rxy(3)
	include 'coord.inc'
	include 'geom.inc'
c
c  compute midpoint
	dpx = 0.5 * (sxy(1) + rxy(1))
	dpy = 0.5 * (sxy(2) + rxy(2))
c
c  compute angle between line azimuth and midpoint
	xx = dpx - gx0
	yy = dpy - gy0
	if (xx.eq.0.0 .and. yy.eq.0.0) then
		phi = 0.0
	else
		phi = atan2(yy,xx)
	endif
	theta = phi - grot
	r = sqrt(xx*xx + yy*yy)
C
C  set cdp and line numbers
	icdp = 1 + nint(r*cos(theta)/gdx)
	ilin = 1 + nint(r*sin(theta)/gdy)
	return
	end
C
C---------------------------------------------------------------------
C
	subroutine geomramp( ldim, ns, ntpg, otr, xv1, xv2 )
C
C  Set data array to a ramp function
C  ldim - leading dimension of trace array
C  ns - samples per trace
C  ntpg - traces per group
C  otr - trace array
C  xv1 - value at first sample
C  xv2 - value at last sample
C
	implicit none
	integer ldim, ns, ntpg
	real otr(ldim,ntpg), xv1, xv2
C  Local declarations
	real scale
	integer i,j
	
	scale = (xv2 - xv1)/float(ns-1)
	do j=1,ntpg
		do i=1,ns
			otr(i,j) = xv1 + scale*float(i-1)
		enddo
	enddo

	return
	end

C*----------------------------------------------------------------------

	subroutine geomatr()
C
C  Set attributes in system commons
C
	implicit none
	include 'coord.inc'
	include 'group.inc'
	include 'para.inc'
	include 'geom.inc'

C  Set data set attributes
	ns = jns
	msr = jmsr
	ntpg = jntpg
	ngpl = jngpl
	nlin = jnlin
	gtype = jtype
	gx0 = x0
	gy0 = y0
	gdx = dx
	gdy = dy
	grot = rot
	pmode = 0

	return
	end
