C*----------------------------------------------------------------------
C*
C*  NAME:
C*    mg3dp, mg3da, mg3db, mg3dc - Seismic Benchmark process MG3D
C*     
C*  SYNOPSIS:
C*    integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
C*    real otr(ldim,maxtrc), ra(nra), sa(nra)
C*    logical abort
C*    call mg3dp( ldim, maxtrc, nra, nsa, abort, ipr )
C*    call mg3da( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                abort, ipr )
C*    call mg3db( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                ntri, ntro, abort, ipr )
C*    call mg3dc( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                ntro, abort, ipr )
C*
C*  DESCRIPTION: 
C*    3D finite difference depth migration
C*
C*  PARAMETERS: 
C*    Record 1
C*    Field  Columns   Description
C*      1     11-20    migration velocity
C*      2     21-25    max frequency in Hz
C*      3     26-30    maximum time
C*      4     31-35    maximum depth
C*      5     36-40    delta z
C*      6     41-46    delta x
C*      7     46-50    delta y
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*     Include files:
C*       ./coord.inc
C*       ./group.inc
C*       ./para.inc
C*       ./mg3d.inc
C*
C*     Libraries:
C*       libbench.a
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        12/31/91 - C. Mosher, ARCO E&PT 
C*
C*      References:
C*
C*      Modified:
C*
C*----------------------------------------------------------------------

C*----------------------------------------------------------------------
C*   NAME:
C*     mg3dp
C*     
C*----------------------------------------------------------------------
C*   DESCRIPTION:
C*
C*   This process reads 'MG3D' parameters, checks syntax, and sets
C*   memory requirements for the processing phase
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*      Subroutines:
C*        getparm - read parameter record
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/06/90 - C. Mosher
C*
C*      References:
C*        SPARC System Documentation
C*
C*      Modified:
C*
C*----------------------------------------------------------------------
C
	subroutine mg3dp( ldim, maxtrc, nra, nsa, abort, ipr )
	implicit none
C
C  Arguments:
C
C  Output leading dimension of trace array required for the PROC step
	integer LDIM
C  Output max number of traces per group to be output during PROC step
	integer MAXTRC
C  Number of words of reserved storage required in PROC step
	integer NRA
C  Number of words of working storage required in PROC step
	integer NSA
C  Abort flag, set to .TRUE. by this process if errors detected
	logical ABORT
C  Print unit number that this routine should use
	integer IPR

C  Common blocks:
	include 'coord.inc'
	include 'group.inc'
	include 'para.inc'
	include 'mg3d.inc'
C  Local declarations:
	character*80 buf
	integer ier,npass,ifn
	real tmax,pad,df
	integer nvx, nvy, nvz
	real vx1, dvx, vy1, dvy, vz1, dvz
	character*8 vtype

C
C  Start of executable statements
C

C  Read velocity, max frequency, max time, max depth, deltas
	vmig = 0.0
	fmax = 0.0
	tmax = 0.0
	zmax = 0.0
	dx = 0.0
	dy = 0.0
	dz = 0.0
	pad = 1.0
	call sysparm( 'vmig', 'F', vmig, ier )
	call sysparm( 'fmax', 'F', fmax, ier )
	call sysparm( 'tmax', 'F', tmax, ier )
	call sysparm( 'zmax', 'F', zmax, ier )
	call sysparm( 'dx', 'F', dx, ier )
	call sysparm( 'dy', 'F', dy, ier )
	call sysparm( 'dz', 'F', dz, ier )
	call sysparm( 'pad', 'F', pad, ier )

C  Set increments in each direction
	if (dx.eq.0.) dx = gdx
	if (dy.eq.0.) dy = gdy
	if (dz.eq.0.) dz = dx

	if (dx.eq.0. .or. dy.eq.0.) then
		write (ipr,*) 'MG3D error: grid dx or dy spacing is zero'
		write (ipr,*) '            Set parameters dx and dy'
		abort = .true.
		return
	endif

C  Set space axis lengths
	nx = ntpg
	ny = ngpl
	if (zmax.eq.0.) zmax = dx*(nx-1)
	nz = 1 + nint(zmax/dz)

C  Set transform lengths
	if (tmax.eq.0.) tmax = msr*(ns-1)
	nt = 1 + nint(tmax/float(msr))
	call seiftlen( nt, pad, n2 )
	nyq = n2/2 + 1

C  Number of frequencies to process
	dt = 0.001*float(msr)
	df = 1./(dt*n2)
	sw = 8.*atan(1.)*df
	if (fmax.ne.0.) then
		nfmax = 1 + nint(fmax/df)
	else
		nfmax = n2/2
	endif
	fmax = df*(nfmax-1)

C  Check for loop over frequency
	ipass = 1
	npass = 1
	call sysparm( 'ipass', 'I', ipass, ier )
	call sysparm( 'npass', 'I', npass, ier )

	if (npass.gt.1) then
		nf = nfmax/npass
		if (nf*npass .lt. nfmax) nf = nf + 1
	else
		nf = nfmax
	endif
	if1 = 1 + nf * (ipass-1)
	ifn = if1 + nf - 1
	if (ifn.gt.nfmax) nf = nfmax-if1+1

C  Check for application of correction factors
	icorr = 0
	call sysparm( 'FDCompInc', 'I', icorr, ier )
	
C  Debug flag
	idbug = 0
	call sysparm( 'debug', 'C', buf, ier )
	if (ier.eq.0 .and. buf(1:3).eq.'yes') idbug = 1
	if (ier.eq.0 .and. buf(1:3).eq.'all') idbug = 2

	write (ipr,9010) vmig,tmax,fmax,zmax,dz,dx,dy,nf,nx,ny,nz
 9010 format (' Process MG3D    : 3D Finite Difference Migration',/,
     +        ' Migration vel   : ',f8.1,/,
     +        ' Max time        : ',f8.1,/,
     +        ' Max frequency   : ',f8.1,/,
     +        ' Max Depth       : ',f8.1,/,
     +        ' Delta z,x,y     : ',3f8.1,/,
     +        ' Grid f,x,y size : ',3i8,/,
     +        ' Depth steps     : ',i8)

	if (nf.lt.nfmax) write (ipr,9020) nfmax,nf,ipass,npass
 9020 format (/,i5,' Total frequencies, will process ',i5, 
     +    ' Frequencies per loop iteration',/,
     +    ' Pass ',i3,' Of ',i3)
	if (icorr.ne.0) write (ipr,*) 
     +    'Will apply FD compensation operator with increment',icorr

	if (vmig.eq.0) then
		call vsbf_info( vtype, nvx, vx1, dvx, nvy, vy1, dvy,
     +			nvz, vz1, dvz, ier )
		if (ier.ne.0) then
			write (ipr,*) 'MG3D Error: VSBF function not found'
			abort = .true.
			return
		endif
		if (vtype(1:7).ne.'Vint-XY') then
			write (ipr,*) 'MG3D Error: VSBF must be type Vint-XY'
			abort = .true.
			return
		endif
		write (ipr,*) 'Will use velocity function from VSBF'
	endif

C  Initial decomposition is over y direction
	call jdcomp( ny, nypn, nylive )
	nyp = nypn*nodes

C  Decomposition during migration is over frequency
	if (nf.lt.nodes) then
		write (ipr,*) 'MG3D Error: less than 1 frequency per node'
		abort = .true.
	endif
	call jdcomp( nf, nfpn, nflive )
	nfp = nfpn*nodes

C  Will output depth slices
	ldim = nx
	maxtrc = max(nx,nypn)

C  Length of 1 frame during input phase
	lframe = 2*nfp*nx

C  Set memory requirements
C  Transform volume
	NRA = 2*nx*nyp*nfpn

C  Work space wavefield extrapolation
	nsa = 12*nx*ny + 8*nx + 8*ny + 2*max(nx,ny)
	nsa = max( nsa, n2+2 )
	if (icorr.ne.0) then
		call seiftlen( nx, 1.0, nxf )
		call seiftlen( ny, 1.0, nyf )
		nsa = max( nsa, nx*ny + 2*nxf*nyf + 2*max(nxf,nyf) )
	endif
	if (vmig.eq.0.) nsa = max( nsa, nvx*nvy )

C  Work space for distributed transpose
	if (nodes.gt.1) nsa = max( nsa, 2*nx*nfpn*nypn)

C  Set output attributes
	call mg3d_attr()

C  Initilize timers
	tload = 0.0
	tcomp = 0.0
	tcomm = 0.0
	tcorr = 0.0
	tvel = 0.0

	return
	end

C*----------------------------------------------------------------------
C*   NAME:
C*     mg3da, mg3db, mg3dc
C*     
C*----------------------------------------------------------------------
C*   DESCRIPTION:
C*
C*   Processing phase routines for Seismic Benchmark process MG3D
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/01/92 - C. Mosher
C*
C*      References:
C*        Seismic Benchmark Users Guide
C*
C*      Modified:
C*
C*----------------------------------------------------------------------

C*----------------------------------------------------------------------
C*  Work phase initialization
C*----------------------------------------------------------------------

	subroutine mg3da( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		abort, ipr )
	implicit none
C  Arguments
	integer ldim, maxtrc, nra, nsa, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort
C  Common blocks
	include 'para.inc'
	include 'mg3d.inc'
C  Local variables

C  Initialize input/output 2D frame counters
	jframe = 0
	kdepth = 0

C  Initialize stored volume to zero
	call vfill( 0.0, ra, 1, 2*nx*nyp*nfpn )

C  Set decompositions over space and frequency
	call jdcomp( ny, nypn, nylive )
	call jdcomp( nf, nfpn, nflive )

C  Set output attributes
	call mg3d_attr()

C  Debug print level 1 by master only
	if (idbug.eq.1 .and. node.ne.master) idbug = 0

	if (idbug.ne.0) write (ipr,*) 'MG3D: Begin data load phase'

	return
	end

C*----------------------------------------------------------------------
C*  The 'b' routine processes data
C*----------------------------------------------------------------------

	subroutine mg3db( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntri, ntro, abort, ipr )
	implicit none
C  Arguments
	integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort
C  Common blocks
	include 'coord.inc'
	include 'group.inc'
	include 'mg3d.inc'
C  Local variables
	real t0, etimer
	integer k

	t0 = etimer(0)	
C  pointer into array of stored data
	k = jframe * lframe + 1
	jframe = jframe + 1

C  Forward 1D fft to (f,x)
	call mg3d_mrcf( ldim, ntri, otr, nfp, ra(k), n2, ns, nf, if1, sa )

C  Transpose to (x,f)
	call tran21c( nfp, nx, ra(k) )

C  No traces passed out
	ntro = 0
	tload = tload + etimer(1)

	return
	end

C*----------------------------------------------------------------------
C*  Migrate and output migrated traces
C*----------------------------------------------------------------------
c
	subroutine mg3dc( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntro, abort, ipr )
	implicit none
C  Arguments
	integer ldim, maxtrc, nra, nsa, ntro, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort
C  Common blocks
	include 'coord.inc'
	include 'group.inc'
	include 'para.inc'
	include 'mg3d.inc'
C  Local variables
	integer ier
	real z, t0, etimer, flopse, flopsc, flopsm, ratee, ratec

C  Transpose data if first time through
	if (kdepth.eq.0) then
C		call jtrace(1)
		call jsync()
		if (idbug.ne.0) write (ipr,*) 'MG3D: Data load completed'
C  Stored data volume is ( x, f, y ), with y spread across nodes.
C  Use transpose operations to spread frequencies across nodes.
C  ( x, f, y ) -> ( x, y, f )
		t0 = etimer(0)
		call dtran132c( nx, nfp, nyp, ra, sa )
		tload = tload + etimer(1)
		if (idbug.ne.0) write (ipr,*) 'MG3D: Initial transpose complete'
C  Initialize image time
		timage = 0.0
C  If up to number of lines, quit
	else if (kdepth.ge.nz) then
		ntro = 0
		if (node.ne.master) return
		call sysohdr('MG3D')
C  Megaflops from regression on Cray YMP HPM data
		flopse = 1.e-6 * float(nf*nz) * ( 220.0 * float(nx*ny) )
		ratee = flopse/tcomp
		if (icorr.ne.0) then
			flopsc = 1.e-6 * 6.5 * 2. * float(nf*nz/icorr)
     +			* float(nxf) * alog(float(nxf))
     +			* float(nyf) * alog(float(nyf))
			ratec = flopsc/tcorr
		else
			flopsc = 0.0
			ratec = 0.0
		endif
		flopsm = flopse + flopsc
		write (ipr,9010) tload, tcomp, tcorr, tcomm, flopsm, 
     +       ratee, ratec, flopsm/(tcomp + tcorr + tcomm)
 9010   format (' 3D Depth Migration Statistics',/,
     +          ' Load time (sec) : ',f10.1,/,
     +          ' Extrap time     : ',f10.1,/,
     +          ' FDComp time     : ',f10.1,/,
     +          ' Communicate time: ',f10.1,/,
     +          ' Flt pt ops (*E6): ',g12.4,/,
     +          ' Extrap Mflops   : ',f10.2,/,
     +          ' FDComp Mflops   : ',f10.2,/,
     +          ' Effective Mflops: ',f10.2)
		return
	endif

C  Propagate one depth step
	ier = 0
	z = z1 + dz*float(kdepth)
	call mg3d_zstep( ra, sa, ldim, otr, z, ipr, ier )
	if (ier.ne.0) then
		write (ipr,*) 'MG3D Error: velocity file access failed'
		abort = .true.
		return
	endif

C  Set output attributes
	call mg3d_attr()
	kdepth = kdepth + 1
	igroup = kdepth
	iline = 1
	ntro = ny
	abort = .false.

	return
	end
C
C*----------------------------------------------------------------------
C*  Support subroutines
C*----------------------------------------------------------------------
C
C*----------------------------------------------------------------------
C
	subroutine mg3d_zstep( ra, sa, ldim, otr, z, ipr, ier )
C 
C  Propagate wavefield 1 depth step
C
	implicit none
C  Common blocks
	include 'para.inc'
	include 'mg3d.inc'
C  Arguments
	integer ldim, ipr, ier
	complex ra(*)
	real sa(*), otr(ldim,*), z
C  Local declarations
	integer iupd,jf1,jfn,ksa,kwk
	real frac,vbar,sbar,amax
	real t0, etimer

	ier = 0

C  Up/down flag and velocity fraction for extrapolators
	iupd = 1
	frac = 0.5

C  Work space
	ksa = nx*ny + 1
	kwk = ksa + 2*nxf*nyf
	
C  Frequency index range
	call jrange( nfp, jf1, jfn )
C  If no frequencies to be processed on this node, set lower frequency
C  greater than the upper to cause frequency loops to be skipped
	if (jf1.eq.0) jf1 = 1

C  Output zero time if first depth step
	if (z.le.0.) then
		if (idbug.ne.0) then
			call mg3d_amax( nx, nyp, nfpn, ra, amax )
			write (ipr,9010) z, vbar, amax
 9010       format (' Depth ',f8.1,' vbar ',f8.1,' amax ',g12.4)
		endif
		t0 = etimer(0)
		call mg3d_image( nx, ny, nx, nyp, ra, sa, jf1, jfn, sw,
     + 	   frac, timage )
		tcomp = tcomp + etimer(1)
		t0 = etimer(0)
		call mg3d_add( nx, node, nodes, nyp, nypn, sa, ldim, otr )
		tcomm = tcomm + etimer(1)
		return
	endif

C  Get velocity
	t0 = etimer(0)
	call mg3d_vmig(nx, x1, dx, ny, y1, dy, z, sa, vbar, sa(ksa), ier)
	tvel = tvel + etimer(1)
	if (ier.ne.0) return
	if (idbug.ne.0) then
		call mg3d_amax( nx, nyp, nfpn, ra, amax )
		write (ipr,9010) z, vbar, amax
	endif

	t0 = etimer(0)

C  Extrapolate in x and y directions
	call mg3d_xtrap( nx, ny, nx, nyp, ra, dx, dy, dz, jf1, jfn, sw, 
     +    iupd, frac, sa, sa(ksa) )

C  Thin lens term
	call mg3d_lens( nx, ny, nx, nyp, ra, sa, vbar, jf1, jfn, sw,
     +  dz, iupd, frac )

C  Correction factors
	if (icorr.ne.0) then
		sbar = sbar + 1./vbar
		if (mod(kdepth,icorr).eq.0) then
			sbar = sbar/float(icorr)
			if (idbug.ne.0) write (ipr,*) 'Apply correction term'
			t0 = etimer(0)
			call mg3d_corr( nx, ny, nx, nyp, ra, sbar, jf1, jfn,
     +			sw, dx, dy, dz*icorr, nxf, nyf, sa(ksa), sa(kwk) )
			tcorr = tcorr + etimer(1)
			sbar = 0.0
		endif
	endif

C  Apply image condition
	timage = timage + 2.*dz/vbar
	call mg3d_image( nx, ny, nx, nyp, ra, sa, jf1, jfn, sw,
     +    frac, timage )

	tcomp = tcomp + etimer(1)

C  Accumulate partial sums across processors
	t0 = etimer(0)
	call mg3d_add( nx, node, nodes, nyp, nypn, sa, ldim, otr )
	tcomm = tcomm + etimer(1)

	return
	end
C
C*----------------------------------------------------------------------
C
	subroutine mg3d_lens( nx, ny, l1, l2, q0, vel, vbar, kf1, kfn, sw, 
     +    dz, iupd, frac )
C
C  Apply thin lens term 
C
C  nx,ny - grid size
C  q0(l1,l2,*) - input/output wavefield
C  vel(nx,ny) - input velocity field
C  vbar - input average velocity
C  kf1,kfn - frequency index range
C  sw - sample rate in radians
C  dz - depth step size
C  iupd - up/down propagation flag
C  frac - velocity fraction, set to 1.0 for full velocity
C
	implicit none
C  Arguments
	integer nx,ny,l1,l2,kf1,kfn,iupd
	real vel(nx,ny),vbar,sw,dz,frac
	complex q0(l1,l2,*)
C  Local variables
	integer i,j,k,kf
	real dzf,dzv,omega,arg

C  Loop over x location and apply phase factor
	dzf = dz/frac
	dzv = dzf/vbar
	do kf=kf1,kfn
		omega = iupd*sw*float(kf-1)
		k = kf-kf1+1
		do j=1,ny
			do i=1,nx
				arg = omega*( dzf/vel(i,j) - dzv )
				q0(i,j,k) = q0(i,j,k) * 
     +             cmplx( cos(arg), sin(arg) )
			enddo
		enddo
	enddo

	return
	end
C
C*----------------------------------------------------------------------
C
	subroutine mg3d_image( nx, ny, l1, l2, q0, sa, kf1, kfn, sw, 
     +    frac, timage )
C
C  Apply image condition
C
C  nx,ny - grid size
C  q0(nx,ny,*) - input/output wavefield
C  sa(nx,ny) - work array for image partial sums
C  kf1,kfn - frequency index range
C  sw - sample rate in radians
C  frac - velocity fraction, set to 1.0 for full velocity
C  timage - image time
C
	implicit none
C  Arguments
	integer nx,ny,l1,l2,kf1,kfn
	real q0(2,l1,l2,*)
	real sa(nx,ny)
	real sw,frac,timage
C  Local variables
	integer i,j,k,kf
	real arg, carg, sarg

C  Initialize array for accumulating local image sum
	call vfill( 0.0, sa, 1, nx*ny )
C  Loop over frequency band on this processor
	do kf = kf1,kfn
C  Compute frequency * image time using global frequency index
		arg = sw*(kf-1)*timage
		carg = cos(arg)
		sarg = sin(arg)
C  Set local index into wavefield array
		k = kf - kf1 + 1
C  Accumulate real part of phase shifted wavefield
		do j=1,ny
			do i=1,nx
				sa(i,j) = sa(i,j) + carg * q0(1,i,j,k)
     +                            - sarg * q0(2,i,j,k)
			enddo
		enddo
	enddo

	return
	end
C
C*----------------------------------------------------------------------
C
	subroutine mg3d_xtrap( nx, ny, l1, l2, q0, dx, dy, dz, kf1, kfn,
     +    sw, iupd, frac, vel, sa )
C
C  FX domain post-stack wavefield extrapolator
C  Apply 2D operator to subarrays in one direction
C  Apply twice with transpose for approximate 3D extrapolation
C
C  nx,ny - grid size
C  l1,l2 - leading dimensions of wavefield array
C  complex q0(l1,l2,*) - input wavefield at z=z1
C  dx,dy,dz - x, y, z step sizes
C  kf1,kfn - frequency index range
C  sw - frequency sampling (rad/sec)
C  ipud - up/down propagation flag, 1=up, -1=down
C  frac - velocity fraction
C  real vel(nx,ny) - input array of velocity values
C  real wv(ny,nx) - work space for transposed velocity array
C  complex sa(*) - work array, length 6*max(nx,ny)
C
	implicit none
C  Argument declarations
	integer nx,ny,l1,l2,kf1,kfn,iupd
	real dx,dy,dz,sw,frac,vel(nx,ny)
	complex q0(l1,l2,*), sa(*)
C  Local declarations
	integer k, kf
	integer ka, kb, kaa, kbb, kze, kbnx1, kbnxn, kbbnx1, kbbnxn
	integer kbny1, kbnyn, kbbny1, kbbnyn, kzf
	real omega

C  Work array pointers
	ka = 1
	kb = ka + nx*ny
	kaa = kb + nx*ny
	kbb = kaa + nx*ny
	kze = kbb + nx*ny
	kbnx1 = kze + nx*ny
	kbnxn = kbnx1 + ny
	kbbnx1 = kbnxn + ny
	kbbnxn = kbbnx1 + ny
	kbny1 = kbbnxn + ny
	kbnyn = kbny1 + nx
	kbbny1 = kbnyn + nx
	kbbnyn = kbbny1 + nx
	kzf = kbbnyn + nx

C  Extrapolate down one depth step for all non-zero frequencies
	do kf=max(kf1,2),kfn

C  Set frequency using global index
		omega = sw*(kf-1)
C  Set local index into wavefield array
		k = kf - kf1 + 1

C  Set up system
	call seiptri3d( nx, ny, vel, sa(ka), sa(kb), sa(kaa), sa(kbb),
     &      sa(kbnx1), sa(kbnxn), sa(kbbnx1), sa(kbbnxn),
     &      sa(kbny1), sa(kbnyn), sa(kbbny1), sa(kbbnyn),
     &      omega, dz, dx, frac, iupd )

C  Solve system
	call seictri3d( q0(1,1,k), sa(ka), sa(kb), sa(kaa), sa(kbb),
     &      sa(kbnx1), sa(kbnxn), sa(kbbnx1), sa(kbbnxn),
     &      sa(kbny1), sa(kbnyn), sa(kbbny1), sa(kbbnyn),
     &      sa(kze), sa(kzf), nx, ny )

	enddo

	return
	end
C
C*----------------------------------------------------------------------
C
	subroutine mg3d_corr( nx, ny, l1, l2, q0, sbar, jf1, jfn, sw, 
     +    dx, dy, dz, nxf, nyf, sa, wk )
C
C  Apply correction term 
C
C  nx,ny - grid size
C  q0(l1,l2,*) - input/output wavefield
C  sbar - input average slowness (reciprocal velocity)
C  jf1,jfn - frequency index range
C  sw - sample rate in radians
C  dx, dy - grid delta factors
C  dz - depth step size
C  nxf,nyf - fourier transform lengths in x and y directions
C  sa - work array, 2*(nxf*nyf + max(nxf,nyf) )
C  wk - work array, complex length max(nxf,nyf)
C
	implicit none
C  Arguments
	integer nx,ny,l1,l2,jf1,jfn,nxf,nyf
	real sbar,sw,dx,dy,dz
	complex q0(l1,l2,*), sa(nxf,nyf), wk(*)
C  Local variables
	integer jf,kf
	real w
C  Coefficients from FXMG_PREPTRI
	real ALPHA, BETA, GAMMA
	parameter (ALPHA=0.478242060, BETA=0.376369527, GAMMA=0.101) 

C  Loop over frequencies stored on this processor
	do jf=max(2,jf1),jfn
C  Compute local index into distributed memory array
		kf = jf - jf1 + 1
C  Copy to work array
		call seiccopy( nxf, nyf, sa, l1, l2, q0(1,1,kf), nx, ny )
C  Call forward 2D FFT routine for the current frequency slice
		call seif2cc( nxf, nyf, 1, 1, nxf, sa, wk )
C  Apply correction factor
		w = sw*(jf-1)
		call mg3d_lihzhi( nxf, nyf, sa, dx, dy, dz, w, sbar,
     +        ALPHA, BETA, GAMMA )
C  Inverse 2D FFT
		call seif2cc( nxf, nyf, -1, -1, nxf, sa, wk )
C  Copy back to wavefield array
		call seiccopy( l1, l2, q0(1,1,kf), nxf, nyf, sa, nx, ny )
	enddo

	return
	end
C
C-----------------------------------------------------------------------
C
	subroutine mg3d_lihzhi( nxf, nyf, p, dx, dy, dz, w, sbar,
     +    alpha, beta, gamma )
C
C  Implements equations 15-17 of Lih-Zhiming's Finite difference
C  compensation algorithm, Geophysics v. 56-10, 1991, p. 1653:
C
C  p(kx,ky,w,z+dz) = p(kx,ky,w,z) * exp(i*(kz-kbar)*dz) * 
C        (a+ib)/(a-ib) * (c+id)/(c-id)
C  where
C
C  nxf,nyf = transform lengths in x and y
C  p(nxf,nyf) = input/output wavefield
C  dx,dy = grid delta values
C  dz = depth step size
C  w = angular frequency
C  sbar = slowness (full value, half-velocity taken into account here)
C  alpha, beta, gamma = finite difference coefficients
C
C  kx,ky = horizontal wavenumbers, i.e. index * 2*pi/(nxf*dx)
C  kbar = w*2.*sbar
C  kz = sqrt( kbar^2 - kx^2 - ky^2 )
C  kxhat = ( 2 - 2*cos(kx*dx) )/(dx^2)
C  a = ( 1 - (gamma*dx^2 + beta/kbar^2)*kxhat^2 )/dz
C  b = (alpha*kxhat^2)/(2*kbar)
C  c,d = same as a,b but for y
C
	implicit none
C  Arguments
	integer nxf,nyf
	real dx,dy,dz,w,sbar,alpha,beta,gamma
	complex p(0:nxf-1,0:nyf-1)
C  Local variables
	integer i,j,nx2,ny2
	real dkx,dky,kxhat2,kyhat2,kbar,kbar2,lbar,dzi,dxi2,dyi2
	real a,b,c,d,e,afact,bfact,cfact,dfact
	real kx,kx2,ky,ky2,kz
	real TWOPI
	parameter (TWOPI = 2*3.141592654)
	complex ctemp

C  Constants
	nx2 = nxf/2
	ny2 = nyf/2
	dkx = TWOPI/(dx*nxf)
	dky = TWOPI/(dy*nyf)
	kbar = 2.0*w*sbar
	kbar2 = kbar*kbar
	lbar = 1./kbar
	dzi = 1./dz
	dxi2 = 0.5/(dx*dx)
	dyi2 = 0.5/(dy*dy)
	afact = gamma*dx*dx + beta*lbar*lbar
	cfact = gamma*dy*dy + beta*lbar*lbar
	bfact = 0.5*lbar*alpha
	dfact = bfact

C  Loop and apply correction term
	do j=0,nyf-1
C  Set signed wavenumber
		if (j.le.ny2) then
			ky = dky*j
		else
			ky = dky*(j-nyf)
		endif
C  Factors for y direction
		ky2 = ky*ky
		kyhat2 = dyi2*(1.-cos(ky*dy))
		c = dzi*(1.-cfact*kyhat2)
		d = dfact*kyhat2
		do i=0,nxf-1
C  Set signed wavenumber
			if (i.le.nx2) then
				kx = dkx*i
			else
				kx = dkx*(i-nxf)
			endif
C  Factors for x direction
			kx2 = kx*kx
			kxhat2 = dxi2*(1.-cos(kx*dx))
			a = dzi*(1.-afact*kxhat2)
			b = bfact*kxhat2	
C  Apply correction term
			if ( kbar2 .ge. kx2+ky2 ) then
				kz = sqrt( kbar2 - kx2 - ky2 )
				e = dz*(kz-kbar)
				ctemp = cmplx(cos(e),sin(e))
     +				* (cmplx(a,b)/cmplx(a,-b))
     +				* (cmplx(c,d)/cmplx(c,-d))
				p(i,j) = p(i,j) * ctemp
			else
				p(i,j) = (0.,0.)
			endif
		enddo
	enddo

	return
	end
C
C-----------------------------------------------------------------------
C  

	subroutine mg3d_vmig(nvx, vx1, dvx, nvy, vy1, dvy, z, 
     +    vel, vbar, sa, ier ) 
C
C  Set migration velocity.  Normally, this comes from a seismic
C  data base system.  For benchmark purposes, use constant velocity.
C
C  nvx,vx1,dvx - x locations
C  nvy,vy1,dvy - y locations
C  z - depth where velocity is requested
C  vel(nvx,nvy) - output migration velocity
C  vbar - output average velocity
C  sa - work array
C
	implicit none
C  Arguments
	integer nvx,nvy,ier
	real vx1, dvx, vy1, dvy, z, vel(nvx,nvy), vbar, sa(*)
	include 'mg3d.inc'
C  Local variables
	integer i,j
	real vtemp

C Set constant velocity if specified in parameters
	if (vmig.ne.0) then
		do j=1,nvy
			do i=1,nvx
				vel(i,j) = vmig
			enddo
		enddo
	else
C  Interpolate velocity from VSBF
		call vsbf_lint( nvx, vx1, dvx, nvy, vy1, dvy, z, 
     +    nvx, vel, sa, ier )
	endif

C  Compute average velocity
	vbar = 0.0
	do j=1,nvy
		vtemp = 0.0
		do i=1,nvx
			vtemp = vtemp + vel(i,j)
		enddo
		vbar = vbar + vtemp / float(nvx)
	enddo
	
	vbar = vbar / float(nvy)

	return
	end
C
C-----------------------------------------------------------------------
C
	subroutine mg3d_add( nx, node, nodes, nyp, nypn, sa, ldim, otr )
C
C  Sum partial images across nodes
C  nx - number of x values
C  node - local node number
C  nodes - number of nodes in processor complex
C  nyp - number of y values spread across nodes on output
C  nypn - number of y values per node on output
C  sa - input partial image array dimensioned (nx,nyp,nodes).
C       Will be transposed across nodes before the array is used,
C       thus the dimension statement below.
C  ldim - leading dimension of output summed array
C  otr - output completed sum, (nx,nypn) values on each processor,
C       y dimension spread across nodes
C
	implicit none
	integer nx, node, nodes, nyp, nypn, ldim
	real sa(nx,nodes,*), otr(ldim,*)
	integer i,j,k

C  Handle single node case
	if (nodes.eq.1) then
		do j=1,nyp
			do i=1,nx
				otr(i,j) = sa(i,1,j)
			enddo
		enddo
		return
	endif

C  On input there is a partial image on each node:
C     (nx,nyp,nodes)
C  Use distributed transpose operation to convert to:
C     (nx,nodes,nyp)
C  After which we may sum over the middle dimension locally.
	
C  Transpose across processors
	call dtran132( nx, nyp, nodes, sa, otr )

C  Sum partial images
	call vfill( 0.0, otr, 1, ldim*nypn )
	do k=1,nypn
		do i=1,nx
			do j=1,nodes
				otr(i,k) = otr(i,k) + sa(i,j,k)
			enddo
		enddo
	enddo
	
	return
	end
C
C-----------------------------------------------------------------------
C
      subroutine mg3d_mrcf( LDA, M, A, LDB, B, N2, NS, NF, IF1, SA )
C
C  Perform multiple real - to - complex forward fourier transforms
C
C  lda - leading dimension of input data to be transformed
C  m - second dimension of input/output arrays (number of transforms)
C  a(lda,m) - input real array
C  ldb - leading dimension of output complex transform array.
C  b(ldb,m) - output complex array
C  n2 - transform length (must be .ge. ns)
C  ns - number of input samples. a(ns+1:lda,1:m) is ignored.
C  nf - number of output frequencies. b(nf+1:ldb,1:m) is undefined.
C  if1 - starting output frequency
C  sa - work array, length n2 + 2
C
	real a(lda,m), sa(*)
	complex b(ldb,m)
C
C  Loop over input traces
	nz = n2-ns
	do j=1,m
C  Copy to work array
		call scopy( ns, a(1,j), 1, sa, 1 )
C  Zero fill
		if (nz.gt.0) call vfill( 0.0, sa(ns+1), 1, nz )
C  Forward transform
		call seircft( n2, sa )
C  Copy to output array
		call ccopy( nf, sa(if1), 1, b(1,j), 1 )
	enddo
C
	return
	end
C
C*----------------------------------------------------------------------
C
	subroutine mg3d_amax( nx, ny, nf, ra, amax )
	complex ra(nx,ny,nf)
	amax = 0.0
	do k=1,nf
	do j=1,ny
	do i=1,nx
		amax = max( amax, cabs(ra(i,j,k)) )
	enddo
	enddo
	enddo
	return
	end
C
C-----------------------------------------------------------------------
C
	subroutine mg3d_attr()
C
C  Set datacube attributes
C
	implicit none
	include 'coord.inc'
	include 'group.inc'
	include 'para.inc'
	include 'mg3d.inc'

C  Data attributes for depth slices
	ns = nx
	fsr = dx
	msr = nint(dx)
	ntpg = ny
	ngpl = nz
	nlin = 1
	gtype = 'zslc    '
	lgroup = igroup
	lline = iline
	ictype = 0

C  Decomposition has y spread across processors
	pmode = IPANEL
	ntpn = nypn
	nlive = nylive

	return
	end

C
C-----------------------------------------------------------------------
