C*----------------------------------------------------------------------
C*
C*  NAME:
C*    nmocp, nmoca, nmocb, nmocc - Seismic Processing Benchmark 
C*        process NMOC
C*     
C*  SYNOPSIS:
C*    integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
C*    real otr(ldim,maxtrc), ra(nra), sa(nra)
C*    logical abort
C*    call nmocp( ldim, maxtrc, nra, nsa, abort, ipr )
C*    call nmoca( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                abort, ipr )
C*    call nmocb( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                ntri, ntro, abort, ipr )
C*    call nmocc( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                ntro, abort, ipr )
C*
C*  DESCRIPTION: 
C*    Apply normal moveout correction
C*
C*  PARAMETERS: 
C*    Record 1
C*    Field  Columns   Description
C*      1     11-20    Velocity for normal moveout application
C*      2     21-30    Max stretch in percent of input sampling
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*     Include files:
C*       ./coord.inc
C*       ./nmoc.inc
C*
C*     Libraries:
C*       libbench.a
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        12/31/91 - C. Mosher, ARCO E&PT 
C*
C*      References:
C*
C*      Modified:
C*
C*----------------------------------------------------------------------

C*----------------------------------------------------------------------
C*   NAME:
C*     nmocp
C*     
C*----------------------------------------------------------------------
C*   DESCRIPTION:
C*
C*   This process reads 'NMOC' parameters, checks syntax, and sets
C*   memory requirements for the processing phase
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*      Subroutines:
C*        getparm - read parameter record
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/06/90 - C. Mosher
C*
C*      References:
C*        SPARC System Documentation
C*
C*      Modified:
C*
C*----------------------------------------------------------------------
C
	subroutine nmocp( ldim, maxtrc, nra, nsa, abort, ipr )
C
C  Arguments:
C
C  Output leading dimension of trace array required for the PROC step
	integer LDIM
C  Output max number of traces per group to be output during PROC step
	integer MAXTRC
C  Number of words of reserved storage required in PROC step
	integer NRA
C  Number of words of working storage required in PROC step
	integer NSA
C  Abort flag, set to .TRUE. by this process if errors detected
	logical ABORT
C  Print unit number that this routine should use
	integer IPR

C
C  Common blocks:
C
C  parameters for process NMOC
	include 'nmoc.inc'
C  data structure definition
	include 'group.inc'

C
C  Local declarations:
C
C  parameter record buffer
	character*80 buf
C  parameter record index
	integer index

C
C  Start of executable statements
C

C  Get parameter record with velocity
	index = 1
	call getparm('NMOC',index,buf,ier)
	if (ier.ne.0) then
		write (ipr,*) 'No parameter records for process NMOC'
		abort = .true.
		return
	endif

C  Read velocity value and max stretch in percent
	read (buf,'(10x,2f10.0)') vnmo,smax
	if (vnmo.eq.0) then
		write (ipr,*) 'NMOC: velocity must be non-zero'
		abort = .true.
		return
	endif

C  Multiply velocity by sample rate in seconds
	vel = vnmo * 0.001 * float(msr)

C  Set sloth value 
	sloth = 1./(vel*vel)

C  Min sample difference
	if (smax.eq.0.) smax = 50
	smin = 1./(1. + 0.01 * smax)

	write (ipr,9010) vnmo,smax
 9010 format (' Process NMOC    : Normal Moveout Correction',/,
     +        ' Moveout velocity: ',f8.1,/,
     +        ' Max stretch (%) : ',f8.1)

C  Set memory requirements
	ldim = ns
	maxtrc = ntpg
	nra = 0
C  Need work space to hold 1 trace
	nsa = ns

	return
	end

C*----------------------------------------------------------------------
C*   NAME:
C*     nmoca, nmocb, nmocc
C*     
C*----------------------------------------------------------------------
C*   DESCRIPTION:
C*
C*   Processing phase routines for Seismic Benchmark process NMOC
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/01/92 - C. Mosher
C*
C*      References:
C*        Seismic Benchmark Users Guide
C*
C*      Modified:
C*
C*----------------------------------------------------------------------

C*----------------------------------------------------------------------
C*  The 'a' routine does nothing
C*----------------------------------------------------------------------

	subroutine nmoca( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		abort, ipr )
	integer ldim, maxtrc, nra, nsa, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort

	return
	end

C*----------------------------------------------------------------------
C*  The 'b' routine applies nmo
C*----------------------------------------------------------------------

	subroutine nmocb( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntri, ntro, abort, ipr )
	integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort

	include 'nmoc.inc'
	include 'group.inc'
	include 'coord.inc'
C
C  Number of samples - 1
	ns1 = ns-1

C  Loop over number of traces
	do j = 1, ntri
C  Compute trace offset squared
		x2 = (sxyz(1,j)-rxyz(1,j))**2 + (sxyz(2,j)-rxyz(2,j))**2
C  Stretch factor
		fact = x2*sloth
C  Copy input trace to work array
		do i=1,ns
			sa(i) = otr(i,j)
			otr(i,j) = 0.0
		enddo
C  Intial sample
		s1 = sqrt( 1 + x2*sloth )
C  Loop over number of output samples
		do i=1,ns
C  Compute floating point sample
			xi = float(i-1)
			xs = sqrt( xi*xi + x2*sloth )	
C  Stretch value
			ds = xs - s1
			s1 = xs
C  Input sample number
			k = ifix(xs)
C  Set output value if minimum sample difference is exceeded
			if (k.lt.ns1 .and. ds.ge.smin) then
				dk = xs - k
				otr(i,j) = dk * sa(k+2) + (1.-dk) * sa(k+1)
			endif
		enddo
	enddo

	ntro = ntri
	return
	end

C*----------------------------------------------------------------------
C*  'c' routine does nothing
C*----------------------------------------------------------------------
c
	subroutine nmocc( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntro, abort, ipr )
	integer ldim, maxtrc, nra, nsa, ntro, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort

	call sysmsg('Process NMOC completed')

	return
	end

C*----------------------------------------------------------------------
