C*----------------------------------------------------------------------
C*
C*  NAME:
C*    stakp, staka, stakb, stakc - Seismic Processing Benchmark 
C*        process STAK
C*     
C*  SYNOPSIS:
C*    integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
C*    real otr(ldim,maxtrc), ra(nra), sa(nra)
C*    logical abort
C*    call stakp( ldim, maxtrc, nra, nsa, abort, ipr )
C*    call staka( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                abort, ipr )
C*    call stakb( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                ntri, ntro, abort, ipr )
C*    call stakc( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                ntro, abort, ipr )
C*
C*  DESCRIPTION: 
C*    Sum seismic traces into output midpoint locations
C*
C*  PARAMETERS: 
C*    Record 1
C*    Field  Columns   Description
C*      1      7       Debug flag
C*      2     11-20    X value of starting midpoint
C*      3     21-30    Number of midpoints to include in stack
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*     Include files:
C*       ./coord.inc
C*       ./group.inc
C*       ./para.inc
C*       ./stak.inc
C*
C*     Libraries:
C*       libbench.a
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/11/92 - C. Mosher, ARCO E&PT 
C*
C*      References:
C*
C*      Modified:
C*
C*----------------------------------------------------------------------

C*----------------------------------------------------------------------
C*   NAME:
C*     stakp
C*     
C*----------------------------------------------------------------------
C*   DESCRIPTION:
C*
C*   This process reads 'STAK' parameters, checks syntax, and sets
C*   memory requirements for the processing phase
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*      Subroutines:
C*        getparm - read parameter record
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/06/90 - C. Mosher
C*
C*      References:
C*        SPARC System Documentation
C*
C*      Modified:
C*
C*----------------------------------------------------------------------
C
	subroutine stakp( ldim, maxtrc, nra, nsa, abort, ipr )
	implicit none
C
C  Arguments:
C
C  Output leading dimension of trace array required for the PROC step
	integer LDIM
C  Output max number of traces per group to be output during PROC step
	integer MAXTRC
C  Number of words of reserved storage required in PROC step
	integer NRA
C  Number of words of working storage required in PROC step
	integer NSA
C  Abort flag, set to .TRUE. by this process if errors detected
	logical ABORT
C  Print unit number that this routine should use
	integer IPR

C
C  Common blocks:
C
	include 'coord.inc'
	include 'group.inc'
	include 'para.inc'
	include 'stak.inc'
	
C
C  Local declarations:
C
	character*80 buf
	integer index,ier

C
C  Start of executable statements
C

C  Get parameter record with velocity
	index = 1
	call getparm('STAK',index,buf,ier)
	if (ier.ne.0) then
		write (ipr,*) 'No parameter records for process STAK'
		abort = .true.
		return
	endif

C  Read start midpoint and number of midpoints
	read (buf,'(6x,i1,3x,f10.0,i10)') idbug, xmid0, nmid
	if (nmid.eq.0) then
		write (ipr,*) 'STAK: Number of midpoints must be non-zero'
		abort = .true.
		return
	endif

	write (ipr,9010) xmid0, nmid
 9010 format (' Process STAK    : Stack traces',/,
     +        ' Start midpoint X: ',f8.1,/,
     +        ' No. midpoints   : ',i8)
	if (idbug.ne.0) write (ipr,*) 'Debug print is on'

C  Compute traces per node for decomposition
	call jdcomp( nmid, jntpn, jnlive )
	jnmid = jntpn*nodes
	jnlin = nlin

C  Set memory requirements
	ldim = ns
	maxtrc = nmid
C  Reserved array to hold stacked section
	nra = ns*jnmid
C  Work space for transpose and substacks
	if (nodes.gt.1) then
		nsa = ns*jntpn
	else
		nsa = 0
	endif
C  If more than 1 line, need work space to hold input traces
	if (nlin.gt.1) then
		ksa = nsa + 1
		nsa = nsa + ns*ntpg
	else
		ksa = 1
	endif

C  Set attributes 
	call stak_attr()

	return
	end

C*----------------------------------------------------------------------
C*   NAME:
C*     staka, stakb, stakc
C*     
C*----------------------------------------------------------------------
C*   DESCRIPTION:
C*
C*   Processing phase routines for Seismic Benchmark process STAK
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/01/92 - C. Mosher
C*
C*      References:
C*        Seismic Benchmark Users Guide
C*
C*      Modified:
C*
C*----------------------------------------------------------------------

C*----------------------------------------------------------------------
C*  The 'a' routine initializes stack grid
C*----------------------------------------------------------------------

	subroutine staka( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		abort, ipr )
	implicit none
	integer ldim, maxtrc, nra, nsa, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort

	include 'coord.inc'
	include 'group.inc'
	include 'para.inc'
	include 'stak.inc'

C  Debug flag
	if (idbug.ne.0 .and. node.eq.master) then
		idbug = ipr
	else
		idbug = 0
	endif

C  Compute traces per node for decomposition
	call jdcomp( nmid, jntpn, jnlive )

C  Initialize stack grid to zero
	call vfill( 0.0, ra, 1, ns*jnmid )

C  Initialize line counter
	jline = 0

C  Set stacked flag
	stacked = .false.

C  Set output attributes
	call stak_attr()

	return
	end

C*----------------------------------------------------------------------
C*  The 'b' routine sums traces into output location
C*----------------------------------------------------------------------

	subroutine stakb( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntri, ntro, abort, ipr )
	implicit none
C  Arguments
	integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort
C  Common blocks
	include 'coord.inc'
	include 'group.inc'
	include 'stak.inc'

C  Check for new line
	if (jline.ne.iline .and. iline.gt.1) then
C  Save input traces for new line
		call stak_save( ldim, ntri, otr, ns, sa(ksa) )
C  Set attributes for output stacked line
		call stak_attr()
		jline = iline
		iline = 1
		igroup = jline-1
		ntro = nmid
		if (idbug.ne.0) write (ipr,9010) igroup
 9010   format (' STAKB: Output group ',i5)
C  Stack current line across nodes
		call stak_line( nra, ra, nsa, sa, ldim, otr )
C  Re-initialize stack grid
		call vfill( 0.0, ra, 1, ns*jnmid )
C  Start accumulation of new line using saved traces
		call stak_add( ns, ntri, sa(ksa), ns, ra )
C  Set output attributes
		call stak_attr()
	else
C  Accumulate current set of traces
		call stak_add( ldim, ntri, otr, ns, ra )
C  Don't send any traces out
		ntro = 0
	endif

	return
	end

C*----------------------------------------------------------------------
C*  'c' routine outputs last stacked section
C*----------------------------------------------------------------------
c
	subroutine stakc( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntro, abort, ipr )
	implicit none
	integer ldim, maxtrc, nra, nsa, ntro, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort

	include 'coord.inc'
	include 'group.inc'
	include 'para.inc'
	include 'stak.inc'

C  Check for all done
	if (stacked) then
		ntro = 0
		call sysmsg('Process STAK completed')
	else
C  Set output attaributes
		call stak_attr()
		ntro = nmid
		if (jline.eq.0) then
			igroup = 1
		else
			igroup = jline
		endif
		iline = 1
		stacked = .true.
C  Output last line
		call stak_line( nra, ra, nsa, sa, ldim, otr )
		if (idbug.ne.0) write (ipr,9010) iline
 9010   format (' STAKC: Output line ',i5)
	endif

	return
	end

C*----------------------------------------------------------------------
C*  Process STAK support routines
C*----------------------------------------------------------------------

	subroutine stak_line( nra, ra, nsa, sa, ldim, otr )
C
C  Sum partial stacks on each processor across nodes
C  nra - length of reserved aray
C  ra - sub-stack on each processor
C  nsa - length of work array
C  sa - work array 
C  otr(ldim,*) - output array containing completed sm
C
	implicit none
C  Arguments
	integer nra, nsa, ldim
	real ra(nra), sa(nsa), otr(ldim,*)
C  Common blocks
	include 'coord.inc'
	include 'group.inc'
	include 'para.inc'
	include 'stak.inc'
C  Local declarations
	integer lblk,j

C  There is now a 2D partial stack array on each node.
C  Treat as a 3D array distributed over the third dimension:
C        (ns  x  nmid/p)  x  p  |  p
C         t      x1         x2     h
C  where p is the number of processors. Use low-level transpose
C  operation to swap last two dimensions:
C        (ns  x  nmid/p)  x  p  |  p
C         t      x1          h    x2
C  After which we may sum over h, leaving a 2D array
C  distributed over the second dimension:
C        ns  x  nmid/p  |  p
C         t      x1       x2

C  Set sub-block length of partial stack array
	lblk = ns * jntpn
C  Call transpose routine
	call jtran23( lblk, ra, sa )
C  Sum over offset
	call stak_sum( ns, jntpn, nodes, ra, ldim, otr )

C  Set coordinates for this block
	do j=1,jnlive
		sxyz(1,j) = xmid0 + gdx*(jntpn*node + j - 1)
		sxyz(2,j) = gdy*(iline-1)
		sxyz(3,j) = 0.0
		rxyz(1,j) = sxyz(1,j)
		rxyz(2,j) = sxyz(2,j)
		rxyz(3,j) = sxyz(3,j)
	enddo

	return
	end

C*----------------------------------------------------------------------

	subroutine stak_save( ldim, ntri, otr, ns, sa )
C
C  Save input traces in temp array
C  ldim - leading dimension of trace array
C  ntri - number of input traces
C  otr - input trace array
C  ns - samples per trace
C  sa - temp array
C
	implicit none
C  Arguments
	integer ldim, ntri, ns
	real otr(ldim,ntri), sa(ns,*)
C  Local variables
	integer i,j

	do j=1,ntri
		do i=1,ns
			sa(i,j) = otr(i,j)
		enddo
	enddo
	
	return
	end

C*----------------------------------------------------------------------

	subroutine stak_add( ldim, ntri, otr, ns, ra )
C
C  Sum trace into stack grid
C  ldim - leading dimension of trace array
C  ntri - number of input traces
C  otr - input trace array
C  ns - samples per trace
C  ra - stack accumulator
C
	implicit none
C  Arguments
	integer ldim, ntri, ns
	real otr(ldim,ntri), ra(ns,*)
C  Common blocks
	include 'coord.inc'
	include 'stak.inc'
C  Local variables
	real xm
	integer i,j,jm

C  Loop over number of traces
	do j = 1, ntri
C  Compute trace midpoint
		xm = 0.5 * (sxyz(1,j) + rxyz(1,j))
C  Midpoint index
		jm = 1 + nint( (xm - xmid0)/gdx )
C  Accumulate if in range
		if (jm.ge.1 .and. jm.le.nmid) then
			do i=1,ns
				ra(i,jm) = ra(i,jm) + otr(i,j)
			enddo
		endif
	enddo
	
	return
	end

C*----------------------------------------------------------------------

	subroutine stak_sum( ns, ntpn, nodes, ra, ldim, otr )
C
C  Sum sub-block of traces over number of nodes
C  ns - samples per trace
C  ntpn - traces per node
C  nodes - number of nodes
C  ra - stack accumulator
C  ldim - leading dimension of trace array
C  otr - output trace array
C
	implicit none
C  Arguments
	integer ns, ntpn, nodes, ldim
	real otr(ldim,ntpn), ra(ns,ntpn,nodes)
C  Local variables
	integer i,j,k

C  Initialize output array
	do j=1,ntpn
		do i=1,ns
			otr(i,j) = ra(i,j,1)
		enddo
	enddo

C  Sum remaining values across nodes
	do k=2,nodes
		do j=1,ntpn
			do i=1,ns
				otr(i,j) = otr(i,j) + ra(i,j,k)
			enddo
		enddo
	enddo
	
	return
	end

C*--------------------------------------------------------------------

	subroutine stak_attr()
C
C  Set output stacked data attributes
C
	include 'group.inc'
	include 'para.inc'
	include 'stak.inc'

C  Group attributes
	ntpg = nmid
	ngpl = jnlin
	nlin = 1
	gtype = 'xoff    '

C  Panel attributes
	pmode = 1
	ntpn = jntpn
	nlive = jnlive

	return
	end
