C*----------------------------------------------------------------------
C*
C*  NAME:
C*    tr12p, tr12a, tr12b, tr12c - Seismic Processing Benchmark 
C*        process TR12
C*     
C*  SYNOPSIS:
C*    integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
C*    real otr(ldim,maxtrc), ra(nra), sa(nra)
C*    logical abort
C*    call tr12p( ldim, maxtrc, nra, nsa, abort, ipr )
C*    call tr12a( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                abort, ipr )
C*    call tr12b( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                ntri, ntro, abort, ipr )
C*    call tr12c( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                ntro, abort, ipr )
C*
C*  DESCRIPTION: 
C*    Transpose trace array in-line.
C*    Note that coordinates are lost in this operation.
C*
C*  PARAMETERS: 
C*    None required
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*     Include files:
C*       ./coord.inc
C*       ./group.inc
C*       ./tr12.inc
C*
C*     Libraries:
C*       libbench.a
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        12/31/91 - C. Mosher, ARCO E&PT 
C*
C*      References:
C*
C*      Modified:
C*
C*----------------------------------------------------------------------

C*========================= START OF SUBROUTINE ========================
C*   NAME:
C*     tr12p
C*     
C*----------------------------------------------------------------------
C*   DESCRIPTION:
C*
C*   This process reads 'TR12' cards, checks syntax, and sets
C*   memory requirements for the processing phase
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*      Subroutines:
C*        sysparm - get parameter value
C*        jdcomp - set parallel data decomposition
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/06/90 - C. Mosher
C*
C*      References:
C*        SPARC System Documentation
C*
C*      Modified:
C*
C*----------------------------------------------------------------------
C
	subroutine tr12p( ldim, maxtrc, nra, nsa, abort, ipr )
	implicit none
C
C*----------------------------------------------------------------------
C*  ARGUMENTS:
C*----------------------------------------------------------------------
C
C  Output leading dimension of trace array required for the PROC step
	integer LDIM
C  Output max number of traces per group to be output during PROC step
	integer MAXTRC
C  Number of words of reserved storage required in PROC step
	integer NRA
C  Number of words of working storage required in PROC step
	integer NSA
C  Abort flag, set to .TRUE. by this process if errors detected
	logical ABORT
C  Print unit number that this routine should use
	integer IPR

C*----------------------------------------------------------------------
C*  Common blocks     
C*----------------------------------------------------------------------

	include 'coord.inc'
	include 'group.inc'
	include 'para.inc'
	include 'tr12.inc'

C*----------------------------------------------------------------------
C*  Local declarations
C*----------------------------------------------------------------------

C*----------------------------------------------------------------------
C*  Start of executable statements
C*----------------------------------------------------------------------

C  Set size of output data
	jns = ntpg
	jntpg = ns
	jngpl = ngpl
	jnlin = nlin

	write (ipr,9010) ns,ntpg,ngpl,nlin,
     +    jns, jntpg, jngpl, jnlin
 9010 format (' Process TR12    : Transpose Seismic Data Set',/,
     +        ' Input  Data Size: ',4i6,/,
     +        ' Output Data Size: ',4i6)

C  Set memory requirements
	nra = 0

C  Need work space for distributed transpose if in panel mode
	if (pmode.ne.0) then
C  Set values for output decomposition 
		call jdcomp( ns, jntpn, jnlive )
C  For transpose, we need lengths that are a multiple of nodes
		n1 = jntpn * nodes
		n2 = ntpn * nodes
C  Set work space for transpose
		nsa = n1*n2/nodes
		ksa = nsa + 1
		nsa = nsa + ntpn*jntpn
	else
C  Set default values to no decomposition
		jntpn = jntpg
		jnlive = jntpg
		nsa = jns*jntpg
	endif

C  Set output attributes and trace array size
	call tr12atr()
	ldim = jns
	maxtrc = jntpn

	return
	end

C*========================= START OF SUBROUTINE ========================
C*   NAME:
C*     tr12a, tr12b, tr12c
C*     
C*----------------------------------------------------------------------
C*   DESCRIPTION:
C*
C*   Processing phase routines for tr12sposing seismic bencmark
C*     data stream
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*      Subroutines:
C*        tr1221 - 2D in-place transpose
C*        dtr1221 - 2D distributed transpose
C*        sbfcopy - 2D array copy with reshape
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/01/92 - C. Mosher
C*
C*      References:
C*        Seismic Benchmark Users Guide
C*
C*      Modified:
C*
C*----------------------------------------------------------------------
C
	subroutine tr12a( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		abort, ipr )
	implicit none
C  Arguments
	integer ldim, maxtrc, nra, nsa, ipr
	logical abort
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
C  Include blocks
	include 'coord.inc'
	include 'group.inc'
	include 'para.inc'
	include 'tr12.inc'

C  Set output decomposition if in panel mode
	if (pmode.ne.0) call jdcomp( jntpg, jntpn, jnlive )

C  Set output attributes
	call tr12atr()

	return
	end

C*----------------------------------------------------------------------
C*  The 'b' routine transposes data stream
C*----------------------------------------------------------------------

	subroutine tr12b( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntri, ntro, abort, ipr )
	implicit none
C  Arguments
	integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort
C  Common blocks
	include 'coord.inc'
	include 'group.inc'
	include 'para.inc'
	include 'tr12.inc'
C  Local variables

C  Distributed transpose if in panel mode
	if (pmode.eq.1) then
C  Copy input data to work array
		call sbfcopy( n1, ntpn, sa, ldim, nlive, otr )
C  Transpose across processors
		call dtran21( n1, n2, sa, sa(ksa) )
C  Copy back to output array
		call sbfcopy( ldim, jntpn, otr, n2, jnlive, sa )
	else 
C  Transpose input array
		call sbfcopy( ns, ntpg, sa, ldim, ntri, otr )
		call tran21( ns, ntpg, sa )
		call sbfcopy( ldim, ns, otr, ntpg, ns, sa ) 
	endif

C  Set output attributes
	ntro = jnlive
	call tr12atr()

C  Set dummy coordinates
	call tr12_coord( ntro, sxyz, rxyz )

	return
	end

C*----------------------------------------------------------------------
C*  Close files and quit in 'c' routine
C*----------------------------------------------------------------------
c
	subroutine tr12c( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntro, abort, ipr )
	implicit none
	integer ldim, maxtrc, nra, nsa, ntro, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort

	include 'group.inc'
	include 'para.inc'
	include 'tr12.inc'

	ntro = 0
	call sysmsg("Process TR12 Completed")

	return
	end

C*---------------------------------------------------------------------

	subroutine tr12_coord( ntro, sxyz, rxyz )
C 
C  Set dummy coordinates for output data
C
	integer ntro, j
	real sxyz(3,ntro), rxyz(3,ntro)

	do j=1,ntro
		sxyz(1,j) = float(j)
		sxyz(2,j) = 0.0
		sxyz(3,j) = 0.0	
		rxyz(1,j) = float(j)
		rxyz(2,j) = 0.0
		rxyz(3,j) = 0.0	
	enddo

	return
	end

C*---------------------------------------------------------------------

	subroutine tr12atr()
C
C  Set process TR12 attributes
C
	include 'group.inc'
	include 'para.inc'
	include 'tr12.inc'

C  Data attributes
	ns = jns
	ntpg = jntpg
	ngpl = jngpl
	nlin = jnlin

C  Parallel decomposition attributes
	if (pmode.ne.0) then
		ntpn = jntpn
		nlive = jnlive
	endif

	return
	end
