/*--------------------------------------------------------------------*/
/*  NAME:
/*    message.c - message passing routines for ARCO/Perfect benchmark
/*
/*  VERSION:
/*    Default version for PVM3 with NOP's defined for single processor
/*    Shared memory support for 'jtran23' on SYSV-SMP platforms
/*
/*  Routines defined in this module:
/*    jpinit - parallel initialization
/*    jmaster - return master node number
/*    jnodes - return number of nodes in system
/*    jdim - return hypercube dimension
/*    jnode - return node number
/*    jsync - global sync
/*    jkill - kill all processes
/*    jsend - send message (synchronous)
/*    jrecv - receive message (synchronous)
/*    jasend - send message (async)
/*    jarecv - receive message (async)
/*    jamsgwt - async message wait
/*    jbrdcst - broadcast message 
/*    jtrace - toggle tracing
/*    jdcomp - set simple decomposition
/*    jrange - return index range for simple decomposition
/*    jtran23 - tiled transpose for simple decomposition
/*     
/*--------------------------------------------------------------------*/
/*
/*  HISTORY:
/*
/*     Written:
/*       01/19/92 - C. Mosher, ARCO E&PT 
/*
/*     References:
/*       Intel iPSC/860 System Documentation
/*
/*     Modified:
/*
/*--------------------------------------------------------------------*/

/*--------------------------------------------------------------------*/
/*  Include files
/*--------------------------------------------------------------------*/

#include <stdio.h>

#ifdef PVM
#include "pvm3.h"
#endif /* PVM */

#ifdef SHMEM
/* Structures for controlling semaphores and shared memory */
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>

/* Base key values for semaphore and memory structures */
#define SEM_BASE 1
#define SHM_BASE 1
#define SHM_MAXSEG 62499840

/* Semphore and memory structures */
static key_t sem_key = (key_t)SEM_BASE;
static int sem_id = -1;
static int sem_sync = -1;
static key_t shm_key = (key_t)SHM_BASE;
static int *shm_id = (int *)NULL;
static char **shm_addr = (char **)NULL;
static int shm_len = 0;
static int syncval = 1;
static union {
	int val;
	struct semid_ds *buf;
	ushort *array;
} sem_arg,sem_val;

extern int errno;

#endif /* SHMEM */

/*--------------------------------------------------------------------*/
/*  Global variables
/*--------------------------------------------------------------------*/

/* Trace flag */
static int trace = 0;

/* Node numbers, master node, task id array */
static int node, nodes, master, *tids;

/* Base message number for jmsgrng */
static int basemsg = 16384;

/* Data decomposition parameters */
static int nglob=0, ntpn=0, nlive=0;

#ifdef PVM
/* local copy of instance name */
static char pvmname[256];
/* internal counter for creating unique PVM barrier names */
static int barnum = 0;
#endif /* PVM */

#ifdef cray
#define ISCRAY
#endif

#ifdef ISCRAY
#define jpinit_ JPINIT
#define jpterm_ JPTERM
#define jmaster_ JMASTER
#define jnodes_ JNODES
#define jnode_ JNODE
#define jkill_ JKILL
#define jsync_ JSYNC
#define jsend_ JSEND
#define jrecv_ JRECV
#define jasend_ JASEND
#define jarecv_ JARECV
#define jamsgwt_ JAMSGWT
#define jbrdcst_ JBRDCST
#define jtrace_ JTRACE
#define jdcomp_ JDCOMP
#define jrange_ JRANGE
#define jtran23_ JTRAN23
#endif

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jpinit - Initialize parallel processing
/*     
/*  SYNOPSIS:
/*
/*    char *path;
/*    int iret;
/*    iret = jpinit( path );
/*
/*  Initiates parallel processing environment.  For machines that
/*  require host or node initialization, the 'path' variable can
/*  be used to specify the executable to be loaded.
/*
/*  PVM Notes:
/*  Assumes standard PVM installation, and that 'pvmd' has been
/*  started.  Use the 'pvm_config' routine to find out how many
/*  machines are in the cluster.  The number of nodes started is
/*  set to the number of machines in the cluster.  If evironment
/*  variable 'PVM_NODES' is set, this number is used instead.
/*
/*--------------------------------------------------------------------*/

int jpinit( path )
char *path;
{
#ifdef PVM
	int iret,i,j,k,mytid,narch,nhost,ntask,parent,flag;
	char *p,*getenv();
	struct hostinfo *info;
#endif /* PVM */
#ifdef SHMEM
	int len,nbits;
#endif /* SHMEM */

	nodes = 1;
	node = 0;
	master = 0;

#ifdef PVM
/* Set executable name */
	strncpy( pvmname, path, 255 );

/* Enroll program in PVM */
	mytid = pvm_mytid();
	parent = pvm_parent();
/*
	iret = pvm_joingroup( pvmname );
*/

/* Set spawn flag */
	flag = PvmTaskHost;
	if ( (p=getenv("PVM_DEBUG")) != (char *)NULL)
		flag += PvmTaskDebug;

/* Parent (master) task spawns additional tasks */
	if (parent < 0) {
		node = master;
/* Get node stats */
		iret = pvm_config( &nhost, &narch, &info );
		if (iret < 0) {
			fprintf(stderr,"SeisPerf/PVM: Initialization failed\n");
			pvm_perror("jpinit: pvm_config");
			return( -1 );
		}
/* Check for multiple tasks per node */
		ntask = 1;
		if ( (p=getenv("PVM_NTASK")) != (char *)NULL) {
			sscanf(p,"%d",&ntask);
			if (ntask < 1 || ntask > 1024) {
				fprintf(stderr,"SeisPerf/PVM: Initialization failed\n");
				fprintf(stderr,"jpinit: invalid PVM_NTASK value\n");
				return(-1);
			}
		}
		nodes = ntask*nhost;
/* Allocate space for task list */
		tids = (int *)malloc( nodes * sizeof(int) );
/* Set master node task id */
		tids[0] = mytid;
/* Spawn node processes */
		if (nodes > 1) {
			k = 1;
/* Spawn tasks on other hosts first */
			for (i=1; i<nhost; i++) {
				for (j=0; j<ntask; j++) {
					iret = pvm_spawn( pvmname, (char **)0, flag,
						info[i].hi_name, 1, &tids[k] );
					if (iret < 0 || tids[k] < 0) {
						fprintf(stderr,
							"SeisPerf/PVM: spawn failed for task %d host %s\n",
							k,info[i].hi_name);
						fflush(stderr);
						pvm_perror("jpinit");
						return( -1 );
					}
					k++;
				}
			}
/* Additional tasks on master node */
			for (j=1; j<ntask; j++) {
				iret = pvm_spawn( pvmname, (char **)0, flag,
					info[0].hi_name, 1, &tids[k] );
				if (iret < 0 || tids[k] < 0) {
					fprintf(stderr,
							"SeisPerf/PVM: spawn failed for task %d host %s\n",
							k,info[0].hi_name);
					fflush(stderr);
					pvm_perror("jpinit");
					return( -1 );
				}
				k++;
			}
/* Broadcast number of nodes and task id list for the cluster */
			pvm_initsend(PvmDataRaw);
			pvm_pkbyte( &nodes, sizeof(int), 1);
			pvm_mcast( &tids[1], nodes-1, 1 );
			pvm_initsend(PvmDataRaw);
			pvm_pkbyte(tids, nodes*sizeof(int), 1);
			pvm_mcast( &tids[1], nodes-1, 1 );
		}
		fprintf(stderr,"SeisPerf/PVM Initialized: %s for %d nodes\n",
			pvmname,nodes);
		fflush(stderr);
	}
	else {
/* Other nodes receive number of nodes and task list */
		jrecv( 1, &nodes, sizeof(int) );
		tids = (int *)malloc(nodes * sizeof(int));
		jrecv( 1, tids, nodes*sizeof(int) );
/* Determine node index number in task list */
		for (i=1; i<nodes; i++) {
			if ( mytid == tids[i] ) {
				node = i;
				break;
			}
		}
		fprintf(stderr,"SeisPerf/PVM slave: node = %d\n",node); 
		fflush(stderr);
	}

#ifdef SHMEM
	if (nodes <= 1) return(0);
/* Shared memory initialization */
	sem_key = getuid();
/* Find power of 2 greater than number of nodes */
	i = 2;
	nbits = 1;
	while (i < nodes) {
		i = 2*i;
		nbits++;
	}
	sem_key = getuid() << nbits;
	shm_key = sem_key;
	if (node == master) {
/* Create semaphore structures for use by 'jtran23' */
		sem_id = semget( sem_key, nodes, 0600 | IPC_CREAT );
		sem_sync = semget( sem_key+1, nodes, 0600 | IPC_CREAT );
		len = nodes*sizeof(ushort);
		sem_val.array = (ushort *)malloc( len );
		memset( (char *)sem_val.array, (char)0, len ); 
/* Initialize semaphore values */
		if (sem_id >= 0) {
			if (semctl( sem_id, 0, SETALL, sem_val ) != 0)
				sem_id = -1;
		}
		if (sem_sync >= 0) {
			if (semctl( sem_sync, 0, SETALL, sem_val ) != 0)
				sem_sync = -1;
		}
/* Sync to allow other nodes to access semaphores */
		msgsync();
	}
	else {
/* Other nodes wait for semaphores to be available */
		msgsync();
		sem_id = semget( sem_key, nodes, 0600 );
		sem_sync = semget( sem_key+1, nodes, 0600 );
		len = nodes*sizeof(ushort);
		sem_val.array = (ushort *)malloc( len );
	}
	if (sem_id < 0 || sem_sync < 0) {
		fprintf(stderr,"SeisPerf/SHMEM allocation failed node %d errno %d\n",node,errno);
		perror("perror");
		fflush(stderr);
		return(-1);
	}

/* Allocate space for shared memory segment id's and addresses */
	len = nodes*sizeof(int);
	shm_id = (int *)malloc( len );
	for (i=0; i<nodes; i++)
		shm_id[i] = -1;
	shm_addr = (char **)malloc( len );
	memset( (char *)shm_addr, (char)0, len );

/* Final sync to assure all structures are allocated and in place */
	msgsync();

#endif /* SHMEM */
	
#endif /* PVM */

	return( 0 );
}

#ifndef ISCRAY
int jpinit_( path, len_path )
char *path;
int len_path;
{
	char *p;
	int iret;

	p = (char *)malloc( len_path + 1 );
	strncpy( p, path, len_path );
	p[len_path] = '\0';
	iret = jpinit( p );
	free(p);
	return( iret );
}
#endif
#ifdef ISCRAY
#include "fortran.h"
int jpinit_( path )
_fcd path;
{
	char *p, *malloc();
	int len, iret;

	len = _fcdlen(path);
	p = malloc( len + 1 );
	strncpy( p, _fcdtocp(path), len );
	p[len] = '\0';
	iret = jpinit( p );
	free(p);
	return( iret );
}
#endif

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jpterm - terminate parallel processing
/*     
/*  SYNOPSIS:
/*
/*    ier = jpterm();
/*
/*--------------------------------------------------------------------*/

int jpterm()
{

#ifdef SHMEM
/* Release semaphores and shared memory */
	if (nodes > 1 ) {
		if (node == master) {
			if (sem_id >= 0)
				semctl( sem_id, 0, IPC_RMID, 0 );
			if (sem_sync >= 0)
				semctl( sem_sync, 0, IPC_RMID, 0 );
		}
		if (shm_id[node] >= 0)
			shmctl( shm_id[node], IPC_RMID, 0 );
	}
#endif /* SHMEM */

#ifdef PVM
	pvm_exit();
#endif /* PVM */

	return( 0 );
}

int jpterm_()
{
	return( jpterm() );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jmaster - return master node number
/*     
/*  SYNOPSIS:
/*
/*    integer master
/*    master = jmaster( )
/*
/*--------------------------------------------------------------------*/

int jmaster()
{
	if (trace)
		printf("jmaster node %d\n",node);
	return(0);
}

int jmaster_()
{
	return( jmaster() );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jnodes - return number of nodes in message passing complex
/*     
/*  SYNOPSIS:
/*
/*    integer nodes
/*    nodes = jnodes( )
/*
/*--------------------------------------------------------------------*/

int jnodes()
{
	if (trace)
		printf("jnodes  node %d\n",node);
	return( nodes );
}

int jnodes_()
{
	return( nodes );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jnode - return processor node number
/*     
/*  SYNOPSIS:
/*
/*    integer node
/*    node = jnode( )
/*
/*--------------------------------------------------------------------*/

int jnode()
{
	if (trace)
		printf("jnode   node %d\n",node);
	return( node );
}

int jnode_()
{
	return( node );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jkill - issue kill signal to entire processor complex
/*     
/*  SYNOPSIS:
/*
/*    call jkill( )
/*
/*--------------------------------------------------------------------*/

int jkill()
{
	int i,iret;

	iret = 0;
	if (trace)
		printf("jkill   node %d\n",node);

#ifdef PVM
	for (i=0; i<nodes; i++)
		if (node != i) iret = pvm_kill( tids[i] );
#endif /* PVM */

#ifdef SHMEM
	if (nodes <= 1) return(iret);
/* Release semaphores and shared memory */
	if (sem_id >= 0)
		semctl( sem_id, 0, IPC_RMID, 0 );
	if (sem_sync >= 0)
		semctl( sem_sync, 0, IPC_RMID, 0 );
	for (i=0; i<nodes; i++) {
		if (shm_id[i] >= 0) {
			shmctl( shm_id[i], IPC_RMID, 0 );
			shm_id[i] = -1;
		}
	}
#endif /* SHMEM */

	return( iret );
}

int jkill_()
{
	return( jkill() );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jsync - issue global sync signal to entire processor complex
/*     
/*  SYNOPSIS:
/*
/*    call jsync( )
/*
/*--------------------------------------------------------------------*/

int jsync()
{
	int ltrace;
	if (trace)
		printf("jsync   node %d\n",node);fflush(stdout);

#ifdef PVM
#ifndef SHMEM
	ltrace = trace;
	trace = 0;
	msgsync();
	trace = ltrace;
#endif /* ndef SHMEM */
#ifdef SHMEM
	semsync();
#endif /* SHMEM */
#endif /* PVM */

	return( 0 );
}

int jsync_()
/* Fortran entry point */
{
	return( jsync() );
}

#ifdef PVM

int msgsync()
/* Message passing based sync */
{
	int i,buf;

/* Master node waits for message from all others */
	buf = 0;
	if (node == master)
		for (i=1; i<nodes; i++)
			jrecv( 99+i, &buf, sizeof(int) );
	else
		jsend( 99+node, &buf, sizeof(int), master );

/* Broadcast to complete sync */
	jbrdcst(99, &buf, sizeof(int), master );

	return( 0 );
}

#ifdef SHMEM

int semsync()
/* Semaphore based sync routine */
{
	int i,j,ier,ival;
	unsigned flag;

	if (nodes <= 1) return(0);

/* Master waits for all other nodes to set semaphore */
	if (node == master) {
		do {
/* Get semaphore array */
			if (semctl(sem_sync, 0, GETALL, sem_val) != 0) 
				goto BOMB;
/* Check for any value not set */
			flag = 1;
			for (i=1; i<nodes; i++)
				if ((int)sem_val.array[i] != syncval) flag = 0;
		} while ( flag == 0 );
/* Now master sets his semaphore to same value */
		sem_arg.val = syncval;
		if (semctl( sem_sync, master, SETVAL, sem_arg ) != 0) 
			goto BOMB;
	}

/* Other nodes set value, then wait for master to set same */
	else {
/* Set semaphore to current barrier number */
		ier = 1;
		sem_arg.val = syncval;
		if (semctl( sem_sync, node, SETVAL, sem_arg ) != 0) 
			goto BOMB;
		if (trace) 
			printf("semsync: node %d set sem_sync = %d\n",node,syncval);
/* Loop and wait for master semaphore to be set */
		ier = 2;
		do {
/* Get master semaphore until value is set */
			ival = semctl( sem_sync, 0, GETVAL, 0 );
			if (ival < 0) goto BOMB;
		} while ( ival != syncval );
	}

	syncval = (syncval + 1)%256;
	if (trace) 
		printf("semsync: node %d next syncval = %d\n",node,syncval);
			
	return(0);

BOMB:
	fprintf(stderr,
		"semsync: SHMEM failure node %d ier %d errno = %d\n",
		node,ier,errno);
	perror("perror");
	fflush(stderr);
	jkill();
	exit(-1);
}

#endif /* SHMEM */

#endif /* PVM */

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jsend - send message
/*     
/*  SYNOPSIS:
/*
/*    integer type, length, dest
/*    character*1 buf(length)
/*    call jsend( type, buf, length, dest )
/*
/*--------------------------------------------------------------------*/

int jsend( type, buf, length, dest )
int type, length, dest;
char *buf;
{
	int i,iret;

	iret = 0;
	if (trace)
		printf("jsend   node %d: type, buf, length, dest = %d 0x%x %d %d\n",
			node,type,buf,length,dest);fflush(stdout);

#ifdef PVM
	if (dest != -1) {
		pvm_initsend(PvmDataRaw);
		if (length) pvm_pkbyte(buf, length, 1);
		iret= pvm_send(tids[dest], type);
	}
	else
		for (i=0; i<nodes; i++) {
			if (node != i) {
				pvm_initsend(PvmDataRaw);
				if (length) pvm_pkbyte(buf, length, 1);
				iret= pvm_send(tids[i], type);
			}
		}
#endif /* PVM */

	return( iret );
}

int jsend_( type, buf, length, node )
int *type, *length, *node;
char *buf;
{
	return( jsend( *type, buf, *length, *node ) );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jrecv - receive next message
/*     
/*  SYNOPSIS:
/*
/*    integer type, length
/*    character*1 buf(length)
/*    call jrecv( type, buf, length )
/*
/*--------------------------------------------------------------------*/

int jrecv( type, buf, length )
int type, length;
char *buf;
{
	int len,iret;

	iret = 0;
	if (trace)
		printf("jrecv   node %d: type, buf, length = %d 0x%x %d\n",
			node,type,buf,length);fflush(stdout);

#ifdef PVM
/* wait for message */
	iret = pvm_recv(-1, type);
/* Get length */
	pvm_bufinfo(iret, &len, 0, 0);
/* Unpack data */
	if (len) iret= pvm_upkbyte(buf, length, 1);
#endif /* PVM */

	return( iret );
}

int jrecv_( type, buf, length )
int *type, *length;
char *buf;
{
	return( jrecv( *type, buf, *length ) );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jasend - asynchronous send message
/*     
/*  SYNOPSIS:
/*
/*    integer msg, type, length, dest
/*    character*1 buf(length)
/*    msg = jasend( type, buf, length, dest )
/*
/*--------------------------------------------------------------------*/

int jasend( type, buf, length, dest )
int type, length, dest;
char *buf;
{
	int iret;

	iret = 0;
	if (trace)
		printf("jasend   node %d: type, buf, length, dest = %d 0x%x %d %d\n",
			node,type,buf,length,dest);fflush(stdout);

#ifdef PVM
	jsend( type, buf, length, dest );
	iret = type;
#endif /* PVM */

	return( iret );
}

int jasend_( type, buf, length, node )
int *type, *length, *node;
char *buf;
{
	return( jasend( *type, buf, *length, *node ) );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jarecv - asynchronous message receive
/*     
/*  SYNOPSIS:
/*
/*    integer msg, type, node, length
/*    character*1 buf(length)
/*    msg = jarecv( type, buf, length, node )
/*
/*--------------------------------------------------------------------*/

int jarecv( type, buf, length )
int type, length;
char *buf;
{
	int iret;

	iret = 0;
	if (trace)
		printf("jarecv   node %d: type, buf, length = %d 0x%x %d\n",
			node,type,buf,length);fflush(stdout);

#ifdef PVM
	iret = jrecv( type, buf, length );
#endif /* PVM */

	return( iret );
}

int jarecv_( type, buf, length )
int *type, *length;
char *buf;
{
	return( jarecv( *type, buf, *length ) );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jamsgwt - asynchronous message receive
/*     
/*  SYNOPSIS:
/*
/*    integer id
/*    call jamsgwt( id )
/*
/*--------------------------------------------------------------------*/

int jamsgwt( id )
int id;
{
	int iret;

	iret = 0;
	if (trace)
		printf("jamsgwt   node %d: id = %d\n", node, id);fflush(stdout);

#ifdef PVM
/* Loop until message is available */
	while ( (iret= pvm_probe(-1, id)) == 0)
		;
	if (iret > 0) iret = 0;
#endif /* PVM */

	return( iret );
}

int jamsgwt_( id )
int *id;
{
	return( jamsgwt( *id ) );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jprobe - Return flag indicating whether a message has
/*             been received
/*     
/*  SYNOPSIS:
/*    integer msgid, msg
/*    msg = jprobe( msgid )
/*
/*  DESCRIPTION:
/*    Returns zero if no message of the specified type is available.
/*    Returns the message type if one is available.  'id' may be
/*    set to '-1', indicating receipt of any message type.
/*    Returns -1 on errors.
/*
/*--------------------------------------------------------------------*/

int jprobe( msgid )
int msgid;
{
	int iret, bufid, msgtag, bytes, tid;

	iret = 0;
	if (trace)
		printf("jprobe   node %d: msgid = %d\n", node, msgid);fflush(stdout);

#ifdef PVM
/* Probe for message */
	bufid = pvm_probe( -1, msgid );
/* Return with zero if no matching message */
	if (bufid == 0) 
		return( iret );
	else if (bufid < 0)
		return( -1);
/* If match, return message id */
	iret = pvm_bufinfo( bufid, &bytes, &msgtag, &tid );
	if (iret == 0) 
		iret = msgtag;
	else
		iret = -1;
#endif /* PVM */

	return( iret );
}

int jprobe_( id )
int *id;
{
	return( jprobe( *id ) );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jbrdcst - broadcast message to all nodes
/*     
/*  SYNOPSIS:
/*
/*    integer type, length, sender
/*    character*1 buf(length)
/*    call jbrdcst( type, buf, length, sender )
/*
/*--------------------------------------------------------------------*/

int jbrdcst( type, buf, length, sender )
int type, length, sender;
char *buf;
{
	int iret;

	iret = 0;
	if (trace)
		printf("jbrdcst node %d: type, buf, length, sender = %d 0x%x %d %d\n",
			node,type,buf,length,sender);fflush(stdout);

#ifdef PVM
	if (node == sender)
		iret = jsend( type, buf, length, -1 );
	else
		iret = jrecv( type, buf, length );
#endif /* PVM */

	return( iret );
}

jbrdcst_( type, buf, length, sender )
int *type, *length, *sender;
char *buf;
{
	return( jbrdcst( *type, buf, *length, *sender ) );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jtrace - Toggle trace option
/*     
/*  SYNOPSIS:
/*
/*    integer iop
/*    call jtrace( iop )
/*
/*--------------------------------------------------------------------*/

int jtrace( iop )
int iop;
{
	trace = iop;
	if (trace)
		printf("jtrace  node %d: tracing is on, nodes, master = %d %d\n",node,nodes,master);
	else
		printf("jtrace  node %d: tracing is off\n",node);

	fflush(stdout);
	return(0);
}

int jtrace_( iop )
int *iop;
{
	return( jtrace( *iop ) );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jdcomp - Simple domain decomposition
/*    Allocates a block of data for each processor
/*     
/*  SYNOPSIS:
/*
/*    integer n, jntpn, jnlive
/*    call jdcomp( n, jntpn, jnlive )
/*
/*  DESCRIPTION:
/*    Divides 'n' by number of processors, and allocates 'jntpn'
/*    values to nodes 0 to n-2, and 'jnlive' values to node n-1.
/*    If n is an even multiple of the number of processors, then
/*    jnlive = jntpn.
/*
/*--------------------------------------------------------------------*/

int jdcomp( n, jntpn, jnlive )
int n, *jntpn, *jnlive;
{
	int nrem, nfull;

/* Compute number of traces to place on last processor */
	nrem = n % nodes;
	if (nrem > 0) {
		ntpn = (n/nodes) + 1;
		nfull = n/ntpn;
		if (node >= nfull) {
			nlive = n - (ntpn*node);
			if (nlive < 0) nlive = 0;
		}
		else
			nlive = ntpn;
	}
	else {
		ntpn = n/nodes;
		nlive = ntpn;
	}

	nglob = n;
	*jntpn = ntpn;
	*jnlive = nlive;

	if (trace)
		printf("jdcomp  node %d: nglob, ntpn, nlive = %d %d %d\n",
			node,nglob,ntpn,nlive );fflush(stdout);

	return(0);
}

/* Fortran interface */

int jdcomp_( n, jntpn, jnlive )
int *n, *jntpn, *jnlive;
{
	return( jdcomp( *n, jntpn, jnlive ) );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jrange - simple index range computation
/*     
/*  SYNOPSIS:
/*    integer n, j1, jn
/*    call jrange( n, j1, jn )
/*
/*  DESCRIPTION:
/*    Allocates 'n' values across processors using 'jdcomp', then
/*    returns the global index range of the values
/*    allocated to the current processor.  If no values are on this
/*    processor, returns -1 and a range of j1 = jn = 0.
/*
/*--------------------------------------------------------------------*/

int jrange( n, j1, jn )
int n, *j1, *jn;
{
	int iret;

/* Set decomposition if it's not done yet */
	if (n != nglob) jdcomp( n, &ntpn, &nlive );

/* Set range on this processor */
	if (nlive > 0) {
		*j1 = ntpn*node + 1;
		*jn = *j1 + nlive - 1;
		iret = 0;
	}
	else {
		*j1 = *jn = 0;
		iret = -1;
	}

	if (trace)
		printf("jrange node %d: n,j1,jn = %d %d %d\n",node,n,*j1,*jn);

	return( iret );
}

/* Fortran interface */

int jrange_( n, j1, jn )
int *n, *j1, *jn;
{
	return( jrange( *n, j1, jn ) );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jmsgrng - allocate a range of message numbers
/*     
/*  SYNOPSIS:
/*    integer nmsg, msgbase
/*    msgbase = jmsgrng( nmsg )
/*
/*  DESCRIPTION:
/*    Allocates 'nmsg' message numbers starting at the base number
/*    defined in 'basemsg'. All processors must call this routine
/*    or the allocation will get out of sync.
/*
/*--------------------------------------------------------------------*/

int jmsgrng( nmsg )
int nmsg;
{
	int iret;

	if (trace)
		printf("jmsgrng node %d: nmsg, basemsg = %d %d\n",
			node, nmsg, basemsg );

/* Set base message number for return */
	iret = basemsg;
	basemsg += nmsg;

	return( iret );
}

/* Fortran interface */

int jmsgrng_( nmsg )
int *nmsg;
{
	return( jmsgrng( *nmsg ) );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jtran23 - Distributed matrix transpose primitive
/*     
/*  SYNOPSIS:
/*    integer n, a, buf
/*    call jtran23( n, a, buf )
/*
/*  DESCRIPTION:
/*    Transposes last two dimensions of an (n,p,p) matrix, where 'p'
/*    is the number of processors, and the last dimension is 
/*    allocated across processors as in 'jdcomp'.  Uses a standard 
/*    tiling algorithm, developed by Intel SSD. See 'trans.f'
/*    for examples of how this primitive can be used to apply more
/*    general transpose operations.
/*
/*--------------------------------------------------------------------*/

#if !defined(SHMEM) && !defined(SERIAL)

/* This version uses message passing to exchange tiles of data */
/* All to all communication takes place */

int jtran23( n, a, buf )
int n, *a, *buf;
{
#	define MSG 10000
	int myswap, i, len, p, pmax;
	int *ap, mask;

	if (nodes <= 1) return(0);

/* Length in bytes of 1 row (tile) of array */
	len = n * sizeof(int);
 
/* Find power of 2 => number of nodes */
	pmax = 1;
	while (pmax < nodes)
		pmax = 2*pmax;

/* Loop and exchange tiles */
	for (i=1; i<pmax; i++) {
/* Processor to swap with is XOR of processor number and i */
		myswap = node ^ i;
/* Restrict swap in case nodes is not a power of 2 */
		if (myswap < nodes) {
/* tile number to send is same as target processor number */
			ap = a + n*myswap;
/* Swap tiles */
			jsend(MSG+i, ap, len, myswap);
			jrecv(MSG+i, buf, len);
			memcpy( ap, buf, len );
		}
	}

	return(0);
#	undef MSG
}

#endif /* MESSAGE */

/*--------------------------------------------------------------------*/

#ifdef SERIAL

/* This version serializes the tile exchange to prevent overloading */
/* the PVM demon and network for large numbers of processors */

int jtran23( n, a, buf )
int n, *a, *buf;
{
#	define MSG 10000
	int inode, iswap, i, j, len, pmax;
	int *ap, *done;

	if (trace) printf("jtran23: node %d nodes %d n d%\n",node,nodes,n);
	if (nodes <= 1) return(0);

/* Length in bytes of 1 row (tile) of array */
	len = n * sizeof(int);

/* Find power of 2 => number of nodes */
	pmax = 1;
	while (pmax < nodes)
		pmax = 2*pmax;

/* Allocate space for done flags */
	done = (int *)malloc( nodes * sizeof(int) );

/* Loop and exchange tiles */
	for (i=1; i<pmax; i++) {
		memset( done, 0, nodes*sizeof(int) );
		if (trace) printf("jtran23: node %d tile %d\n",node,i);
/* Use loop over nodes to serialize the swap */
		for (inode=0; inode<nodes; inode++) {
/* Processor swap pair is XOR of processor number and tile number i */
			iswap = inode ^ i;
			if (trace) printf("jtran23: inode %d iswap %d done %d\n",
				inode,iswap,done[inode]);
/* Participate in swap if needed */
			if ((inode == node) && (iswap < nodes) && 
					(done[node] == 0)) {
/* Originating node does a send-receive */
				ap = a + n*iswap;
				jsend(MSG+i, ap, len, iswap);
				jrecv(MSG+i, buf, len);
				memcpy( ap, buf, len );
/* Mark this tile done */
				done[inode] = 1;
			}
			else if ((iswap == node) && (done[node] == 0)) {
/* Target node does a receive-send */
				ap = a + n*inode;
				jrecv(MSG+i, buf, len);
				jsend(MSG+i, ap, len, inode);
				memcpy( ap, buf, len );
/* Mark this tile done */
				done[iswap] = 1;
			}
/* Sync at bottom of loop */
			jsync();
		}
	}

	free(done);
	return(0);
#	undef MSG
}

#endif /* SERIAL */

/*--------------------------------------------------------------------*/

#ifdef SHMEM

/* System V Shared memory segment implementation */

jtran23( n, a, buf )
int n, *a, *buf;
{
	int myswap, i, j, len, pmax, ier;
	int *ap;

	if (nodes <= 1) return(0);

/* Length in bytes of 1 row (tile) of array */
	len = n * sizeof(int);

/* Initialize shared memory resources */
	ier = shm_init(len);
	if (ier) goto BOMB;
 
/* Find power of 2 => number of nodes */
	pmax = 1;
	while (pmax < nodes)
		pmax = 2*pmax;

/* Loop and exchange tiles */
	for (i=1; i<pmax; i++) {

/* Processor to swap with is XOR of processor number and i */
		myswap = node ^ i;
/* Restrict swap in case nodes is not a power of 2 */
		if (myswap < nodes) {
			if (trace)
				printf("jtran23 node %d swap with %d\n",node,myswap);
/* tile number to send is same as target processor number */
			ap = a + n*myswap;

/* Copy data to local segment */
			memcpy( shm_addr[node], ap, len );
			semsync();

/* Copy data out of target segment */
			memcpy( ap, shm_addr[myswap], len );
				
		}
		else
			semsync();

/* Sync on each loop iteration */
		semsync();
	}

	return(0);

BOMB:
	fprintf(stderr,"jtran23: SHMEM failure, node %d ier %d errno %d\n",
		node, ier, errno);
	perror("perror");
	fflush(stderr);
	jkill();
	exit(-1);
}

/*--------------------------------------------------------------------*/

int shm_init( len )
/*
	Initialize shared memory resources
*/
int len; 	/* Requested length for shared memory segments */
{
	int j;

	if (trace) printf("shm_init node %d len %d\n",node,len);

/* check for max segment size exceeded */
	if (len > SHM_MAXSEG) return(-1);
	if (len <= shm_len) return(0);

/* Detach any existing segments */
	for( j=0; j<nodes; j++) {
		if ( shm_id[j] > 0 )
			if ( shmdt( shm_addr[j] ) ) return(-1);
		if (j!=node) shm_id[j] = -1;
	}

/* Sync to be sure all other nodes have detached */
	semsync();

/* Delete existing segment if it exists */
	if (shm_id[node] >= 0) {
		if ( shmctl( shm_id[node], IPC_RMID, 0 ) ) return(-2);
		shm_id[node] = -1;
	}

/* Create new segment */
	shm_id[node] = shmget( (key_t)(shm_key+node), 
		len, 0600 | IPC_CREAT | IPC_EXCL);
	if (shm_id[node] < 0) return(-3);
	shm_len = len;

/* Sync to be sure all segments have been created */
	semsync();

/* Attach segments */
	for (j=0; j<nodes; j++) {
		if (j != node) {
/* Get id for other nodes */
			shm_id[j] = shmget( (key_t)(shm_key+j), 
				len, 0600 );
			if (shm_id[j] < 0) return(-4);
		}
/* Attach segment */
		shm_addr[j] = (char *)shmat( shm_id[j], (char *)0, 0 );
		if (shm_addr[j] == (char *)(-1)) return(-5);
	}

	return(0);

}
	
#endif /* SHMEM */

/*--------------------------------------------------------------------*/

/* Fortran interface */

int jtran23_( n, a, buf )
int *n, *a, *buf;
{
	return( jtran23( *n, a, buf ) );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jshift - Distributed array shift primitive
/*     
/*  SYNOPSIS:
/*    integer nshift, la, a(la), buf(la)
/*    call jtran23( nshift, la, a, buf )
/*
/*  DESCRIPTION:
/*    Shifts a vector across processors.  'nshift' is the number of
/*    processors that data is to be shifted, 'la' is the length of
/*    the array to be shifted, 'a' is the input/output array, and
/*    'buf' is a work array, length 'la'.
/*
/*--------------------------------------------------------------------*/

int jshift( nshift, la, a, buf )
{
#	define MSG 11000
	int myswap, len;

	if (nodes <= 1 || nshift == 0) return(0);

/* number of bytes to be shifted */
	len = la * sizeof(int);
 
/* Processor to swap with */
	myswap = (node + nshift) % nodes;

/* Swap vectors */
	jsend(MSG+myswap, a, len, myswap);
	jrecv(MSG+node, buf, len);

/* Copy received data to output array */
	memcpy( a, buf, len );

	return(0);
#	undef MSG
}

/*--------------------------------------------------------------------*/

/* Fortran interface */

int jshift_( nshift, la, a, buf )
int *nshift, *la, *a, *buf;
{
	return( jshift( *nshift, *la, a, buf ) );
}

/*--------------------------------------------------------------------*/
