/*--------------------------------------------------------------------*/
/*  FILE:
/*    sbfio.c - seismic benchmark file I/O
/*
/*  VERSION:
/*    Initial version, no support for different data types
/*
/*  Routines defined in this module:
/*    sbf_creat - create SBF dataset
/*    sbf_open - open SBF dataset
/*    sbf_close - close SBF dataset
/*    sbf_paths - set leading path components for SBF files
/*    sbf_nfiles - set number of files used for a SBF dataset
/*    sbf_rdframe - read x-frame from SBF file
/*    sbf_wrframe - write x-frame to SBF file
/*    sbf_rdframet - read transposed (y-frame) from SBF dataset
/*    sbf_wrframet - write transposed (y-frame) to SBF dataset
/*    sbf_rdframep - read paneled frame) from SBF dataset
/*    sbf_wrframep - write paneled frame to SBF dataset
/*    sbf_rdcoord - read coordinates from SBF dataset
/*    sbf_wrcoord - write coordinates to SBF dataset
/*    sbf_rdcoordt - read transposed coordinates from SBF dataset
/*    sbf_wrcoordt - write transposed coordinates to SBF dataset
/*    sbf_rdcoordp - read paneled coordinates from SBF dataset
/*    sbf_wrcoordp - write paneled coordinates to SBF dataset
/*    sbf_rdparm - read parameters from SBF .HDR file
/*    sbf_wrparm - write parameters to SBF .HDR file
/*    sbf_qattr - query attribute existence in SBF .ATR file
/*    sbf_rdattr - read attribute from SBF .ATR file
/*    sbf_wrattr - write attribute to SBF .ATR file
/*    sbf_copy - memory to memory frame copy
/*    sbf_ftoc - float to char frame conversion
/*    sbf_stat - file status
/*    sbf_unlink - Remove SBF files
/*    sbf_perror - write error message explanation
/*     
/*--------------------------------------------------------------------*/
/*
/*  HISTORY:
/*
/*     Written:
/*       10/08/93 - C. Mosher, ARCO E&PT 
/*
/*     References:
/*
/*     Modified:
/*
/*--------------------------------------------------------------------*/

/* SBF Structure definition */
#include "sbfio.h"

/* NFS File locking */
#ifdef LOCK 
#	include <unistd.h>
#endif

/* Space for data set name */
static char dsn[512];

/* SBF error number */
int sbf_error;

/* Multiple file support */
static int nfiles = 0;
static int npaths = 0;
static char **lpaths = (char **)NULL;

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_creat - Create SBF file
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    char *path;
/*    int ns, ntpf, nfpv, nvol;
/*    sbf = sbf_creat( path, ns, ntpf, nfpv, nvol );
/*
/*  DESCRIPTION:
/*    Creates a new Seismic Benchmark File.  Uses 'path' to create
/*    3 files: path.HDR, path.XYZ, path.TRC, containing ASCII header 
/*    information, binary coordinates, and binary traces. The header
/*    file contains 'keyword=string' values of the input arguments,
/*    the coordinate file contains (x,y,z) locations of sources and
/*    receivers for each trace, and the trace file contains seismic
/*    traces.  The binary files are written to disk without record
/*    marks, 4 bytes per word.  The coordinate file has 6 words 
/*    (source x,y,z, receiver x,y,z) for each trace.  Seismic traces
/*    have 'ns' samples per trace, and are written as frames with
/*    'ntpf' traces per frame. 'nfpv' frames make up a volume, and
/*    the file contains 'nvol' individual 3D volumes.  Frames have
/*    'C' dimensions [ntpf][ns], 'Fortran' dimensions (ns,ntpf).
/*
/*    Additional file specific information can be written to the
/*    HDR file using the 'sbf_wparm' routines.
/*
/*    For files greater than 2 Gbytes, the .TRC files are broken
/*    up into 2GB chunks, and named .TRC0, .TRC1, etc.
/*
/*    This same mechanism can be used to support software striping.
/*    Use 'sbf_paths' to specify paths to multiple disks.  The
/*    .TRC files will be allocated to each disk in round-robin order.
/*
/*    Returns file pointer on success, NULL on errors.
/*
/*--------------------------------------------------------------------*/

SBFIO *sbf_creat( path, ns, ntpf, nfpv, nvol )
char *path;
int ns,ntpf,nfpv,nvol;
{
	SBFIO *sbf;
	FILE *fopen();
	char trc[8],base[64];
	int i, ipath, len, nfpc;
	float flen;

/* Allocate space for structures */
	sbf_error = SBF_MEM;
	sbf = (SBFIO *)malloc(sizeof(SBFIO));
	if (sbf == (SBFIO *)NULL) return(sbf);
	memset( sbf, 0, sizeof(SBFIO) );
	sbf->magic = SBF_MAGIC;

/* Return error if parameter file exists */
	strncpy(dsn,path,500);
	strcat(dsn,".HDR");
	if (sbf_stat(dsn) == 0) {
		sbf_error = SBF_CREAT;
		sbf_close( sbf );
		return( (SBFIO *)NULL );
	}

/* Create parameter and coordinate files */
	sbf->fp = fopen( dsn, "w");
	if (sbf->fp == (FILE *)NULL) {
		sbf_error = SBF_CREAT;
		sbf_close( sbf );
		return( (SBFIO *)NULL );
	}
	strncpy(dsn,path,500);
	strcat(dsn,".XYZ");
	sbf->cfp = open(dsn, O_RDWR | O_CREAT | O_EXCL, 0644);
	if (sbf->cfp < 0) {
		sbf_error = SBF_CRTB;
		sbf_close( sbf );
		return((SBFIO *)NULL);
	}
	strncpy(dsn,path,500);
	strcat(dsn,".ATR");
	sbf->afp = open(dsn, O_RDWR | O_CREAT | O_EXCL, 0644);
	if (sbf->afp < 0) {
		sbf_error = SBF_CRTB;
		sbf_close( sbf );
		return((SBFIO *)NULL);
	}

/* Set lengths in SBF structure */
	sbf->ns = ns;
	sbf->ntpf = ntpf;
	sbf->nfpv = nfpv;
	sbf->nvol = nvol;

/* Trace length and coordinate record lengths */
	sbf->trlen = ns*sizeof(float);
	sbf->crlen = 6*sizeof(float);

/* Trace and coordinate frame lengths */
	sbf->tflen = ntpf * sbf->trlen;
	sbf->cflen = ntpf * sbf->crlen;

/* Define multi-file values */

/* Default: Create files using default chunk size */ 
	if (nfiles == 0) { 
/* Set length of frame in words */
		len = ns*ntpf;
/* Compute number of frames that fit in a chunck */
		nfpc = MAXCHUNK / len;
		if (nfpc*len < MAXCHUNK) nfpc++;
/* Compute number of chunks in the total file */
		len = nfpv * nvol;
		sbf->nfiles = len / nfpc;
		if (sbf->nfiles * nfpc < len) sbf->nfiles++;
	}
	else {
/* Use preset value, but check to be be sure it's not too big */
		flen = (float)(ns*ntpf) * (float)(nfpv*nvol) / (float)nfiles;
		if (flen > (float)(MAXCHUNK)) {
			sbf_error = SBF_SIZE;
			sbf_close( sbf );
			return((SBFIO *)NULL);
		}
		sbf->nfiles = nfiles;
	}

/* Set number of frames per file */
	len = nfpv * nvol;
	sbf->nfpf = len / sbf->nfiles;
	if (sbf->nfpf * sbf->nfiles < len) sbf->nfpf = sbf->nfpf + 1;

/* Allocate space for multiple file pointers */
	len = sbf->nfiles * sizeof(int);
	sbf->mtfp = (int *)malloc( len );
	memset( sbf->mtfp, 0, len );
		
/* Store file structure and write arguments to header file */
	sbf_wrparm( sbf, "SamplesPerTrace", 'I', &ns );
	sbf_wrparm( sbf, "TracesPerFrame", 'I', &ntpf );
	sbf_wrparm( sbf, "FramesPerVolume", 'I', &nfpv );
	sbf_wrparm( sbf, "Volumes", 'I', &nvol );
	sbf_wrparm( sbf, "NumberOfFiles", 'I', &(sbf->nfiles) );
	sbf_wrparm( sbf, "FramesPerFile", 'I', &(sbf->nfpf) );

/* Create individual trace files */
	for (i=0; i<sbf->nfiles; i++) {

/* Attach leading path component, if it exists */
		if (npaths) {
			ipath = i%npaths;
			strcpy(dsn,lpaths[ipath]);
/* Add base name from input path */
			sbf_base(path,base);
			strcat(dsn,"/");
			strcat(dsn,base);
		}
		else
/* Attach local path */
			strncpy(dsn,path,500);

/* Attach qualifier */
		strcat(dsn,".");
		sprintf(trc,"TRC%d",i);
		if (sbf->nfiles == 1)
			strcat(dsn,"TRC");
		else
			strcat(dsn,trc);

/* Create the trace file, return on errors */
		sbf->mtfp[i] = open(dsn, O_RDWR | O_CREAT | O_EXCL, 0644);
		if (sbf->mtfp[i] < 0) {
			sbf_error = SBF_CRTB;
			sbf_close( sbf );
			return((SBFIO *)NULL);
		}

/* Write file names to header file */
		sbf_wrparm( sbf, trc, 'C', dsn );
	}

/* Return pointer to SBFIO structure */
	sbf_error = 0;
	return( sbf );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_open - Open SBF file
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    char *path;
/*    char *mode;
/*    sbf = sbf_open ( path, mode );
/*
/*  DESCRIPTION:
/*    Opens an existing Seismic Benchmark File.
/*    Returns file pointer on success, NULL on errors.
/*
/*--------------------------------------------------------------------*/

SBFIO *sbf_open( path, mode )
char *path;
char *mode;
{
	FILE *fp,*fopen();
	SBFIO *sbf;
	char rmode[4], trc[8];
	int ier, i, len;
	unsigned irw;

/* Allocate space for structures */
	sbf = (SBFIO *)malloc(sizeof(SBFIO));
	if (sbf == (SBFIO *)NULL) {
		sbf_error = SBF_MEM;
		return(sbf);
	}
	memset( sbf, 0, sizeof(SBFIO) );
	sbf->magic = SBF_MAGIC;

/* Set read/write modes */
	if (strncmp(mode,"rw",2) == 0) {
		irw = O_RDWR;
		strcpy( rmode, "r+" );
	}
	else {
		irw = O_RDONLY;
		strcpy( rmode, "r" );
	}

/* Check to see if header file exists */
	strncpy(dsn,path,500);
	strcat(dsn,".HDR");
	if (sbf_stat(dsn)) {
		sbf_close( sbf );
		sbf_error = SBF_OPNH;
		return( (SBFIO *)NULL );
	}
	sbf->fp = fopen( dsn, rmode );
	if (sbf->fp == (FILE *)NULL) { 
		sbf_close( sbf );
		sbf_error = SBF_OPNH;
		return((SBFIO *)NULL);
	}
	strncpy(dsn,path,500);
	strcat(dsn,".XYZ");
	sbf->cfp = open(dsn, irw );
	if (sbf->cfp < 0) {
		sbf_close( sbf );
		sbf_error = SBF_OPNB;
		return((SBFIO *)NULL);
	}
	strncpy(dsn,path,500);
	strcat(dsn,".ATR");
	sbf->afp = open(dsn, irw );
	if (sbf->afp < 0) {
		sbf_close( sbf );
		sbf_error = SBF_OPNB;
		return((SBFIO *)NULL);
	}

/* Read structure values */
	ier = ifetch( sbf->fp, "SamplesPerTrace", &(sbf->ns) );
	if (ier == 0)
		ier = ifetch( sbf->fp, "TracesPerFrame", &(sbf->ntpf) );
	if (ier == 0)
		ier = ifetch( sbf->fp, "FramesPerVolume", &(sbf->nfpv) );
	if (ier == 0)
		ier = ifetch( sbf->fp, "Volumes", &(sbf->nvol) );
	if (ier == 0)
		ier = ifetch( sbf->fp, "NumberOfFiles", &(sbf->nfiles) );
	if (ier == 0)
		ier = ifetch( sbf->fp, "FramesPerFile", &(sbf->nfpf) );

	if (ier) {
		sbf_close(sbf);
		sbf_error = SBF_INFO;
		return((SBFIO *)NULL);
	}

/* Open individual trace files */
	len = sbf->nfiles * sizeof(int);
	sbf->mtfp = (int *)malloc( len );
	memset( sbf->mtfp, 0, len );

	for (i=0; i<sbf->nfiles; i++) {
/* Read file name from header file */
		sprintf(trc,"TRC%d",i);
		sbf_rdparm( sbf, trc, 'C', dsn );
/* Open trace file */
		sbf->mtfp[i] = open( dsn, irw );
		if (sbf->mtfp[i] < 0) {
			sbf_error = SBF_CRTB;
			sbf_close( sbf );
			return((SBFIO *)NULL);
		}
	}

/* Trace length and coordinate record lengths */
	sbf->trlen = sbf->ns * sizeof(float);
	sbf->crlen = 6 * sizeof(float);

/* Trace and coordinate frame lengths */
	sbf->tflen = sbf->ntpf * sbf->trlen;
	sbf->cflen = sbf->ntpf * sbf->crlen;

	sbf_error = 0;
	return( sbf );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_close - Close an open SBF file
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    int ier;
/*    ier = sbf_close( sbf );
/*
/*  DESCRIPTION:
/*    CLoses an open Seismic Benchmark File.
/*
/*--------------------------------------------------------------------*/

sbf_close( sbf )
SBFIO *sbf;
{
	int i;

/* Check for valid structure values */
	if (sbf == (SBFIO *)NULL) return(-1);
	if (sbf->magic != SBF_MAGIC) return(-1);
/* Close files if open */
	for (i=0; i<sbf->nfiles; i++) 
		if (sbf->mtfp[i]) close(sbf->mtfp[i]);
	if (sbf->afp) close(sbf->afp);
	if (sbf->cfp) close(sbf->cfp);
	if (sbf->fp != (FILE *)NULL) fclose(sbf->fp);
/* Free structure memory */
	if (sbf->mtfp) free(sbf->mtfp);
	free(sbf);
	return(0);
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_nfiles - Set number of TRC files to use
/*     
/*  SYNOPSIS:
/*    int numfiles;
/*    sbf_close( numfiles );
/*
/*  DESCRIPTION:
/*    Sets number of .TRC files to uses to manage SBF data sets
/*    with trace data exceeding the 2GB file size limit
/*
/*--------------------------------------------------------------------*/

sbf_nfiles( numfiles )
int numfiles;
{
	nfiles = numfiles;
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_paths - Set path names for individual SBF files
/*     
/*  SYNOPSIS:
/*    char *pathfile;
/*    int ier;
/*    ier = sbf_paths( pathfile );
/*
/*  DESCRIPTION:
/*    Reads leading path component for .TRC files that make up
/*    an SBF dataset. This path component is pre-pended to the paths
/*    supplied to 'sbf_creat' and 'sbf_open'.  The file 'pathfile'
/*    should contain path components separated by white space, i.e.:
/*    /disk1/tmp  /disk2/tmp  /disk3/tmp  /disk4/tmp
/*    The .TRC files that make up the SBF dataset are allocated to
/*    the paths in round-robin order.
/*
/*--------------------------------------------------------------------*/

int sbf_paths( pathfile )
char *pathfile;
{
	FILE *fp, *fopen();
	int i,n;

/* Open file containing path names */
	fp = fopen( pathfile, "r" );
	if (fp == (FILE *)NULL) return(-1);

/* Count number of path names */
	n = 0;
	while (fscanf(fp,"%s",dsn) == 1)
		n++;
	
	if (n == 0) return(-1);

/* Free existing pointers, if they exist */
	for (i=0; i<npaths; i++)
		if (lpaths[i]) free(lpaths[i]);
	if (lpaths) free(lpaths);

/* Set new value for number of paths */
	npaths = n;

/* Allocate space for path name pointer array */
	lpaths = (char **)malloc( npaths * sizeof(int) );
	if (lpaths == (char **)NULL) return(-1);
	memset( lpaths, 0, npaths*sizeof(int) );

/* Loop and read path names */
	rewind(fp);
	for (i=0; i<npaths; i++) {
		if (fscanf(fp,"%s",dsn) != 1) return(-1);
		lpaths[i] = (char *)malloc( strlen(dsn) + 1 );
		if (lpaths[i] == (char *)NULL) return(-1);
		strcpy(lpaths[i],dsn);
	}

/* Close path file and return */
	fclose(fp);
	return(0);

}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_rdframe - Read a frame from an SBF dataset
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    int frame;
/*    float buf;
/*    int ier;
/*    ier = sbf_rdframe( sbf, frame, buf );
/*
/*  DESCRIPTION:
/*    Read a frame from and open SBF dataset.
/*    Returns 0 on success, negative on errors.
/*
/*--------------------------------------------------------------------*/

int sbf_rdframe( sbf, frame, buf )
SBFIO *sbf;
int frame;
float *buf;
{
	int fpos;

/* Seek to frame location */
	fpos = sbf_fseek( sbf, frame );
	if (fpos < 0) return(fpos);

/* Read frame */
	if ( read( sbf->tfp, buf, sbf->tflen ) != sbf->tflen ) {
		sbf_error = SBF_READ;
		return( -1 );	
	}
	
	return(0);
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_wrframe - Write a frame to an SBF dataset
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    int frame;
/*    float buf;
/*    int ier;
/*    ier = sbf_wrframe( sbf, frame, buf );
/*
/*  DESCRIPTION:
/*    Writes a frame to an open SBF dataset.
/*    Returns 0 on success, negative on errors.
/*
/*--------------------------------------------------------------------*/

int sbf_wrframe( sbf, frame, buf )
SBFIO *sbf;
int frame;
float *buf;
{
	int fpos;

/* Seek to frame location */
	fpos = sbf_fseek( sbf, frame );
	if (fpos < 0) return(fpos);

/* Write frame with lock protection */
#ifdef LOCK
	if ( lockf(sbf->tfp, F_LOCK, sbf->tflen) ) {
		sbf_error = SBF_LOCK;
		return(-1);
	}
#endif
	if ( write( sbf->tfp, buf, sbf->tflen ) != sbf->tflen ) {
		sbf_error = SBF_WRIT;
		return( -1 );	
	}
#ifdef LOCK
	if ( lockf( sbf->tfp, F_ULOCK, -(sbf->tflen) ) ) {
		sbf_error = SBF_LOCK;
		return(-1);
	}
#endif
	
	return(0);
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_rdframet - Read a transposed frame from an SBF dataset
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    int tframe;
/*    float buf;
/*    int ier;
/*    ier = sbf_rdframet( sbf, tframe, buf );
/*
/*  DESCRIPTION:
/*    Read a transposed frame from an open SBF dataset.
/*    A transposed frame has 'C' dimensions:
/*       [FramesPerVolume][SamplesPerFrame]
/*    Transposed frame number has the range 0:TracesPerFrame-1
/*    Returns 0 on success, negative on errors.
/*
/*--------------------------------------------------------------------*/

int sbf_rdframet( sbf, tframe, buf )
SBFIO *sbf;
int tframe;
float *buf;
{
	int i,fpos;

/* Check for valid structure */
	if (sbf->magic != SBF_MAGIC) {
		sbf_error = SBF_BAD;
		return( -1 );
	}

/* Loop and read traces */
	for (i=0; i<sbf->nfpv; i++) {
/* Seek to current location */
		if (sbf_tseek( sbf, i, tframe ) != 0) return(-1);
/* Read a single trace */
		if ( read( sbf->tfp, buf, sbf->trlen ) != sbf->trlen ) {
			sbf_error = SBF_READ;
			return( -1 );	
		}
/* Increment buffer position to next trace */
		buf += sbf->ns;
	}
	
	return(0);
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_wrframet - Write a transposed frame to an SBF dataset
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    int tframe;
/*    float buf;
/*    int ier;
/*    ier = sbf_wrframet( sbf, tframe, buf );
/*
/*  DESCRIPTION:
/*    Write a transposed frame to an open SBF dataset.
/*    A transposed frame has 'C' dimensions:
/*       [FramesPerVolume][SamplesPerFrame]
/*    Transposed frame number has the range 0:TracesPerFrame-1
/*    Returns 0 on success, negative on errors.
/*
/*--------------------------------------------------------------------*/

int sbf_wrframet( sbf, tframe, buf )
SBFIO *sbf;
int tframe;
float *buf;
{
	int i,fpos;

/* Check for valid structure */
	if (sbf->magic != SBF_MAGIC) {
		sbf_error = SBF_BAD;
		return( -1 );
	}

/* Loop and write traces */
	for (i=0; i<sbf->nfpv; i++) {
/* Seek to current location */
		if (sbf_tseek( sbf, i, tframe ) != 0) return(-1);

/* Write trace with lock protection */
#ifdef LOCK
		if ( lockf(sbf->tfp, F_LOCK, sbf->trlen) ) {
			sbf_error = SBF_LOCK;
			return(-1);
		}
#endif
		if ( write( sbf->tfp, buf, sbf->trlen ) != sbf->trlen ) {
			sbf_error = SBF_WRIT;
			return( -1 );	
		}
#ifdef LOCK
		if ( lockf(sbf->tfp, F_ULOCK, -(sbf->trlen)) ) {
			sbf_error = SBF_LOCK;
			return(-1);
		}
#endif

/* Increment buffer position by trace length */
		buf += sbf->ns;
	}
	
	return(0);
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_rdframep - Read a frame from an SBF dataset in paneled mode
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    int frame;
/*    float buf;
/*    int ier;
/*    ier = sbf_rdframep( sbf, frame, buf );
/*
/*  DESCRIPTION:
/*    Read a frame in paneled mode from an open SBF dataset.
/*    A paneled frame is spread across processors in the cluster.
/*    The YAMPL routine 'jdcomp' is used to set the number of 
/*    traces on each processor.  'jrange' returns the index range
/*    stored on each processor.
/*    Returns 0 on success, negative on errors.
/*
/*--------------------------------------------------------------------*/

int sbf_rdframep( sbf, frame, buf )
SBFIO *sbf;
int frame;
float *buf;
{
	int i,fpos,len,jt1,jtn;

/* Check for valid structure */
	if (sbf->magic != SBF_MAGIC) {
		sbf_error = SBF_BAD;
		return( -1 );
	}

/* Get index range */
	jrange( sbf->ntpf, &jt1, &jtn );

/* Seek to first trace position */
	if (sbf_tseek( sbf, frame, jt1-1 ) != 0) return(-1);

/* Set length to read */
	len = sbf->trlen * (jtn-jt1+1);

/* Read traces */
	if ( read( sbf->tfp, buf, len ) != len ) {
		sbf_error = SBF_READ;
		return( -1 );	
	}
	
	return(0);
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_wrframep - Write a frame from an SBF dataset in paneled mode
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    int frame;
/*    float buf;
/*    int ier;
/*    ier = sbf_rdframep( sbf, frame, buf );
/*
/*  DESCRIPTION:
/*    Write a frame in paneled mode an open SBF dataset.
/*    A paneled frame is spread across processors in the cluster.
/*    The YAMPL routine 'jdcomp' is used to set the number of 
/*    traces on each processor.  'jrange' returns the index range
/*    stored on each processor.
/*    Returns 0 on success, negative on errors.
/*
/*--------------------------------------------------------------------*/

int sbf_wrframep( sbf, frame, buf )
SBFIO *sbf;
int frame;
float *buf;
{
	int i,len,jt1,jtn;

/* Check for valid structure */
	if (sbf->magic != SBF_MAGIC) {
		sbf_error = SBF_BAD;
		return( -1 );
	}

/* Get index range */
	jrange( sbf->ntpf, &jt1, &jtn );

/* Seek to first trace position */
	if (sbf_tseek( sbf, frame, jt1-1 ) != 0) return(-1);

/* Set length to read */
	len = sbf->trlen * (jtn-jt1+1);

/* Write traces with lock protection */
#ifdef LOCK
	if ( lockf(sbf->tfp, F_LOCK, len) ) {
		sbf_error = SBF_LOCK;
		return(-1);
	}
#endif
	if ( write( sbf->tfp, buf, len ) != len ) {
		sbf_error = SBF_WRIT;
		return( -1 );	
	}
#ifdef LOCK
	if ( lockf(sbf->tfp, F_ULOCK, -len) ) {
		sbf_error = SBF_LOCK;
		return(-1);
	}
#endif
	
	return(0);
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_rdcoord - Read coordinates for a frame from an SBF dataset
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    int frame;
/*    float buf;
/*    int ier;
/*    ier = sbf_rdcoord( sbf, frame, buf );
/*
/*  DESCRIPTION:
/*    Reads coordinates for a frame from an open SBF dataset.
/*    Returned array has 'Fortran' dimensions (6,TracesPerFrame),
/*    and contains ( Source x,y,z, Receiver x,y,z ) for each trace.
/*    Returns 0 on success, negative on errors.
/*
/*--------------------------------------------------------------------*/

int sbf_rdcoord( sbf, frame, buf )
SBFIO *sbf;
int frame;
float *buf;
{
	int fpos;

/* Seek to coordinate record location */
	fpos = sbf_cseek( sbf, frame );
	if (fpos < 0) return(fpos);

/* Read coordinates */
	if ( read( sbf->cfp, buf, sbf->cflen ) != sbf->cflen ) {
		sbf_error = SBF_READ;
		return( -1 );	
	}
	
	return(0);
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_wrcoord - Write coordinates for a frame to an SBF dataset
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    int frame;
/*    float buf;
/*    int ier;
/*    ier = sbf_wrcoord( sbf, frame, buf );
/*
/*  DESCRIPTION:
/*    Writes coordinates for a frame to an open SBF dataset.
/*    Returns 0 on success, negative on errors.
/*
/*--------------------------------------------------------------------*/

int sbf_wrcoord( sbf, frame, buf )
SBFIO *sbf;
int frame;
float *buf;
{
	int ier;

/* Seek to coordinate record location */
	ier = sbf_cseek( sbf, frame );
	if (ier < 0) return(ier);

/* Write coordinates with lock protection */
#ifdef LOCK
	if ( lockf(sbf->cfp, F_LOCK, sbf->cflen) ) {
		sbf_error = SBF_LOCK;
		return(-1);
	}
#endif
	if ( write( sbf->cfp, buf, sbf->cflen ) != sbf->cflen ) {
		sbf_error = SBF_WRIT;
		return( -1 );	
	}
#ifdef LOCK
	if ( lockf(sbf->cfp, F_ULOCK, -(sbf->cflen)) ) {
		sbf_error = SBF_LOCK;
		return(-1);
	}
#endif
	
	return(0);
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_rdcoordt - Read transposed coordinates from an SBF dataset
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    int frame;
/*    float buf;
/*    int ier;
/*    ier = sbf_rdcoordt( sbf, frame, buf );
/*
/*  DESCRIPTION:
/*    Reads transposed coordinates from an open SBF dataset.
/*    Returns 0 on success, negative on errors.
/*
/*--------------------------------------------------------------------*/

int sbf_rdcoordt( sbf, frame, buf )
SBFIO *sbf;
int frame;
float *buf;
{
	int i,fpos;

/* Check for valid structure */
	if (sbf->magic != SBF_MAGIC) {
		sbf_error = SBF_BAD;
		return( -1 );
	}

/* Set first trace position */
	fpos = sbf->crlen * (frame + sbf->vpos * sbf->ntpf);

/* Loop and read coordinate records */
	for (i=0; i<sbf->nfpv; i++) {
/* Seek to current location */
		if (lseek( sbf->cfp, fpos, 0 ) != fpos) {
			sbf_error = SBF_SEEK;
			return(-1);
		}
/* Read a single coordinate record */
		if ( read( sbf->cfp, buf, sbf->crlen ) != sbf->crlen ) {
			sbf_error = SBF_READ;
			return( -1 );	
		}
/* Increment position by frame length to next trace */
		fpos += sbf->cflen;
		buf += 6;
	}
	
	return(0);
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_wrcoordt - Write transposed coordinates to an SBF dataset
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    int frame;
/*    float buf;
/*    int ier;
/*    ier = sbf_wrcoordt( sbf, frame, buf );
/*
/*  DESCRIPTION:
/*    Writes transposed coordinates to an open SBF dataset.
/*    Returns 0 on success, negative on errors.
/*
/*--------------------------------------------------------------------*/

int sbf_wrcoordt( sbf, frame, buf )
SBFIO *sbf;
int frame;
float *buf;
{
	int i,fpos;

/* Check for valid structure */
	if (sbf->magic != SBF_MAGIC) {
		sbf_error = SBF_BAD;
		return( -1 );
	}

/* Set first trace position */
	fpos = sbf->crlen * (frame + sbf->vpos * sbf->ntpf);

/* Loop and write coordinate records */
	for (i=0; i<sbf->nfpv; i++) {
/* Seek to current location */
		if (lseek( sbf->cfp, fpos, 0 ) != fpos) {
			sbf_error = SBF_SEEK;
			return(-1);
		}

/* Write a single coordinate record with lock protection */
#ifdef LOCK
		if ( lockf(sbf->cfp, F_LOCK, sbf->crlen) ) {
			sbf_error = SBF_LOCK;
			return(-1);
		}
#endif
		if ( write( sbf->cfp, buf, sbf->crlen ) != sbf->crlen ) {
			sbf_error = SBF_READ;
			return( -1 );	
		}
#ifdef LOCK
		if ( lockf(sbf->cfp, F_ULOCK, -(sbf->crlen)) ) {
			sbf_error = SBF_LOCK;
			return(-1);
		}
#endif

/* Increment position by frame length to next trace */
		fpos += sbf->cflen;
		buf += 6;
	}
	
	return(0);
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_rdcoordp - Read coords from an SBF dataset in paneled mode
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    int frame;
/*    float buf;
/*    int ier;
/*    ier = sbf_rdcoordp( sbf, frame, buf );
/*
/*  DESCRIPTION:
/*    Read coords for a frame in paneled mode an open SBF dataset.
/*    A paneled frame is spread across processors in the cluster.
/*    The YAMPL routine 'jdcomp' is used to set the number of 
/*    traces on each processor.  'jrange' returns the index range
/*    stored on each processor.
/*    Returns 0 on success, negative on errors.
/*
/*--------------------------------------------------------------------*/

int sbf_rdcoordp( sbf, frame, buf )
SBFIO *sbf;
int frame;
float *buf;
{
	int i,fpos,len,jt1,jtn;

/* Check for valid structure */
	if (sbf->magic != SBF_MAGIC) {
		sbf_error = SBF_BAD;
		return( -1 );
	}

/* Get index range */
	jrange( sbf->ntpf, &jt1, &jtn );

/* Set starting position and write length */
	fpos = sbf->crlen * (jt1-1) + sbf->cflen * (frame + sbf->vpos);
	len = sbf->crlen * (jtn - jt1 + 1);

/* Seek to location */
	if (lseek( sbf->cfp, fpos, 0 ) != fpos) {
		sbf_error = SBF_SEEK;
		return(-1);
	}
/* Read coords */
	if ( read( sbf->cfp, buf, len ) != len ) {
		sbf_error = SBF_READ;
		return( -1 );	
	}
	
	return(0);
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_wrcoordp - Write coords to an SBF dataset in paneled mode
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    int frame;
/*    float buf;
/*    int ier;
/*    ier = sbf_wrcoordp( sbf, frame, buf );
/*
/*  DESCRIPTION:
/*    Write coords for a frame in paneled mode to an open SBF dataset.
/*    A paneled frame is spread across processors in the cluster.
/*    The YAMPL routine 'jdcomp' is used to set the number of 
/*    traces on each processor.  'jrange' returns the index range
/*    stored on each processor.
/*    Returns 0 on success, negative on errors.
/*
/*--------------------------------------------------------------------*/

int sbf_wrcoordp( sbf, frame, buf )
SBFIO *sbf;
int frame;
float *buf;
{
	int i,fpos,len,jt1,jtn;

/* Check for valid structure */
	if (sbf->magic != SBF_MAGIC) {
		sbf_error = SBF_BAD;
		return( -1 );
	}

/* Get index range */
	jrange( sbf->ntpf, &jt1, &jtn );

/* Set starting position and write length */
	fpos = sbf->crlen * (jt1-1) + sbf->cflen * (frame + sbf->vpos);
	len = sbf->crlen * (jtn - jt1 + 1);

/* Seek to location */
	if (lseek( sbf->cfp, fpos, 0 ) != fpos) {
		sbf_error = SBF_SEEK;
		return(-1);
	}
/* Write coords */
#ifdef LOCK
	if ( lockf(sbf->cfp, F_LOCK, len) ) {
		sbf_error = SBF_LOCK;
		return(-1);
	}
#endif
	if ( write( sbf->cfp, buf, len ) != len ) {
		sbf_error = SBF_WRIT;
		return( -1 );	
	}
#ifdef LOCK
	if ( lockf(sbf->cfp, F_ULOCK, -len) ) {
		sbf_error = SBF_LOCK;
		return(-1);
	}
#endif
	
	return(0);
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_setvol - Set current volume in an open SBF dataset
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    int volume;
/*    int ier;
/*    ier = sbf_setvol( sbf, volume );
/*
/*  DESCRIPTION:
/*    Sets location of current volume in SBF .XYZ and .TRC files.
/*    Returns 0 on success, negative on errors.
/*
/*--------------------------------------------------------------------*/

int sbf_setvol( sbf, volume )
SBFIO *sbf;
int volume;
{

/* Check for valid structure */
	if (sbf->magic != SBF_MAGIC) {
		sbf_error = SBF_BAD;
		return( -1 );
	}

/* Check range */
	if (volume < 0 || volume >= sbf->nvol) {
		sbf_error = SBF_SEEK;
		return(-1);
	}

/* Set start frame location for the specified volume */
	sbf->vpos = volume * sbf->nfpv;

	return(0);
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_fseek - Seek to frame location in an open SBF dataset
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    int frame;
/*    int ier;
/*    ier = sbf_fseek( sbf, frame );
/*
/*  DESCRIPTION:
/*    Seeks to location of frame in the SBF .TRC file.
/*    Returns 0 on success, negative on errors.
/*
/*--------------------------------------------------------------------*/

int sbf_fseek( sbf, frame )
SBFIO *sbf;
int frame;
{
	unsigned fpos;
	int ifile, iframe;

/* Check for valid structure */
	if (sbf->magic != SBF_MAGIC) {
		sbf_error = SBF_BAD;
		return( -1 );
	}

/* Set file and frame numbers */
	frame = frame + sbf->vpos;
	ifile = frame/sbf->nfpf;
	sbf->tfp = sbf->mtfp[ifile];
	iframe = frame % sbf->nfpf;

/* Seek to location */
	fpos = sbf->tflen * (unsigned)iframe;
	if (lseek( sbf->tfp, fpos, 0 ) != fpos) {
		sbf_error = SBF_SEEK;
		return(-1);
	}

	return(0);
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_tseek - Seek to trace location in an open SBF dataset
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    int frame;
/*    int trace;
/*    int ier;
/*    ier = sbf_tseek( sbf, frame, trace );
/*
/*  DESCRIPTION:
/*    Seeks to location of trace in the SBF .TRC files.
/*    Returns 0 on success, negative on errors.
/*
/*--------------------------------------------------------------------*/

int sbf_tseek( sbf, frame, trace )
SBFIO *sbf;
int frame, trace;
{
	unsigned fpos;
	int ifile, iframe;

/* Check for valid structure */
	if (sbf->magic != SBF_MAGIC) {
		sbf_error = SBF_BAD;
		return( -1 );
	}

/* Set file and frame numbers */
	frame = frame + sbf->vpos;
	ifile = frame/sbf->nfpf;
	sbf->tfp = sbf->mtfp[ifile];
	iframe = frame % sbf->nfpf;

/* Seek to location */
	fpos = sbf->tflen * (unsigned)iframe + sbf->trlen * (unsigned)trace;
	if (lseek( sbf->tfp, fpos, 0 ) != fpos) {
		sbf_error = SBF_SEEK;
		return(-1);
	}

	return(0);
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_cseek - Seek to SBF coordinate record location
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    int frame;
/*    int ier;
/*    ier = sbf_cseek( sbf, frame );
/*
/*  DESCRIPTION:
/*    Seeks to location of coordinate record in the SBF .XYZ file.
/*    Returns 0 on success, negative on errors.
/*
/*--------------------------------------------------------------------*/

int sbf_cseek( sbf, frame )
SBFIO *sbf;
int frame;
{
	int fpos;

/* Check for valid structure */
	if (sbf->magic != SBF_MAGIC) {
		sbf_error = SBF_BAD;
		return( -1 );
	}

/* Seek to location */
	fpos = sbf->cflen * (frame + sbf->vpos);
	if (lseek( sbf->cfp, fpos, 0 ) != fpos) {
		sbf_error = SBF_SEEK;
		return(-1);
	}

	return(0);
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_rdparm - Read parameter from SBF .HDR file
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    char *name;
/*    char type;
/*    (char | int | float) *value;
/*    int ier;
/*    ier = sbf_rdparm( sbf, name, type, value );
/*
/*  DESCRIPTION:
/*    Read parameter from SBF .HDR file
/*    Return value is zero on success, negative on errors
/*
/*--------------------------------------------------------------------*/

sbf_rdparm( sbf, name, type, buf )

SBFIO *sbf;		/* SBF file pointer */
char *name;		/* parameter to read */
char type;		/* parameter type, 'I', 'F', or 'C' */
char *buf;		/* output value */

/*--------------------------------------------------------------------*/
{
	int iret;

/* Retrieve parameter based on type */
	switch (type) {
		case 'C':
			iret = fetch(sbf->fp,name,buf);
			break;
		case 'I':
			iret = ifetch(sbf->fp,name,(int *)buf);
			break;
		case 'F':
			iret = rfetch(sbf->fp,name,(float *)buf);
			break;
		default:
			return(-1);
	}

	if (iret >= 0)
		return(0);
	else {
		sbf_error = SBF_RPRM;
		return(-1);
	}
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_wrparm - Write parameter to SBF .HDR file
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    char *name;
/*    char type;
/*    (char | int | float) *value;
/*    int ier;
/*    ier = sbf_wrparm( sbf, name, type, value );
/*
/*  DESCRIPTION:
/*    Write parameter to SBF .HDR file
/*    Return value is zero on success, negative on errors
/*
/*--------------------------------------------------------------------*/

sbf_wrparm( sbf, name, type, buf )

SBFIO *sbf;		/* SBF file pointer */
char *name;		/* parameter to write */
char type;		/* parameter type, 'I', 'F', or 'C' */
char *buf;		/* input value */
{
	int iret;

/* Write to end of paramter file */
	if (fseek( sbf->fp, 0, 2 ) < 0) {
		sbf_error = SBF_WPRM;
		return(-1);
	}

/* Write parameter based on type */
	switch (type) {
		case 'C':
			iret = fprintf(sbf->fp,"%s=%s\n",name,buf);
			break;
		case 'I':
			iret = fprintf(sbf->fp,"%s=%d\n",name,*((int *)buf));
			break;
		case 'F':
			iret = fprintf(sbf->fp,"%s=%-12G\n",name,*((float *)buf));
			break;
		default:
			return(-1);
	}

	if (iret >= 0)
		return(0);
	else {
		sbf_error = SBF_WPRM;
		return(-1);
	}
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_qattr - Query existence of attribute in SBF .HDR file
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    char *name;
/*    int len;
/*    len = sbf_qattr( sbf, name );
/*
/*  DESCRIPTION:
/*    Returns length of attribute array in bytes on success,
/*    Returns -1 on errors.
/*
/*--------------------------------------------------------------------*/

sbf_qattr( sbf, name )

SBFIO *sbf;		/* SBF file pointer */
char *name;		/* parameter to read */

/*--------------------------------------------------------------------*/
{
	int iret, len;
	char attr[256];	

/* Attach length qualifier to attribute name */
	strncpy( attr, name, 250 );
	strcat( attr, ".len");

/* Retrieve from .HDR file */
	iret = ifetch( sbf->fp,name, &len );

/* Return length on success */
	if (iret == 0)
		return( len );
	else {
		sbf_error = SBF_RPRM;
		return(-1);
	}
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_rdattr - Read attribute array from SBF .ATR file
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    char *name;
/*    int *buf;
/*    int len;
/*    len = sbf_rdattr( sbf, name, buf );
/*
/*  DESCRIPTION:
/*    Reads attribute array from SBF .ATR file.
/*    Returns length of attribute array in bytes on success,
/*    Returns -1 on errors.
/*
/*--------------------------------------------------------------------*/

sbf_rdattr( sbf, name, buf )

SBFIO *sbf;		/* SBF file pointer */
char *name;		/* parameter to read */
int *buf;		/* output buffer containing attribute data */

/*--------------------------------------------------------------------*/
{
	int len, offset;
	char attr[256];	

/* Retrieve length and offset of attribute from SBF .HDR file */
	sbf_error = SBF_ATTR;
	strncpy( attr, name, 250 );
	strcat( attr, ".len");
	if (ifetch( sbf->fp, attr, &len )) return(-1);
	strncpy( attr, name, 240 );
	strcat( attr, ".offset");
	if (ifetch( sbf->fp, attr, &offset )) return(-1);

/* Read attribute from .ATR file */
	if (lseek( sbf->afp, offset, 0 ) != offset ) return(-1);
	if (read( sbf->afp, buf, len ) != len ) return(-1);
	
	sbf_error = 0;
	return(len);
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_wrattr - Write attribute array to SBF .ATR file
/*     
/*  SYNOPSIS:
/*    include "sbfio.h"
/*    SBFIO *sbf;
/*    char *name;
/*    int *buf;
/*    int len;
/*    int ier;
/*    ier = sbf_wrattr( sbf, name, buf, len );
/*
/*  DESCRIPTION:
/*    Writes attribute array to SBF .ATR file.
/*    Returns 0 on success, -1 on errors.
/*
/*--------------------------------------------------------------------*/

sbf_wrattr( sbf, name, buf, len )

SBFIO *sbf;		/* SBF file pointer */
char *name;		/* attribute name */
int *buf;		/* input buffer containing attribute data */
int len;		/* length in bytes of attribute data */

/*--------------------------------------------------------------------*/
{
	int offset;
	char attr[256];	

/* Write length and offset of attribute to SBF .HDR file */
	sbf_error = SBF_ATTR;
	strncpy( attr, name, 250 );
	strcat( attr, ".len");
	if ( sbf_wrparm( sbf, attr, 'I', &len ) ) return(-1);
	offset = lseek( sbf->afp, 0, SEEK_END );
	strncpy( attr, name, 240 );
	strcat( attr, ".offset");
	if ( sbf_wrparm( sbf, attr, 'I', &offset ) ) return(-1);

/* Write attribute to .ATR file */
	if (write( sbf->afp, buf, len ) != len ) return(-1);
	
	sbf_error = 0;
	return(0);
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_copy - Memory to memory frame copy
/*     
/*  SYNOPSIS:
/*    int n1a, n2a, n1b, n2b;
/*    float a[n2a][n1a], b[n2b][n1b];
/*    (void)sbf_copy( n1a, n2a, a, n1b, n2b, b );
/*
/*  DESCRIPTION:
/*    Copies a frame from one 2D array to another, possibly in-place.
/*    Sets a[j][i] = b[j][i], truncating 'b' if destination 'a' is 
/*    smaller, zero-fills if destination is larger.
/*
/*--------------------------------------------------------------------*/

sbf_copy( n1a, n2a, a, n1b, n2b, b )
int n1a, n2a, n1b, n2b;
float *a, *b;
{
	sbfcopy_( &n1a, &n2a, a, &n1b, &n2b, b );
}

/*--------------------------------------------------------------------*/

int sbf_ftoc( ns, ntpg, otr, buf, base, ncolor, smin, smax )

/*--------------------------------------------------------------------*/
/*  Conversion from float to byte format
/*--------------------------------------------------------------------*/

int ns;			/* samples per trace */
int ntpg;		/* traces per group */
float *otr;	     /* input float array, fortran dim (ns,ntpg) */
char *buf;	      /* output char array, transposed (C buf[ns][ntpg]) */
int base;		/* base pixel value */
int ncolor;		/* number of colors */
float smin,smax;		/* Conversion range */

/*--------------------------------------------------------------------*/

{
	int i,j,k,val,scale;
	float fact,*x;

/* Set range if smin == smax == 0 */
	if (smin == 0.0 && smax == 0.0) {
		smin = smax = *otr;
		for (j=0; j<ntpg; j++) {
			x = otr + ns * j;
			for (i=0; i<ns; i++,x++) {
				if (*x > smax) 
					smax = *x;
				else if (*x < smin) 
					smin = *x;
			}
		}
		if (smin < 0.0) {
			smin = -smin;
			if (smin < smax) 
				smin = smax;
			else
				smax = smin;
			smin = -smin;
		}
		else if (smin == smax) {
			smax = smax + 1.0;
			smin = -smax;
		}
	}

/* Loop and scale */
	scale = ncolor-1;
	fact = (float)(scale)/(smax - smin);
	for (j=0; j<ntpg; j++) {
		x = otr + ns * j;
		k = j;
		for (i=0; i<ns; i++,x++) {
			val = (int)(fact * (*x - smin));
			if (val > scale) val = scale;
			if (val < 0) val = 0;
			buf[k] = (char)(val + base);
			k += ntpg;
		}
	}

	return(0);
}
 

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_base - Return base name of SBF path, stripping off leading
/*               path components
/*     
/*  SYNOPSIS:
/*    char *path,*base;
/*    int ier;
/*    ier = sbf_base(path,base);
/*
/*  DESCRIPTION:
/*    Return error code concerning file existence.
/*--------------------------------------------------------------------*/

int sbf_base( path, base )
char *path, *base;
{
	int i,j,len;

/* Find last '/' in input path string */
	len = strlen(path);
	for( i=len-1; i>0; i--) {
		if (path[i] == '/') {
/* Copy from input to output leaving off the leading path */
			strcpy( base, path+i+1 );
			return(0);
		}
	}

/* No leading path component */
	strcpy( base, path );
 
	return(0);
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_stat - File status
/*     
/*  SYNOPSIS:
/*    char *path;
/*    int ier;
/*    ier = sbf_stat(path);
/*
/*  DESCRIPTION:
/*    Return error code concerning file existence.
/*    Value returned in ier:
/*    0 - file exists
/*    1 - file does not exist
/*    2 - illegal path name
/*
/*--------------------------------------------------------------------*/

int sbf_stat(path) 
char *path;
{
#	include <sys/types.h>
#	include <sys/stat.h>
#	include <errno.h>
#	ifndef errno
	extern int errno;
#	endif
	struct stat buf;
	int iret;

/* make stat call to determine file state */
	iret = stat(path,&buf);
	if (iret) {
		if (errno == ENOENT)
/* File does not exist */
			return(1);
		else
/* Path or access error */
			return(2);
	}
	else
/* File exists */
		return(0);

}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    sbf_unlink - Remove SBF files
/*     
/*  SYNOPSIS:
/*    char *path;
/*    int ier;
/*    ier = sbf_unlink(path);
/*
/*  DESCRIPTION:
/*    Removes the .TRC, .XYZ, and .HDR files for a SBF dataset.
/*    Returns zero on success, non-zero on errors.
/*
/*--------------------------------------------------------------------*/

int sbf_unlink(path) 
char *path;
{
	int i, ier, nf, ipath;
	char trc[8],base[64];

/* Remove the three aux files */
	ier = 0;
	strncpy(dsn,path,500);
	strcat(dsn,".HDR");
	ier = ier | unlink(dsn);
	strncpy(dsn,path,500);
	strcat(dsn,".XYZ");
	ier = ier | unlink(dsn);
	strncpy(dsn,path,500);
	strcat(dsn,".ATR");
	ier = ier | unlink(dsn);

/* Remove the trace files */
	nf = nfiles;
	if (nf == 0) nf = 1;
	for (i=0; i<nf; i++) {
/* Attach leading path component, if it exists */
		if (npaths) {
			ipath = i%npaths;
			strcpy(dsn,lpaths[ipath]);
/* Add base name from input path */
			sbf_base(path,base);
			strcat(dsn,"/");
			strcat(dsn,base);
		}
		else
/* Attach local path */
			strncpy(dsn,path,500);

/* Attach qualifier */
		strcat(dsn,".");
		sprintf(trc,"TRC%d",i);
		if (nf == 1)
			strcat(dsn,"TRC");
		else
			strcat(dsn,trc);

/* Remove file */
		ier = ier | unlink(dsn);
	}

	return(ier);
}

/*--------------------------------------------------------------------*/

int sbf_perror(stream,s)

/*--------------------------------------------------------------------*/
/*	Write short explanation of error to an I/O stream.
/*	If stream is null, write to standard error.
/*--------------------------------------------------------------------*/

FILE *stream;	/* I/O steam */
char *s;		/* string to precede explanation */

/*--------------------------------------------------------------------*/
{
	FILE *fp;

	if (stream == (FILE *)NULL)
		fp = stderr;
	else
		fp = stream;

/* Set default message */
	if (s == (char *)NULL) s = "SBF_io:";

	switch (sbf_error) {
		case 0:
			fprintf(fp,"%s last error was zero\n",s);
			break;
		case SBF_MEM:
			fprintf(fp,
			"%s SBF io structure memory allocation failed\n",s);
			break;
		case SBF_OPNH:
			fprintf(fp,
			"%s could not open SBF header file\n",s);
			perror(0);
			break;
		case SBF_OPNB:
			fprintf(fp,
			"%s could not open SBF binary file\n",s);
			perror(0);
			break;
		case SBF_CRTH:
			fprintf(fp,
			"%s could not create SBF header file\n",s);
			perror(0);
			break;
		case SBF_CRTB:
			fprintf(fp,
			"%s could not create SBF binary file\n",s);
			perror(0);
			break;
		case SBF_READ:
			fprintf(fp,
			"%s error reading SBF binary file\n",s);
			perror(0);
			break;
		case SBF_WRIT:
			fprintf(fp,
			"%s error writing SBF binary file\n",s);
			perror(0);
			break;
		case SBF_SEEK:
			fprintf(fp,
			"%s error seeking to location in SBF binary file\n",s);
			perror(0);
			break;
		case SBF_INFO:
			fprintf(fp,
			"%s required parms missing in SBF header file\n",s);
			break;
		case SBF_CREAT:
			fprintf(fp,
			"%s cannot create, SBF file already exists\n",s);
			perror(0);
			break;
		case SBF_RPRM:
			fprintf(fp,
			"%s error reading parameter from .HDR file\n",s);
			perror(0);
			break;
		case SBF_WPRM:
			fprintf(fp,
			"%s error writing parameter to .HDR file\n",s);
			perror(0);
			break;
		case SBF_RANGE:
			fprintf(fp,
			"%s function parameter is out of range\n",s);
			break;
		case SBF_BAD:
			fprintf(fp,
			"%s SBF structure is corrupted\n",s);
			break;
		case SBF_LOCK:
			fprintf(fp,
			"%s SBF file lock failed\n",s);
			perror(0);
			break;
		case SBF_SIZE:
			fprintf(fp,
			"%s file size too large, increase nfiles\n",s);
			break;
		case SBF_ATTR:
			fprintf(fp,
			"%s Error while accessing .ATR file\n",s);
			perror(0);
			break;
		default:
			fprintf(fp,"%s invalid error number\n",s);
	}
	return(0);
}

/*--------------------------------------------------------------------*/
/* Reverse byte order for endian differences
/*--------------------------------------------------------------------*/

int sbf_swap( buf, n )

char *buf;
int n;
{
	int j;
	char temp;

    for (j=0; j<4*n; j+=4) {
		temp=buf[j];
		buf[j]=buf[j+3];
		buf[j+3]=temp;
		temp=buf[j+1];
		buf[j+1]=buf[j+2];
		buf[j+2]=temp;
	}

}
