C*======================= START OF SUBROUTINE ==========================
C*
C*  NAME:
C*    syncini - Initialize optimal sinc interpolation operators
C*     
C*  SYNOPSIS:
C*    integer lr
C*    real bw
C*    real coef(lr,101)
C*    real sa(4*lr+4)
C*    call syncini( lr, bw, coef, sa )
C*
C*  DESCRIPTION: 
C*    Generates 101 convolution operators of length 'lr' for use
C*    in interpolating band limited time series.  Fractional band-
C*    width (relative to Nyquist) is specified by the parameter 'bw',
C*    valid range ( 0.0, 1.0 ).  Convolutional operators are returned
C*    in array 'coef'.  The filters can be used to interpolate between
C*    samples to the nearest 0.01 sample unit.  For example:
C*
C*           X1    X2  P X3   X4
C*
C*    Point P is 0.65 samples from X1.  Given lr=4, point P is approx-
C*    imated by the dot product of (X0...X3) and coef(1,65...4,65).
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*   Math Advantage:
C*     vfill - set array to constant
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        04/15/90 - C. Mosher
C*
C*      References:
C*        C.J. Sicking, PhD dissertation
C*
C*      Modified:
C*        07/14/90 - C. Mosher
C*                   Converted names to OSPARC convention
C*
C*----------------------------------------------------------------------
      SUBROUTINE SYNCINI(LR,BW,COEF,SA)
C  CONSTRUCT SICKING INTERPOLATION FILTERS
C  LR - OPERATOR LENGTHS
C  BW - BANDWIDTH, 0.0 to 1.0
C  COEF(LR,101) - OUTPUT OPERATORS
C  SA - WORK ARRAY, LENGTH 4*LR + 4
      DIMENSION COEF(LR,101),SA(*)
C  COMPUTE FILTERS 2-100
      CALL SYNCGEN(LR,BW,SA,SA(LR+1),SA(2*LR+1),SA(3*LR+3),COEF(1,2))
C  SET FIRST AND LAST FILTERS TO GIVE LOWER AND UPPER SAMPLES
      CALL VFILL(0.0,COEF(1,1),1,LR)
      LR2 = LR/2
      COEF(LR2,1) = 1.0
      CALL VFILL(0.0,COEF(1,101),1,LR)
      COEF(LR2+1,101) = 1.0
      RETURN
      END
C
C
      SUBROUTINE SYNCGEN(LR,BW,RW,G,FF,A,COEF)
      DIMENSION RW(1),G(1),FF(1),A(1),COEF(LR,99)
C**********************************************************************
C
C   AUTHOR   C.J. SICKING
C
C   COFGEN COMPUTES 99 INTERPOLATION FILTERS ON EACH CALL.  EACH
C   FILTER IS LR COEFFICIENTS IN LENGTH.  THESE 99 FILTERS CAN BE
C   USED TO COMPUTE THE VALUE OF A SIGNAL TO THE NEAREST 0.01 SAMPLE
C   INTERVAL.  COMPUTING THE VALUE TO THE NEAREST 0.01 SAMPLE
C   INTERVAL IS ADEQUATE TO MAINTAIN 30 TO 40 DB FIDELITY IN THE
C   INTERPOLATION PROCESS
C
C        X1          X2       P  X3          X4
C
C   IF WE HAVE A SIGNAL WITH SAMPLES AT LOCATIONS X1,X2,X3 AND X4 AND
C   WE WANT THE VALUE OF THE SIGNAL AT LOCATION P, WE WOULD FIRST
C   COMPUTE THE FRACTIONAL DISTANCE BY DIVIDING THE DISTANCE P-X2 BY
C   THE DISTANCE X3-X2 AND THEN ROUND THE RESULT TO THE NEAREST 0.01
C   AND MULTIPLY BY 100.  THE RESULT IS A NUMBER BETWEEN 0 AND 100.
C   THIS IS THE INDEX  TO THE FILTER ARRAY AND IS USED TO DETERMINE
C   WHICH OF THE 99 FILTERS TO USE IN THE INTERPOLATION.  IF THE RESULT
C   IS 0 OR 100, P IS WITHIN PLUS OR MINUS 0.005*(SAMPLES INTERVAL)
C   OF X2 OR X3 AND IS CLOSE ENOUGH TO USE THE
C   KNOWN VALUE WITHOUT GOING THROUGH THE INTERPOLATION STEP.
C
C
C   THE FILTERS COMPUTED ARE WIENER FILTERS.  A CORRELATION FUNCTION
C   IS REQUIRED IN ORDER TO COMPUTE THE FILTERS.  THE FILTERS ARE
C   COMPUTED USING THE MATRIX EQUATION:
C
C               A F = C
C
C   WHERE A IS THE AUTOCORRELATION MATRIX AND C IS THE CROSS CORRELATION
C   MATRIX.  THE EQUATION IS SOLVED USING A MODIFIED VERSION OF
C   SUBROUTINE EUREKA (CALLED SYNCCMP HERE) FROM ROBINSON'S BOOK.
C   THE CORRELATION FUNCTION USED IS A BANDLIMITED SINC FUNCTION.
C   FOR DETAILS OF THE INTERPOLATION PROCEDURE AND HOW TO THE PARAMETERS
C   ARE RELATED SEE SICKING' PH.D. DISSERTATION.
C   LR AND BW ARE INTERRELATED AND THEIR VALUES SHOULD BE CHOSEN BASED
C   ON KNOWGLEDGE OF THE MAXIMUM FREQUENCY PRESENT IN THE SIGNAL AND
C   THE CURVES PRESENTED IN SICKING'S DISSERTATION CHAPTER 4.
C
C   PARAMETERS
C
C   LR    -  NUMBER OF COEFFICIENTS IN THE FILTERS
C   BW    -  MAXIMUM FREQUENCY AS A FRACTION OF NYQUIST
C   RW    -  ARRAY FOR AUTOCORRELATION
C   G     -  ARRAY FOR CROSS CORRELATION
C   FF    -  FILTER ARRAY USED BY SYNCCMP
C   A     -  WORK ARRAY USED BY SYNCCMP
C   COEF  -  COEFFICIENT ARRAY (OUTPUT).  99 FILTERS OF LENGTH LR.
C
C   ARRAYS RW AND G NEED TO BE DIMENSIONED TO LENGTH LR IN THE
C   CALLING PROGRAM.
C   ARRAYS FF AND A NEED TO BE DIMENSIONED TO LENGTH LR+2 IN THE
C   CALLING PROGRAM.
C
C***********************************************************************
C
C   COMPUTE AUTOCORRELATION VALUES
C
C   THE AUTOCORRELATION MATRIX IS THE SAME FOR ALL FILTERS COMPUTED
C   SO IT NEEDS TO BE COMPUTED ONLY ONCE.
C
C   THE PARAMETER CORD IS THE LAG IN NUMBER OF SAMPLE INTERVALS.
C
      PI=3.141592654
      CORD=0.0
      DO 100 I=1,LR
      TX=CORD*PI*BW
      IF(TX.EQ.0.0)GO TO 90
      VAL=SIN(TX)/TX
      GO TO 95
   90 VAL=1.0
   95 RW(I)=VAL
  100 CORD=CORD+1.0
C
C   COMPUTE EACH OF FIRST 50 FILTERS
C
C   LCN IS INCREMENTED FROM 1 TO 50 WHILE FRAC IS INCREMENTED FROM
C   O.01 TO 0.50.  THE CROSS CORRELATION ARRAY IS FILLED ,THE FILTER
C   IS COMPUTED, AND THE FILTER IS STORED IN THE OUTPUT COEF ARRAY.
C
C   THE PARAMETER CORDIS IS THE CORRELATION LAG IN SAMPLE SPACINGS
C   BETWEEN A GIVEN SAMPLE AND THE DESIRED SAMPLE LOCATION.
C
      LCN=1
      FRAC=0.01
  190 CORDIS=-((LR/2)-1)-FRAC
C
C   COMPUTE CROSS CORRELATION VALUES
C
      DO 200 I=1,LR
      TX=CORDIS*PI*BW
      IF(TX.EQ.0.0)GO TO 196
      VAL=SIN(TX)/TX
      GO TO 198
  196 VAL=1.0
  198 G(I)=VAL
  200 CORDIS=CORDIS+1.0
C
C   COMPUTE FILTER AND STORE IN COEF ARRAY
C
      CALL SYNCCMP(LR,RW,G,FF,A)
      DO 210 I=1,LR
  210 COEF(I,LCN)=FF(I)
C
C   INCREMENT FRAC AND LCN AND RETURN TO COMPUTE THE NEXT FILTER.
C
      FRAC=FRAC+0.01
      LCN=LCN+1
      IF(LCN.LT.51)GO TO 190
C
C   THE FIRST 50 FILTERS HAVE BEEN COMPUTED.  FILTER 50 IS THE
C   FILTER TO BE USED FOR MIDPOINT INTERPOLATION.
C
C   THERE IS SYMETRY ABOUT MIDPOINT. REVERSE FIRST 49 FILTERS AND
C   STORE IN LAST 49.
C
      DO 230 I=1,49
      J=50-I
      K=50+I
      DO 220 L=1,LR
      M=LR-L+1
  220 COEF(M,K)=COEF(L,J)
  230 CONTINUE
      RETURN
      END
C
C
      SUBROUTINE SYNCCMP(LR,R,G,F,A)
C**********************************************************************
C
C    MATRIX INVERSION ROUTINE MODIFIED FROM ROBINSON'S BOOK
C
C    PARAMETERS
C
C    LR   -  NUMBER OF FILTER COEFFICIENTS TO BE COMPUTED
C    R    -  AUTOCORRELATION MATRIX
C    G    -  CROSS CORRELATION MATRIX
C    F    -  FILTER COMPUTED BY SYNCCMP
C    A    -  ARRAY CONTAINING SPIKING  DECONVOLUTION FILTER.
C
C*********************************************************************
      DIMENSION R(2),G(2),F(3),A(4)
      V=R(1)
      D=R(2)
      A(1)=1.0
      F(1)=G(1)/V
      Q=F(1)*R(2)
      IF(LR.EQ.1)RETURN
      DO 4 L=2,LR
      A(L)=-D/V
      IF(L.EQ.2)GO TO 2
      L1=(L-2)/2
      L2=L1+1
      IF(L2.LT.2)GO TO 5
      DO 1 J=2,L2
      HOLD=A(J)
      K=L-J+1
      A(J)=A(J)+A(L)*A(K)
    1 A(K)=A(K)+A(L)*HOLD
    5 IF(2*L1.EQ.L-2)GO TO 2
      A(L2+1)=A(L2+1)+A(L)*A(L2+1)
    2 V=V+A(L)*D
      F(L)=(G(L)-Q)/V
      L3=L-1
      DO 3 J=1,L3
      K=L-J+1
    3 F(J)=F(J)+F(L)*A(K)
      IF(L.EQ.LR)RETURN
      D=0.0
      Q=0.0
      DO 4 I=1,L
      K=L-I+2
      D=D+A(I)*R(K)
    4 Q=Q+F(I)*R(K)
      RETURN
      END
C
C*======================= START OF SUBROUTINE ==========================
C*
C*  NAME:
C*    synccnm - Complex sync interpolation normalization coefficients
C*     
C*  SYNOPSIS:
C*    real bw
C*    integer n2,nf
C*    complex cmod(nf), cnrm(101)
C*    call synccnm( bw, n2, nf, cmod, cnrm )
C*
C*  DESCRIPTION: 
C*    Used with optimal sync interpolation operators to perform
C*    interpolation of band-limited frequency domain time series.  
C*    'bw' is the fractional bandwidth of the sync interpolators
C*    (see routine syncini).  'n2' is the fourier transform length,
C*    'nf' is the number of frequencies to be interpolated.  The
C*    data to be interpolated should be premultiplied at each 
C*    frequency by conjg(cmod) to reduce phase variations.  After
C*    interpolation of real and imaginary parts using sync operators,
C*    multiply by cmod to restore original phase, and by
C*    cnrm(index) to compensate for fractional sample shift.  'index' 
C*    is the index of the sync operator used in the interpolation.
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*      Libraries:
C*        FORTRAN intrinsic functions
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        04/15/90 - C. Mosher
C*
C*      References:
C*        B.J. Verwest, Pre-Stack FK Migration Technical Documentation
C*        W.   Harlan, Stanford Exploration Project Rept 30
C*
C*      Modified:
C*        07/14/90 - C. Mosher 
C*                   Converted to OSPARC naming convention
C*
C*----------------------------------------------------------------------
      SUBROUTINE SYNCCNM(BW,N2,NF,CMOD,CNRM)
C
C  GENERATE MODULATION AND NORMALIZATION OPERATORS
C  BW - FRACTIONAL BANDWIDTH FOR OPERATORS (SAME AS SICKING INTERP)
C  N2 - FFT LENGTH
C  NF - MAX NUMBER OF FREQUENCIES TO BE PROCESSED
C  COMPLEX CMOD(NF) - FREQUENCY MODULATION FACTORS
C  COMPLEX CNRM(101) - FRACTIONAL SAMPLE NORMALIZATION FACTORS
C  BASED ON TECHNIQUES WORKED OUT BY B.J. VERWEST, DERIVED FROM
C    W. HARLAN'S NOTE IN SEP 30.
C  THE DATA SHOULD BE PRE-MULTIPLIED BY CONGJ(CMOD(I))
C  TO REDUCE PHASE VARIATION.  AFTER INTERPOLATION, THE DATA SHOULD
C  BE MULTIPLIED BY CMOD TO RESTORE THE ORIGINAL PHASE, AND
C  BY CNRM(IFRAC) TO COMPENSATE FOR THE FRACTIONAL SAMPLE SHIFT.
C
      COMPLEX CMOD(NF),CNRM(101)
C  PI TIMES FRACTIONAL BANDWIDTH
      PIBW = 4.*ATAN(1.)*BW
C  SET MODULATION FACTORS
      DO 10 I=1,NF
          ARG = PIBW*(1-I)
          CMOD(I) = CMPLX(COS(ARG),SIN(ARG))
  10  CONTINUE
C  SET FRACTIONAL SAMPLE PHASE CORRECTIONS
      DO 20 I=1,101
          ARG = PIBW*0.01*(1-I)
          CNRM(I) = CMPLX(COS(ARG),SIN(ARG))
  20  CONTINUE
      RETURN
      END
