C*-------------------------------------------------------------------
C*
C*  Transpose routines for message style programming
C*
C*  Routines defined in this module:
C*
C*    dtran21 - transpose distributed 2D array
C*    dtran132 - transpose last two dimensions of 3D distributed array
C*    tran21 - transpose local 2D array
C*    tran213 - transpose first 2 dimensions local 3D array
C*    tran132 - transpose last 2 dimensions local 3D array
C*
C*    The above routines with suffix 'c' handle complex arrays.
C*
C*-------------------------------------------------------------------
C
C--------------------------------------------------------------------
C
	subroutine dtran21( n1, n2, a, sa )
c
c  Transpose a 2D distributed array, where the 2nd dimension is
c    spread across processors.  In following comments, 'p' is the
c    number of processors as returned by 'jnodes()'.
c  n1,n2 - input array dimensions. Both must be divisible by p.
c  a(n1,n2/p) - input array. Transposed to a(n2,n1/p) on output.
c  sa - work array, length n1/p * n2/p
c
c  We have 3 transpose operations available:
c   tran21 - transpose local (n1,n2) to (n2,n1)
c  tran132 - transpose local (n1,n2,n3) to (n1,n3,n2)
c  jtran23 - transpose last two dimensions of distributed array
c            (n1,p,p), where p is the number of processors.
c
c  Treat as a 3D array distributed over the last dimension.
c  This can be thought of as 'strips' on each processor. 
c         n1 . n2  ->  n1 . n2/p | p
c          x    y       x     y1  y2
c  Use local transpose to exchange x and y1:
c         n1 . n2/p | p  ->  n2/p . n1 | p
c          x     y1  y2        y1    x  y2
c  Now expand the x direction, which can be thought of as 
c  creating a set of 'tiles' on each processor. Use 'jtran23' to
c  transpose across processors:
c         n2/p . n1/p . p | p  ->  n2/p . n1/p . p | p
c           y1     x1  x2  y2        y1     x1  y2  x2
c  Then use tran132 to collect y direction:
c         n2/p . n1/p . p | p  ->  n2/p . p . n1/p | p
c           y1     x1  y2  x2        y1  y2     x1  x2
c  Collapse y to yield transposed array:
c         n2 . n1/p | p
c          y     x1  x2
c
	implicit none
	integer n1, n2, p, node
	real a(*), sa(*)
	integer n1p, n2p
	integer jnode, jnodes

C  Processor numbers
	p = jnodes()
	node = jnode()

C  Strip and tile dimensions
	n1p = n1/p
	n2p = n2/p

C  Local transpose x and y1:
	call tran21( n1, n2p, a )

C  Return if 1 node
	if (p.eq.1) return

C  Global transpose x2 and y2
	call jtran23( n1p * n2p, a, sa )

C  Local transpose x1 and y2
	call tran132( n2p, n1p, p, a, sa )

C  That's it !
	return
	end
C
C-----------------------------------------------------------------------
C
	subroutine dtran21c( n1, n2, a, sa )
c
c  Complex version of distributed matrix transpose
c  See dtran21 for comments.
c
	implicit none
	integer n1, n2, p, node
	complex a(*), sa(*)
	integer n1p, n2p
	integer jnode, jnodes

C  Processor numbers
	p = jnodes()
	node = jnode()

C  Strip and tile dimensions
	n1p = n1/p
	n2p = n2/p

C  Local transpose x and y1:
	call tran21c( n1, n2p, a )
	if (p.eq.1) return

C  Global transpose x2 and y2
	call jtran23( 2 * n1p * n2p, a, sa )

C  Local transpose x1 and y2
	call tran132( 2*n2p, n1p, p, a, sa )

C  That's it !
	return
	end
C
C--------------------------------------------------------------------
C
	subroutine dtran132( n1, n2, n3, a, sa )
c
c  Transpose second and 3rd dimensions of a 3D distributed array
c    with last dimension spread across processors.
c  n1 - leading dimension of array
c  n2,n3 - Second and third dimensions. Both must be divisible by p.
c  p - number of processors
c  a(n1,n2,n3/p) - input array. Transposed to a(n1,n3,n2/p) on output.
c  sa - work array, length n1 * n2/p * n3/p
c  node - node number
c
	implicit none
	integer n1, n2, n3, p, node
	real a(*), sa(*)
	integer n2p,n3p
	integer jnode, jnodes

C  Processor numbers
	p = jnodes()
	node = jnode()

C  Handle single processor case
	if (p.eq.1) then
		call tran132( n1, n2, n3, a, sa )
		return
	endif

C  Strip and tile dimensions
	n2p = n2/p
	n3p = n3/p

c  Treat as a multi-D array distributed over the last dimension.
c         n1 . n2/p . p . n3/p | p
c          x     y1  y2     z1  z2

c  Use local transpose to exchange y2 and z1:
	call tran132( n1*n2p, p, n3p, a, sa )

c         n1 . n2/p . n3/p . p | p
c          x     y1     z1  y2  z2

c  Transpose y2 and z2 across processors:
	call jtran23( n1*n2p*n3p, a, sa )

c         n1 . n2/p . n3 | p
c          x     y1    z  y2

c  Locally transpose y1 and z
	call tran132( n1, n2p, n3, a, sa )

c  All done
c         n1 . n3 . n2/p | p
c          x    z     y1  y2

	return
	end
C
C--------------------------------------------------------------------
C
	subroutine dtran132c( n1, n2, n3, a, sa )
c
c  Complex version of 3D distributed transpose.  See dtran132 for
c    commments.
c
	implicit none
	integer n1,n2,n3
	real a(*),sa(*)
	
	call dtran132( 2*n1, n2, n3, a, sa )

	return
	end
C
C-----------------------------------------------------------------------
C
	subroutine tran213( n1, n2, n3, a )
C
C  Locally tranpose first and second dimensions of a 3D array
C
	implicit none
	integer n1,n2,n3
	real a(n1,n2,n3)
	integer k

	do k=1,n3
		call tran21( n1, n2, a(1,1,k) )
	enddo

	return
	end

C
C-----------------------------------------------------------------------
C
	subroutine tran213c( n1, n2, n3, a )
C
C  Locally tranpose first and second dimensions of a complex 3D array
C
	implicit none
	integer n1, n2, n3
	complex a(n1,n2,n3)
	integer k

	do k=1,n3
		call tran21c( n1, n2, a(1,1,k) )
	enddo

	return
	end

C
C-----------------------------------------------------------------------
C
	subroutine tran21( n, m, a )
C
C  In-place prime factorization transpose for real matrices
C
C  n,m - input array dimensions
C  a - input (n,m) array, on output (m,n)
C
	dimension          a(*),ipf(13),iexp(13),ipwr(13),itemp(13)
	real               a,temp,temp1
C  Return for unity matrices  
	if (n .le. 1 .or. m .le. 1) return
C 
	nm = n*m-1
	if (n .eq. m) then
C  Handle square matrices as separate case
		k = 2
		np1 = n+1
		nm1 = n-1
		do i=n,nm,n
		   l = k+nm1
		   do j=k,i
		      temp = a(j)
		      a(j) = a(l)
		      a(l) = temp
		      l = l+n
			enddo
		   k = k+np1
		enddo
		return
	endif
C  transposition of an n by m matrix amounts to
C  replacing the element at vector position i with the element at
C  position n*i (mod n*m-1). each subcycle of this permutation is
C  completed in order.
   15 call pfact(nm,npf,ipf,iexp,ipwr)
	do 20 i=1,npf
	   itemp(i) = 0
   20 continue
C                                  generate divisors of nm less than
C                                    nm/2
	id = 1
   25 if (id .ge. nm/2) return
	j = nm/id
	do 30 i=1,npf
	   if (itemp(i) .ne. iexp(i)) j = (j/ipf(i))*(ipf(i)-1)
   30 continue
C                                  the starting point of a subcycle is
C                                    divisible only by id and must not
C                                    appear in any other subcycle
	istart = id
   35 nmis = nm-istart
	if (istart .eq. id) go to 50
	isd = istart/id
	do 40 i=1,npf
	   if (itemp(i) .eq. iexp(i)) go to 40
	   if (mod(isd,ipf(i)) .eq. 0) go to 70
   40 continue
	it = istart
   45 it = mod(n*it,nm)
	if (it .lt. istart .or. it .gt. nmis) go to 70
	if (it .gt. istart .and. it .lt. nmis) go to 45
   50 temp = a(istart+1)
	temp1 = a(nmis+1)
	k = istart
   55 kk = mod(n*k,nm)
	l = nm-k
	ll = nm-kk
	j = j-2
C                                  move two elements. the second from
C                                    the negative subcycle. check to
C                                    see if the subcycle is complete.
	if (kk .eq. istart) go to 60
	if (ll .eq. istart) go to 65
C                                  subcycle not complete. move elements
C                                    and update pointers
	a(k+1) = a(kk+1)
	a(l+1) = a(ll+1)
	k = kk
	go to 55
C                                  subcycle complete. recompute id
   60 a(k+1) = temp
	a(l+1) = temp1
	go to 70
   65 a(k+1) = temp1
	a(l+1) = temp
   70 istart = istart+id
	if (j .gt. 0) go to 35
	do 80 i=1,npf
	   if (itemp(i) .eq. iexp(i)) go to 75
	   itemp(i) = itemp(i)+1
	   id = id*ipf(i)
	   go to 25
   75    itemp(i) = 0
	   id = id/ipwr(i)
   80 continue
 	return
	end
C
C-----------------------------------------------------------------------
C
	subroutine tran21c( n, m, a )
C
C  In-place prime factorization transpose for complex matrices
C
C  n,m - input array dimensions
C  a - input (n,m) array, on output (m,n)
C
	dimension a(*),ipf(13),iexp(13),ipwr(13),itemp(13)
	complex  a,temp,temp1
C  See tran21 for coments.
	if (n .le. 1 .or. m .le. 1) return
C 
	nm = n*m-1
	if (n .eq. m) then
		k = 2
		np1 = n+1
		nm1 = n-1
		do i=n,nm,n
		   l = k+nm1
		   do j=k,i
		      temp = a(j)
		      a(j) = a(l)
		      a(l) = temp
		      l = l+n
			enddo
		   k = k+np1
		enddo
		return
	endif
C
   15 call pfact(nm,npf,ipf,iexp,ipwr)
	do 20 i=1,npf
	   itemp(i) = 0
   20 continue
C
	id = 1
   25 if (id .ge. nm/2) return
	j = nm/id
	do 30 i=1,npf
	   if (itemp(i) .ne. iexp(i)) j = (j/ipf(i))*(ipf(i)-1)
   30 continue
C
	istart = id
   35 nmis = nm-istart
	if (istart .eq. id) go to 50
	isd = istart/id
	do 40 i=1,npf
	   if (itemp(i) .eq. iexp(i)) go to 40
	   if (mod(isd,ipf(i)) .eq. 0) go to 70
   40 continue
	it = istart
   45 it = mod(n*it,nm)
	if (it .lt. istart .or. it .gt. nmis) go to 70
	if (it .gt. istart .and. it .lt. nmis) go to 45
   50 temp = a(istart+1)
	temp1 = a(nmis+1)
	k = istart
   55 kk = mod(n*k,nm)
	l = nm-k
	ll = nm-kk
	j = j-2
C
	if (kk .eq. istart) go to 60
	if (ll .eq. istart) go to 65
C
	a(k+1) = a(kk+1)
	a(l+1) = a(ll+1)
	k = kk
	go to 55
C
   60 a(k+1) = temp
	a(l+1) = temp1
	go to 70
   65 a(k+1) = temp1
	a(l+1) = temp
   70 istart = istart+id
	if (j .gt. 0) go to 35
	do 80 i=1,npf
	   if (itemp(i) .eq. iexp(i)) go to 75
	   itemp(i) = itemp(i)+1
	   id = id*ipf(i)
	   go to 25
   75    itemp(i) = 0
	   id = id/ipwr(i)
   80 continue
 	return
	end
C
C---------------------------------------------------------------------
C
	subroutine tran132(ldim,n,m,a,wk)
C  
C  Transpose second and third dimensions of a 3D real array
C    using prime factor algorithm.
C  ldim - first dimension
C  n,m - second and third dimensions
C  a - real array, on input (ldim,n,m), on output (ldim,m,n)
C  wk - real work array, length ldim
C
	integer            ldim,n,m
	real               a(ldim,*),wk(*)
	dimension          ipf(13),iexp(13),ipwr(13),itemp(13)
C
	if (n .le. 1 .or. m .le. 1) go to 9005
C
	nm = n*m-1
	iw1 = ldim + 1
	if (n .ne. m) go to 15
C
	k = 2
	np1 = n+1
	nm1 = n-1
	do 10 i=n,nm,n
	   l = k+nm1
	   do 5 j=k,i
	      call scopy( ldim, a(1,j), 1, wk,     1 )
	      call scopy( ldim, a(1,l), 1, a(1,j), 1 )
	      call scopy( ldim,     wk, 1, a(1,l), 1 )
	      l = l+n
    5    continue
	   k = k+np1
   10 continue
	go to 9005
C
   15 call pfact(nm,npf,ipf,iexp,ipwr)
	do 20 i=1,npf
	   itemp(i) = 0
   20 continue
C
	id = 1
   25 if (id .ge. nm/2) go to 9005
	j = nm/id
	do 30 i=1,npf
	   if (itemp(i) .ne. iexp(i)) j = (j/ipf(i))*(ipf(i)-1)
   30 continue
C
	istart = id
   35 nmis = nm-istart
	if (istart .eq. id) go to 50
	isd = istart/id
	do 40 i=1,npf
	   if (itemp(i) .eq. iexp(i)) go to 40
	   if (mod(isd,ipf(i)) .eq. 0) go to 70
   40 continue
	it = istart
   45 it = mod(n*it,nm)
	if (it .lt. istart .or. it .gt. nmis) go to 70
	if (it .gt. istart .and. it .lt. nmis) go to 45
   50 continue
C
	call scopy( ldim, a(1,istart+1), 1, wk, 1 )
	call scopy( ldim, a(1,nmis+1), 1, wk(iw1), 1 )
C
	k = istart
   55 kk = mod(n*k,nm)
	l = nm-k
	ll = nm-kk
	j = j-2
C
	if (kk .eq. istart) go to 60
	if (ll .eq. istart) go to 65
C
	call scopy( ldim, a(1,kk+1), 1, a(1,k+1), 1 )
	call scopy( ldim, a(1,ll+1), 1, a(1,l+1), 1 )
C
	k = kk
	go to 55
C 
   60 continue
	call scopy( ldim, wk, 1, a(1,k+1), 1 )
	call scopy( ldim, wk(iw1), 1, a(1,l+1), 1 )
	go to 70
   65 continue
	call scopy( ldim, wk(iw1), 1, a(1,k+1), 1 )
	call scopy( ldim, wk, 1, a(1,l+1), 1 )
   70 istart = istart+id
	if (j .gt. 0) go to 35
	do 80 i=1,npf
	   if (itemp(i) .eq. iexp(i)) go to 75
	   itemp(i) = itemp(i)+1
	   id = id*ipf(i)
	   go to 25
   75    itemp(i) = 0
	   id = id/ipwr(i)
   80 continue
 9005 return
	end
C
C---------------------------------------------------------------------
C
	subroutine tran132c(ldim,n,m,a,wk)
C
C  Complex version of tran132
C
	implicit none
	integer ldim,n,m
	real a(*), wk(*)

	call tran132( 2*ldim, n, m, a, wk )

	return
	end
C
C-----------------------------------------------------------------------
C
	subroutine pfact  (n,npf,ipf,iexp,ipwr)
C
C  Return prime factors of an integer
C
C  n - input number to be factored
C  npf - output number of different prime factors
C  ipf - output prime factors
C  iexp - output exponents for each prime factor
C  ipwr - output value of ipf(i)**iexp(I)
C
	integer          ipf(13),iexp(13),ipwr(13)
C                                  first executable statement
	i = 0
	nf = iabs(n)
C                                  if n is -1, 0, or 1, set npf to 0
	if (nf .le. 1) go to 30
	ifc = 0
	id = 2
    5 iq = nf/id
	if (nf .ne. id*iq) go to 15
C                                  id is a factor of nf
	nf = iq
	if (id .le. ifc) go to 10
C                                  id is a new divisor
	i = i+1
C                                  update the prime factor vector,
C                                    the power vector, and set the
C                                    exponent to 1
	ipf(i) = id
	ipwr(i) = id
	iexp(i) = 1
	ifc = id
	go to 5
C                                  id factors n more than once. update
C                                    the power vector and the exponent
C                                    vector.
   10 ipwr(i) = id*ipwr(i)
	iexp(i) = iexp(i)+1
	go to 5
C                                  id is not a factor of nf. if nf is
C                                    not prime, continue to factor.
   15 if (iq .le. id) go to 25
C                                  update the divisor
	if (id .gt. 2) go to 20
	id = 3
	go to 5
   20 id = id+2
	go to 5
C                                  nf is prime or 1
   25 if (nf .le. 1) go to 30
C                                  nf is prime. update the prime
C                                    factor vector, the power vector,
C                                    and the exponent vector.
	i = i+1
	ipf(i) = nf
	ipwr(i) = nf
	iexp(i) = 1
   30 npf = i
	return
	end
