C*----------------------------------------------------------------------
C*
C*  NAME:
C*    vecsubs.f - BLAS and SEG Vector routines for systems
C*       that don't have these
C*     
C*  SYNOPSIS:
C*    integer n, inca, incb
C*    real x, a(n), b(n)
C*    complex ca(n), cb(n)
C*
C*    call scopy( n, a, inca, b, incb )
C*    call ccopy( n, ca, inca, cb, incb )
C*    call sscal( n, x, a, inca )
C*    call vfill( x, a, inca, n )
C*    call vsmul( a, inca, x, b, incb, n )
C*    call vsincos( a, b, n )
C*
C*  DESCRIPTION: 
C*    scopy and ccopy copy vectors with stride.
C*    vfill sets a vector to a constant with stride.
C*    vscsin computes cosine and sine of a vector with strine
C*
C*----------------------------------------------------------------------
C*
C*   HISTORY:
C*
C*      Written:
C*        12/31/91 - C. Mosher, ARCO E&PT 
C*
C*----------------------------------------------------------------------
C
	subroutine scopy( n, a, inca, b, incb )
C
C  Copy a vector with stride
C  n - vector length
C  a - input vector
C  inca - stride for vector a 
C  b - output vector
C  incb - stride for vector b 
C
	implicit none
	integer n, inca, incb, i, ia, ib
	real a(n), b(n)

C  If stride is negative, start from end of vector
	if (inca.lt.0) then
		ia = 1 + inca*(1-n)
	else
		ia = 1
	endif
	if (incb.lt.0) then
		ib = 1 + incb*(1-n)
	else
		ib = 1
	endif
		
C  Loop and copy from a to b
	do i=1,n
		b(ib) = a(ia)
		ia = ia + inca
		ib = ib + incb
	enddo

	return
	end
C
C*----------------------------------------------------------------------
C
	subroutine ccopy( n, a, inca, b, incb )
C
C  Copy a complex vector with stride
C  n - vector length
C  a - input vector
C  inca - stride for vector a 
C  b - output vector
C  incb - stride for vector b 
C
	implicit none
	integer n, inca, incb, i, ia, ib
	complex a(n), b(n)

C  If stride is negative, start from end of vector
	if (inca.lt.0) then
		ia = 1 + inca*(1-n)
	else
		ia = 1
	endif
	if (incb.lt.0) then
		ib = 1 + incb*(1-n)
	else
		ib = 1
	endif
		
C  Loop and copy from a to b
	do i=1,n
		b(ib) = a(ia)
		ia = ia + inca
		ib = ib + incb
	enddo

	return
	end
C*
C*----------------------------------------------------------------------
C
	subroutine sscal( n, x, a, inca )
C
C  Single precision vector scale
C  n - vector length
C  x - real scale factor
C  a - input/output vector
C  inca - stride for vector a 
C
	implicit none
	integer n, inca, i, ia
	real a(n), x

C  If stride is negative, start from end of vector
	if (inca.lt.0) then
		ia = 1 + inca*(1-n)
	else
		ia = 1
	endif
		
C  Loop and scale
	do i=1,n
		a(ia) = x * a(ia)
		ia = ia + inca
	enddo

	return
	end
C
C*----------------------------------------------------------------------
C
	subroutine vfill( x, a, inca, n )
C
C  Set a vector to a constant with stride
C  x - input constant
C  a - output vector
C  inca - stride for vector a 
C  n - vector length
C
	implicit none
	integer n, inca, i, ia
	real x, a(n)

C  If stride is negative, start from end of vector
	if (inca.lt.0) then
		ia = 1 + inca*(1-n)
	else
		ia = 1
	endif
		
C  Loop and set vector
	do i=1,n
		a(ia) = x
		ia = ia + inca
	enddo

	return
	end
C
C*----------------------------------------------------------------------
C
	subroutine vsmul( a, inca, x, b, incb, n )
C
C  Multiply a vector by a scalar
C  a - input vector
C  inca - stride for vector a 
C  x - input scalar
C  b - output vector
C  incb - stride for vector b 
C  n - vector length
C
	implicit none
	integer n, inca, incb, i, ia, ib
	real a(n), b(n), x

C  If stride is negative, start from end of vector
	if (inca.lt.0) then
		ia = 1 + inca*(1-n)
	else
		ia = 1
	endif
	if (incb.lt.0) then
		ib = 1 + incb*(1-n)
	else
		ib = 1
	endif
		
C  Loop and multipy
	do i=1,n
		b(ib) = x * a(ia)
		ia = ia + inca
		ib = ib + incb
	enddo

	return
	end
C
C*----------------------------------------------------------------------
C
	subroutine vsincos(a, c, n)
C  Simple interface to vscsin
C  a - input real vector
C  c - output complex vector, = cmplx(cos(a),sin(a))
C  n - length of a and c
	real a(*)
	complex c(*)
	call vscsin( a, 1, c, 2, n )
	return
	end
C
C*----------------------------------------------------------------------
C
	subroutine vscsin(a, ia, c, ic, n)
C
C  Compute cosine and sine of a vector
C  a - input real vector
C  ia - stride for vector a
C  c - output real vector, on output:
C      c(ic)=cos(a(ia)), c(ic+1)=sin(a(ia)) 
C  ic - stride for input output real vector (set to 2 for complex)
C  n - element count
C
	real a(*), c(*)
	integer ia, ic, n
C
	integer iia, iic
	iia = 1
	iic = 1
	do i=1, n
		c(iic) = cos(a(iia))
		c(iic+1) = sin(a(iia))
	    iic = iic + ic
	    iia = iia + ia
	enddo
c
	return
	end
C
C*----------------------------------------------------------------------
